/*
 * uPD77660 processor support - INTC hardware block
 *
 * Copyright (C) 2011  Renesas Electronics Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/io.h>
#include <linux/sh_intc.h>
#include <asm/hardware/gic.h>
#include <asm/mach-types.h>
#include <asm/mach/arch.h>

/* IRQ pin */
#define INTC_ADDR	0xe6900000
#define ICR1A		(INTC_ADDR + 0x00)	/* 32-bit */
#define ICR2A		(INTC_ADDR + 0x04)	/* 32-bit */
#define ICR3A		(INTC_ADDR + 0x08)	/* 32-bit */
#define ICR4A		(INTC_ADDR + 0x0c)	/* 32-bit */
#define INTREQ00A	(INTC_ADDR + 0x20)	/* 8-bit */
#define INTREQ10A	(INTC_ADDR + 0x24)	/* 8-bit */
#define INTREQ20A	(INTC_ADDR + 0x28)	/* 8-bit */
#define INTREQ30A	(INTC_ADDR + 0x2c)	/* 8-bit */
#define INTMSK00A	(INTC_ADDR + 0x40)	/* 8-bit */
#define INTMSK10A	(INTC_ADDR + 0x44)	/* 8-bit */
#define INTMSK20A	(INTC_ADDR + 0x48)	/* 8-bit */
#define INTMSK30A	(INTC_ADDR + 0x4c)	/* 8-bit */
#define INTMSKCLR00A	(INTC_ADDR + 0x60)	/* 8-bit */
#define INTMSKCLR10A	(INTC_ADDR + 0x64)	/* 8-bit */
#define INTMSKCLR20A	(INTC_ADDR + 0x68)	/* 8-bit */
#define INTMSKCLR30A	(INTC_ADDR + 0x6c)	/* 8-bit */

static DEFINE_SPINLOCK(irqpin_lock);

static int irqpin_set_irq_type(unsigned int irq, unsigned int type)
{
	u32 irqpin = irq - IRQPIN_IRQ_BASE;
	u32 shift = (~irqpin & 0x7) << 2;
	u32 mask, reg;

	switch (type & IRQ_TYPE_SENSE_MASK) {
	case IRQ_TYPE_EDGE_RISING:
		mask = 0x1 << shift;
		__set_irq_handler_unlocked(irq, handle_edge_irq);
		break;
	case IRQ_TYPE_EDGE_FALLING:
		mask = 0x0 << shift;
		__set_irq_handler_unlocked(irq, handle_edge_irq);
		break;
	case IRQ_TYPE_LEVEL_HIGH:
		mask = 0x3 << shift;
		__set_irq_handler_unlocked(irq, handle_level_irq);
		break;
	case IRQ_TYPE_LEVEL_LOW:
		mask = 0x2 << shift;
		__set_irq_handler_unlocked(irq, handle_level_irq);
		break;
	case IRQ_TYPE_EDGE_BOTH:
		mask = 0x4 << shift;
		__set_irq_handler_unlocked(irq, handle_edge_irq);
		break;
	default:
		return -EINVAL;
	}

	reg = (irqpin & 0x10) ? ((irqpin & 0x08) ? ICR4A : ICR3A) :
				((irqpin & 0x08) ? ICR2A : ICR1A);

	spin_lock(&irqpin_lock);
	__raw_writel((__raw_readl(reg) & ~(0xf << shift)) | mask, reg);
	spin_unlock(&irqpin_lock);

	return 0;
}

static void irqpin_irq_ack(unsigned int irq)
{
	u32 irqpin = irq - IRQPIN_IRQ_BASE;
	u8  mask = 1 << (~irqpin & 0x7);
	u32 reg = (irqpin & 0x10) ? ((irqpin & 0x08) ? INTREQ30A : INTREQ20A) :
				    ((irqpin & 0x08) ? INTREQ10A : INTREQ00A);

	__raw_writeb(~mask, reg);
}

static void irqpin_irq_mask(unsigned int irq)
{
	u32 irqpin = irq - IRQPIN_IRQ_BASE;
	u8  mask = 1 << (~irqpin & 0x7);
	u32 reg = (irqpin & 0x10) ? ((irqpin & 0x08) ? INTMSK30A : INTMSK20A) :
				    ((irqpin & 0x08) ? INTMSK10A : INTMSK00A);

	__raw_writeb(mask, reg);
}

static void irqpin_irq_unmask(unsigned int irq)
{
	u32 irqpin = irq - IRQPIN_IRQ_BASE;
	u8  mask = 1 << (~irqpin & 0x7);
	u32 reg = (irqpin & 0x10) ?
			((irqpin & 0x08) ? INTMSKCLR30A : INTMSKCLR20A) :
			((irqpin & 0x08) ? INTMSKCLR10A : INTMSKCLR00A);

	__raw_writeb(mask, reg);
}

static void irqpin_demux(unsigned int irq, struct irq_desc *desc)
{
	struct irq_chip *chip = get_irq_chip(irq);
	u32 status, irqpin_irq;
	u8 req0, req1, req2, req3;

	/* primary controller ack'ing */
	chip->ack(irq);

	do {
		req0 = __raw_readb(INTREQ00A) & ~(__raw_readb(INTMSK00A));
		req1 = __raw_readb(INTREQ10A) & ~(__raw_readb(INTMSK10A));
		req2 = __raw_readb(INTREQ20A) & ~(__raw_readb(INTMSK20A));
		req3 = __raw_readb(INTREQ30A) & ~(__raw_readb(INTMSK30A));

		status = (req0 << 24) | (req1 << 16) | (req2 << 8) | req3;
		if (status == 0)
			break;

		irqpin_irq = IRQPIN_IRQ_BASE;
		while (status) {
			if (status & (1UL << 31))
				generic_handle_irq(irqpin_irq);
			status <<= 1;
			irqpin_irq++;
		}
	} while (1);

	/* primary controller unmasking */
	chip->unmask(irq);
}

static struct irq_chip irqpin_chip = {
	.name     = "IRQ pin",
	.ack      = irqpin_irq_ack,
	.mask     = irqpin_irq_mask,
	.unmask   = irqpin_irq_unmask,
	.set_type = irqpin_set_irq_type,
	.disable  = irqpin_irq_mask,
};

static void setup_irqpin_irq(int base)
{
	int i;

	/* Load h/w initial value 4'b0000 (detected at a falling edge) */
	__raw_writel(0, ICR1A);
	__raw_writel(0, ICR2A);
	__raw_writel(0, ICR3A);
	__raw_writel(0, ICR4A);

	__raw_writeb(0, INTREQ00A);
	__raw_writeb(0, INTREQ10A);
	__raw_writeb(0, INTREQ20A);
	__raw_writeb(0, INTREQ30A);

	__raw_writeb(0xff, INTMSK00A);
	__raw_writeb(0xff, INTMSK10A);
	__raw_writeb(0xff, INTMSK20A);
	__raw_writeb(0xff, INTMSK30A);

	for (i = base; i < base + 32; i++) {
		set_irq_chip(i, &irqpin_chip);
		set_irq_handler(i, handle_edge_irq);
		set_irq_flags(i, IRQF_VALID);
	}
}

static int sh7740_set_wake(unsigned int irq, unsigned int on)
{
	return 0; /* always allow wakeup */
}

#define INTA_CTRL	0xe605807c

void __init sh7740_init_irq(void)
{
	void __iomem *gic_dist_base = __io(0xf2800000);
	void __iomem *gic_cpu_base = __io(0xf2000000);

	/* Change INT_SEL INTCA->GIC */
	__raw_writel(__raw_readl(INTA_CTRL) & ~0x2, INTA_CTRL);

	gic_init(0, 29, gic_dist_base, gic_cpu_base);
	gic_arch_extn.set_wake = sh7740_set_wake;

	/* Setup IRQ-pin cascade_irq */
	setup_irqpin_irq(IRQPIN_IRQ_BASE);
	set_irq_chained_handler(gic_spi(149), irqpin_demux);
}
