/*
 * Function        : Composer driver for SH Mobile
 *
 * Copyright (C) 2011 Renesas Electronics Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 */

#include <linux/module.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/kernel.h>
#include <linux/io.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/wait.h>
#include <linux/semaphore.h>
#include <linux/rwlock.h>
#include <linux/list.h>
#include <linux/kthread.h>

#include <rtapi/screen_graphics.h>

#include <linux/sh_mobile_composer.h>

static int debug = 1;    /* default debug level */

#define   INTERNAL_DEBUG   0	/* enable debug interface        */
				/*   0: none                     */
				/*   1: enable debug.            */
				/*   2: enable debug with trace. */

#define   _TIME_DBG  0		/* generate report processing time. */
#define   _LOG_DBG   0		/* generate debug log.              */
#define   _ERR_DBG   2		/* generate error log.              */



/******************************************************/
/* define protptype                                   */
/******************************************************/
static ssize_t core_read(struct file *filp, char __user *buf, \
		size_t sz, loff_t *off);
static unsigned int core_poll(struct file *filp, \
		struct poll_table_struct *poll);
static int core_ioctl(struct inode *inode, struct file *filep, \
		unsigned int cmd, unsigned long arg);
static int core_open(struct inode *inode, struct file *filep);
static int core_release(struct inode *inode, struct file *filep);


static void work_runblend(struct work_struct *work);
static int  release_buffer(struct composer_fh *fh, int level);
static int  lock_buffer(struct composer_fh *fh, int level);
static int  unlock_buffer(struct composer_fh *fh, int level);
static void wakeup_allfile(struct composer_buffer *buffer);

static int  setbusy(struct composer_fh *fh);
static int  setidle(struct composer_fh *fh);

static int    get_index_from_level(int level, int *idx);
static struct composer_info *get_composer_info(
	struct composer_fh *fh, int level);


/* ioctl */
static int  iocs_close(struct composer_fh *fh, \
	struct cmp_lay_close *arg);

/* parameter check */
static int  chk_ioc_supportpixfmt(struct composer_fh *fh, \
	struct cmp_lay_supportpixfmt *arg);
static int  chk_ioc_open(struct composer_fh *fh, \
	struct cmp_lay_open *arg);
static int  chk_ioc_close(struct composer_fh *fh, \
	struct cmp_lay_close *arg);
static int  chk_ioc_disp(struct composer_fh *fh, \
	struct cmp_lay_disp *arg);
static int  chk_ioc_alpha(struct composer_fh *fh, \
	struct cmp_lay_alpha *arg, int set_flag);
static int  chk_ioc_keycolor(struct composer_fh *fh, \
	struct cmp_lay_keycolor *arg, int set_flag);
static int  chk_ioc_laysize(struct composer_fh *fh, \
	struct cmp_lay_data_size *arg, int set_flag);
static int  chk_ioc_pixfmt(struct composer_fh *fh, \
	struct cmp_lay_pixfmt *arg, int set_flag);
static int  chk_ioc_cropsize(struct composer_fh *fh, \
	struct cmp_lay_cropsize *arg, int set_flag, \
	struct cmp_lay_data_size *laysize);
static int  chk_ioc_croppos(struct composer_fh *fh, \
	struct cmp_lay_croppos *arg, int set_flag, \
	struct cmp_lay_data_size *laysize);
static int  chk_ioc_pos(struct composer_fh *fh, \
	struct cmp_lay_pos *arg, int set_flag);
static int  chk_ioc_size(struct composer_fh *fh, \
	struct cmp_lay_compose_size *arg, int set_flag);
static int  chk_ioc_viewlay(struct composer_fh *fh, \
	struct cmp_viewlay *arg, int idx);
static int  chk_ioc_backcolor(struct composer_fh *fh, \
	struct cmp_lay_backcolor *arg, int set_flag);
static int  chk_ioc_layaddr(struct composer_fh *fh, \
	struct cmp_layaddr *arg, int set_flag);
static int  chk_ioc_start(struct composer_fh *fh);

static void notify_graphics_image_conv(int result, unsigned long user_data);
static void notify_graphics_image_blend(int result, unsigned long user_data);

/******************************************************/
/* define local define                                */
/******************************************************/
#if INTERNAL_DEBUG >= 2
#error not supported.
#endif
#define TRACE_ENTER(FH, ID)
#define TRACE_LEAVE(FH, ID)
#define TRACE_LOG(FH, ID)
#define TRACE_LOG1(FH, ID, VAL1)

#define MAX_OPEN    32
#define MAX_BUFFER  64

#define DEV_NAME      "composer"

/* define for id of struct composer_buffer */
#define BUF_ID_FLG_PIXFMT    0x001
#define BUF_ID_FLG_LAYSIZE   (BUF_ID_FLG_LAYSIZE1|BUF_ID_FLG_LAYSIZE2)
#define BUF_ID_FLG_LAYSIZE1  0x002
#define BUF_ID_FLG_LAYSIZE2  0x004
#define BUF_ID_FLG_CROPSIZE  (BUF_ID_FLG_CROPSIZE1|BUF_ID_FLG_CROPSIZE2)
#define BUF_ID_FLG_CROPSIZE1 0x008
#define BUF_ID_FLG_CROPSIZE2 0x010
#define BUF_ID_FLG_CROPPOS   (BUF_ID_FLG_CROPPOS1|BUF_ID_FLG_CROPPOS2)
#define BUF_ID_FLG_CROPPOS1  0x020
#define BUF_ID_FLG_CROPPOS2  0x040
#define BUF_ID_FLG_POS       (BUF_ID_FLG_POS1|BUF_ID_FLG_POS2)
#define BUF_ID_FLG_POS1      0x080
#define BUF_ID_FLG_POS2      0x100
#define BUF_ID_FLG_SIZE      (BUF_ID_FLG_SIZE1|BUF_ID_FLG_SIZE2)
#define BUF_ID_FLG_SIZE1     0x200
#define BUF_ID_FLG_SIZE2     0x400
#define BUF_ID_SET_BUFFER    0x800

#define BUF_ID_FLG_OUTPLANE   (0x01<<16)


/* define for id of struct composer_info */
#define INFO_ID_OUTPLANE    (0x00<<24)
#define INFO_ID_MAINLAYER   (0x01<<24)
#define INFO_ID_CH1LAYER    (0x02<<24)
#define INFO_ID_CH2LAYER    (0x03<<24)
#define INFO_ID_CH3LAYER    (0x04<<24)
#define INFO_ID_SET_APP0     0x02
#define INFO_ID_SET_APP1     0x04
#define INFO_ID_SET_APP2     0x08
#define INFO_ID_FLG_BLENDING    0x20
#define INFO_ID_FLG_NEEDAPPSET  0x40
#define INFO_ID_FLG_NOTDISP     0x80         /* control by CMP_IOCS_DISP */

/* define for fh_status of struct composer_fh */
#define FH_STATUS_OUTPLANE  (0x01<<16)
#define FH_STATUS_MAINLAYER (0x02<<16)
#define FH_STATUS_CH1LAYER  (0x04<<16)
#define FH_STATUS_CH2LAYER  (0x08<<16)
#define FH_STATUS_CH3LAYER  (0x10<<16)
#define FH_STATUS_BLENDING  (0x01)
#define FH_STATUS_BLENDERR  (0x02)
#define FH_STATUS_POLLBLEND (0x01<<24)
#define FH_STATUS_POLLADDR  (0x02<<24)

/* define for index number of struct composer_fh */
#define IDX_OUTPLANE    0x00
#define IDX_MAINLAYER   0x01
#define IDX_CH1LAYER    0x02
#define IDX_CH2LAYER    0x03
#define IDX_CH3LAYER    0x04

/* macros for general error message */
#if _ERR_DBG >= 2
#define printk_err2(fmt, arg...) \
	do { \
		printk(KERN_ERR DEV_NAME ":E %s: " fmt, __func__, ## arg); \
	} while (0)
#else
#define printk_err2(fmt, arg...)
#endif

/* macros for RT-API related error message */
#if _ERR_DBG >= 1
#define printk_err1(fmt, arg...) \
	do { \
		printk(KERN_ERR DEV_NAME ":E %s: " fmt, __func__, ## arg); \
	} while (0)
#else
#define printk_err1(fmt, arg...)
#endif

#define printk_err(fmt, arg...) \
	do { \
		printk(KERN_ERR DEV_NAME ":E %s: " fmt, __func__, ## arg); \
	} while (0)

/* macros for general log message */
#if _LOG_DBG >= 2
#define printk_dbg2(level, fmt, arg...) \
	do { \
		if (level <= debug) \
			printk(KERN_INFO DEV_NAME ": %s: " fmt, \
				__func__, ## arg); \
	} while (0)
#else
#define printk_dbg2(level, fmt, arg...)
#endif

/* macros for RT-API log message */
#if _LOG_DBG >= 1
#define printk_dbg1(level, fmt, arg...) \
	do { \
		if (level <= debug) \
			printk(KERN_INFO DEV_NAME ": %s: " fmt, \
				__func__, ## arg); \
	} while (0)
#else
#define printk_dbg1(level, fmt, arg...)
#endif

#define printk_dbg(level, fmt, arg...) \
	do { \
		if (level <= debug) \
			printk(KERN_INFO DEV_NAME ": %s: " fmt, \
				__func__, ## arg); \
	} while (0)


/******************************************************/
/* define local variables                             */
/******************************************************/
static const struct file_operations composer_fops = {
	.owner		= THIS_MODULE,
	.read		= core_read,
	.write		= NULL,
	.poll		= core_poll,
	.ioctl		= core_ioctl,
	.open		= core_open,
	.release	= core_release,
};

static struct miscdevice composer_device = {
	.minor = MISC_DYNAMIC_MINOR,
	.name = "composer",
	.fops = &composer_fops,
};

static struct composer_buffer local_buffer[MAX_BUFFER];
static spinlock_t             irqlock;
static DECLARE_MUTEX(sem);
static int                    num_open;
static LIST_HEAD(file_top);
static LIST_HEAD(appshare_top);
static struct workqueue_struct *workqueue;

/******************************************************/
/* local functions                                    */
/******************************************************/
static void initialize_bufferinformation(struct composer_buffer *buf)
{
	spin_lock_init(&buf->buf_lock);
	sema_init(&buf->buf_sem, 1);

	INIT_LIST_HEAD(&buf->buf_link.list);
	buf->buf_link.body = NULL; /* no buffer binding */
	buf->buf_link.fh   = NULL; /* no buffer binding */

	buf->buf_id          = 0;
	buf->keycolor        = CMPKEYCOLOR_OFF;
	buf->backcolor       = 0;
	buf->alpha           = 255;
	buf->data_fmt        = CMPPIXFMT_ARGB8888;
	buf->col_space       = CMPYUVCOLOR_BT601_COMPRESS;
	buf->crop_pos_x      = 0;
	buf->crop_pos_y      = 0;
	buf->crop_size_x     = 0;
	buf->crop_size_y     = 0;
};

static struct composer_fh *allocate_device(void)
{
	struct composer_fh *fh;
	int    i;

	fh = kmalloc(sizeof(*fh), GFP_KERNEL);
	if (NULL == fh)
		return NULL;

	/* initialize handle */
	memset(fh, 0, sizeof(*fh));

	INIT_WORK(&fh->fh_wqtask, work_runblend);

	INIT_LIST_HEAD(&fh->fh_filelist);
	list_add_tail(&fh->fh_filelist, &file_top);

	sema_init(&fh->fh_sem, 1);

	rwlock_init(&fh->fh_rwlock);

	init_waitqueue_head(&fh->fh_wait);
	init_waitqueue_head(&fh->grap_data.wait_notify);

	for (i = 0; i < 5; i++) {
		struct composer_info     *fh_info   = &fh->fh_info[i];
		struct composer_buffer   *fh_buffer = &fh->fh_buffer_id0[i];

		/* initialize open buffer information */

		INIT_LIST_HEAD(&fh_info->wqlink.list);
		fh_info->wqlink.body = &fh->fh_wait;
		fh_info->wqlink.fh   = fh;

		fh_info->id         = 0;
		fh_info->buffer     = NULL;
		fh_info->appinfo[0] = NULL;
		fh_info->appinfo[1] = NULL;
		fh_info->appinfo[2] = NULL;

		/* initialize local buffer information */
		initialize_bufferinformation(fh_buffer);
	}

	fh->fh_status = 0;

	return fh;
}

static void  free_device(struct composer_fh *fh)
{
	int    i;
	static const int fh_status_translate[] = {
		[IDX_OUTPLANE]  = FH_STATUS_OUTPLANE,
		[IDX_MAINLAYER] = FH_STATUS_MAINLAYER,
		[IDX_CH1LAYER]  = FH_STATUS_CH1LAYER,
		[IDX_CH2LAYER]  = FH_STATUS_CH2LAYER,
		[IDX_CH3LAYER]  = FH_STATUS_CH3LAYER
	};
	static const  int level[5] = {
		[IDX_OUTPLANE]  = CMP_OUT_PLANE,
		[IDX_MAINLAYER] = CMP_MAIN_LAYER,
		[IDX_CH1LAYER]  = CMP_CH1_LAYER,
		[IDX_CH2LAYER]  = CMP_CH2_LAYER,
		[IDX_CH3LAYER]  = CMP_CH3_LAYER };

	printk_dbg2(2, "down");
	down(&fh->fh_sem);

	/* release open buffer */
	for (i = 0; i < ARRAY_SIZE(level); i++) {
		struct cmp_lay_close close;

		if (fh->fh_status & fh_status_translate[i]) {
			printk_dbg2(2, "index %d closing.", level[i]);
			close.level = level[i];
			iocs_close(fh, &close);
		} else {
			printk_dbg2(2, "index %d not open.", level[i]);
		}
	}
	list_del_init(&fh->fh_filelist);

	/* unlock semaphore */
	up(&fh->fh_sem);
	wake_up_interruptible_all(&fh->fh_wait);
	kfree(fh);
}

static int    get_index_from_level(int level, int *idx)
{
	static const int fh_level_translate[] = {
		[CMP_OUT_PLANE]  = IDX_OUTPLANE,
		[CMP_MAIN_LAYER] = IDX_MAINLAYER,
		[CMP_CH1_LAYER]  = IDX_CH1LAYER,
		[CMP_CH2_LAYER]  = IDX_CH2LAYER,
		[CMP_CH3_LAYER]  = IDX_CH3LAYER
	};
	if (level < 0 || level >= ARRAY_SIZE(fh_level_translate)) {
		printk_err2("argument level %d invalid.\n", level);
		return CMP_ERR_NG;
	}
	*idx = fh_level_translate[level];
	return CMP_ERR_OK;
}

static struct composer_info *get_composer_info(
	struct composer_fh *fh, int level)
{
	int                      idx;
	static const int fh_status_translate[] = {
		[CMP_OUT_PLANE]  = FH_STATUS_OUTPLANE,
		[CMP_MAIN_LAYER] = FH_STATUS_MAINLAYER,
		[CMP_CH1_LAYER]  = FH_STATUS_CH1LAYER,
		[CMP_CH2_LAYER]  = FH_STATUS_CH2LAYER,
		[CMP_CH3_LAYER]  = FH_STATUS_CH3LAYER
	};

	if (get_index_from_level(level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", level);
		return NULL;
	}

	if ((fh->fh_status &  fh_status_translate[level]) == 0) {
		printk_err2("level %d not opend.\n", level);
		return NULL;
	}

	return &fh->fh_info[idx];
}

static int  bind_buffer(struct composer_fh *fh, int idx, int id)
{
	struct composer_buffer *buf;
	struct composer_info   *info;
	unsigned long flags;

	printk_dbg(2, "bind_buffer: idx:%d id:%d\n", idx, id);

	if (id == 0) {
		/* use private buffer */
		buf = &fh->fh_buffer_id0[idx];
	} else {
		/* use global buffer */
		buf = &local_buffer[id];
	}

	if (idx == IDX_OUTPLANE) {
		if (buf->buf_id & BUF_ID_FLG_OUTPLANE) {
			printk_err2("buffer already opened for output.\n");
			return -EAGAIN;
		}
	}

	info   = &fh->fh_info[idx];

	/* assertion trap */
	BUG_ON(info->buffer != NULL);

	printk_dbg2(2, "spinlock");
	spin_lock_irqsave(&buf->buf_lock, flags);

	/* set bind to buffer */
	info->id       = (idx<<24) | (id<<16) | 0x0;
	info->buffer   = buf;

	spin_unlock_irqrestore(&buf->buf_lock, flags);

	/* lock buffer. */
	if (lock_buffer(fh, idx) < 0) {
		printk_err2("lock_buffer interrupted.\n");

		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buf->buf_lock, flags);
		info->id       = 0;
		info->buffer   = NULL;
		spin_unlock_irqrestore(&buf->buf_lock, flags);

		return -EAGAIN;
	}

	printk_dbg2(2, "spinlock");
	spin_lock_irqsave(&buf->buf_lock, flags);

	/* buffer bind to info */
	list_add_tail(&info->wqlink.list, &buf->buf_link.list);
	if (idx == IDX_OUTPLANE) {
		/* mark this buffer opend for output */
		buf->buf_id |= BUF_ID_FLG_OUTPLANE;
	}

	spin_unlock_irqrestore(&buf->buf_lock, flags);

	unlock_buffer(fh, idx);

	return 0;
}

static int  release_buffer(struct composer_fh *fh, int level)
{
	struct composer_info   *info;
	struct composer_buffer *buffer;
	struct appmem_handle   *app[3];
	int                  info_id;
	unsigned long flags;
	int           i;
	int           idx;

	printk_dbg(2, "release buffer: idx:%d\n", level);

	if (get_index_from_level(level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", level);
		return -EINVAL;
	}

	info   = &fh->fh_info[idx];
	buffer = info->buffer;

	BUG_ON(NULL == buffer);

	if (0 != buffer->buf_sem.count) {
		/* report bug */
		BUG_ON(0 != buffer->buf_sem.count);
	}

	/* acqure lock of buffer */
	printk_dbg2(2, "spinlock");
	spin_lock_irqsave(&buffer->buf_lock, flags);

	/* reset buffer info */
	info_id = info->id;
	app[0]  = info->appinfo[0];
	app[1]  = info->appinfo[1];
	app[2]  = info->appinfo[2];

	info->id = 0;
	info->buffer = NULL;
	info->appinfo[0] = NULL;
	info->appinfo[1] = NULL;
	info->appinfo[2] = NULL;

	/* remove link to current buffer info */
	list_del_init(&info->wqlink.list);

	if (list_empty(&buffer->buf_link.list)) {
		printk_dbg2(2, "clear CMP_IOCS_LAYADDR setting.\n");
		buffer->buf_id &= ~BUF_ID_SET_BUFFER;
	}
	if (idx == IDX_OUTPLANE) {
		/* reset mark of buffer opend for output */
		buffer->buf_id &= ~BUF_ID_FLG_OUTPLANE;
	}

	spin_unlock_irqrestore(&buffer->buf_lock, flags);

	/*****************************************/
	/* unlock_buffer(fh, idx) can not used.  */
	/* threafore, implement unlock directly. */
	/*****************************************/
	up(&buffer->buf_sem);

	/* notify waiting thread */
	wakeup_allfile(buffer);

	/* wakeup current handle */
	wake_up_interruptible(&fh->fh_wait);

	/* release appShare memory */
	for (i = 0; i < ARRAY_SIZE(app); i++) {
		int rc = 0;
		if (app[i]) {
			rc = sh_mobile_appmem_free(app[i]);
			/* ignore error, because recovery from error \
			   is only available by reboot-kernel.       */
		}
		if (rc) {
			/* report error */
			printk_err("sh_mobile_appmem_free return error\n");
		}
	}
	return 0;
}

static int  lock_buffer(struct composer_fh *fh, int level)
{
	struct composer_info   *info   = &fh->fh_info[level];
	struct composer_buffer *buffer = info->buffer;

	BUG_ON(NULL == buffer);

	while (down_trylock(&buffer->buf_sem)) {
		int rc;
		do {
			printk_dbg2(2, "retrying trylock\n");
			rc = wait_event_interruptible_timeout(
				fh->fh_wait,
				buffer->buf_sem.count > 0,
				1*HZ);
		} while (rc == 0);
		if (rc < 0) {
			printk_err2("buffer lock interrupted.\n");
			return rc;
		}
	}

	/* should acquire semaphore. */
	BUG_ON(0 != buffer->buf_sem.count);

	return 0;
}
static int  unlock_buffer(struct composer_fh *fh, int level)
{
	struct composer_info   *info   = &fh->fh_info[level];
	struct composer_buffer *buffer = info->buffer;

	if (NULL == buffer) {
		printk_err2("buffer level %d not binding\n", level);
		return 0;
	}

	if (0 != buffer->buf_sem.count) {
		/* report bug */
		printk_err2("currently not acquire buffer lock\n");
		BUG_ON(0 != buffer->buf_sem.count);
	}

	up(&buffer->buf_sem);

	/* notify waiting thread */
	wakeup_allfile(buffer);

	return 0;
}
static void  wakeup_allfile(struct composer_buffer *buffer)
{
	unsigned long flags;
	struct list_head *list;

	if (NULL == buffer) {
		printk_err2("buffer not opened.\n");
		return;
	}

	printk_dbg2(2, "spinlock");
	spin_lock_irqsave(&buffer->buf_lock, flags);

	/* wakeup waiting thread */
	list_for_each(list, &buffer->buf_link.list)
	{
		struct bufferlink *buflink;

		buflink = list_entry((void *)list, struct bufferlink, list);

		printk_dbg2(2, "wakeup queue ptr:%p\n", buflink->body);
		if (buflink->body) {
			/* wake up waiting thread. */
			wake_up_interruptible_all(buflink->body);
		}
	}

	spin_unlock_irqrestore(&buffer->buf_lock, flags);

	return;
}


static void allfile_status_set(struct composer_buffer *buffer, int status)
{
	unsigned long flags;
	struct list_head *list;

	if (NULL == buffer) {
		printk_err2("buffer not opened.\n");
		return;
	}

	printk_dbg2(2, "spinlock");
	spin_lock_irqsave(&buffer->buf_lock, flags);

	printk_dbg2(2, "status 0x%x append to file\n", status);
	/* wakeup waiting thread */
	list_for_each(list, &buffer->buf_link.list)
	{
		struct bufferlink  *buflink;
		struct composer_fh *fh;

		buflink = list_entry((void *)list, struct bufferlink, list);

		fh = buflink->fh;
		if (fh) {
			fh->fh_status |= status;
			printk_dbg2(2, "set status %x to fh:%p\n", status, fh);
		}
	}

	spin_unlock_irqrestore(&buffer->buf_lock, flags);

	return;
}



static int update_work_linebytes(
	struct composer_grapdata *grap, int idx, struct composer_buffer *buffer)
{
	int c[3], v[3];
	int w;
	if ((buffer->buf_id & BUF_ID_FLG_PIXFMT) != BUF_ID_FLG_PIXFMT) {
		printk_dbg2(2, "pixefmt not initialized.");
		return CMP_ERR_NG;
	}
	if ((buffer->buf_id & BUF_ID_FLG_LAYSIZE) != BUF_ID_FLG_LAYSIZE) {
		printk_dbg2(2, "laysize not initialized.");
		return CMP_ERR_NG;
	}
	if (idx >= ARRAY_SIZE(grap->work_linebyte)) {
		printk_err2("idx invalid.");
		BUG_ON(idx >= ARRAY_SIZE(grap->work_linebyte));
		return CMP_ERR_NG;
	}

	/* set pixel data size. fixpoint format 31.1 */
	switch (buffer->data_fmt) {
	case CMPPIXFMT_ARGB8888:
	default:
		c[0] = 4*2; c[1] = 0; c[2] = 0; break;
	case CMPPIXFMT_RGB888:
		c[0] = 3*2; c[1] = 0; c[2] = 0; break;
	case CMPPIXFMT_RGB565:
		c[0] = 2*2; c[1] = 0; c[2] = 0; break;
	case CMPPIXFMT_YUV420SP:
	case CMPPIXFMT_YUV422SP:
		c[0] = 1*2; c[1] = 1*2; c[2] = 0; break;
	case CMPPIXFMT_YUV420PL:
		c[0] = 1*2; c[1] = 1; c[2] = 1; break;
	}
	printk_dbg2(2, "data_fmt:%d  " \
		"its pixel size is [31.1] %d, %d, %d\n", \
		buffer->data_fmt, c[0], c[1], c[2]);

	/* calculate linebyte */
	w = buffer->data_x;
	v[0] = w * c[0]/2;
	v[1] = w * c[1]/2;
	v[2] = w * c[2]/2;

	printk_dbg2(2, "linebyte: %d, %d, %d\n", \
		v[0], v[1], v[2]);

	grap->work_linebyte[idx]    = v[0];
	grap->work_linebyte_c0[idx] = v[1];
	grap->work_linebyte_c1[idx] = v[2];
	grap->work_pixelsize[idx]    = c[0];
	grap->work_pixelsize_c0[idx] = c[1];
	grap->work_pixelsize_c1[idx] = c[2];
	return CMP_ERR_OK;
}

static int update_work_imagesize(
	struct composer_grapdata *grap, int idx, struct composer_buffer *buffer)
{
	int w, h;
	int c_x, c_y, c_w, c_h;
	int rc = CMP_ERR_OK;

	if ((buffer->buf_id & BUF_ID_FLG_LAYSIZE) != BUF_ID_FLG_LAYSIZE) {
		printk_dbg2(2, "laysize not initialized.");
		rc = CMP_ERR_NG;
		goto err_exit;
	}
	if (idx >= ARRAY_SIZE(grap->work_linebyte)) {
		printk_err2("idx invalid.");
		BUG_ON(idx >= ARRAY_SIZE(grap->work_linebyte));
		rc = CMP_ERR_NG;
		goto err_exit;
	}

	w = buffer->data_x;
	h = buffer->data_y;

	/* set default crop */
	c_x = 0;
	c_y = 0;
	c_w = w;
	c_h = h;

	/* update crop */
	if ((buffer->buf_id & BUF_ID_FLG_CROPPOS) == BUF_ID_FLG_CROPPOS) {
		c_x = buffer->crop_pos_x;
		c_y = buffer->crop_pos_y;
	}
	if ((buffer->buf_id & BUF_ID_FLG_CROPSIZE) == BUF_ID_FLG_CROPSIZE) {
		c_w = buffer->crop_size_x;
		c_h = buffer->crop_size_y;
	}

	printk_dbg2(2, "image size:%d x %d, crop (%d, %d)-(%d, %d)\n", \
		w, h, c_x, c_y, c_w, c_h);

	grap->work_rect_x[idx] = c_x;
	grap->work_rect_y[idx] = c_y;
	grap->work_rect_w[idx] = c_w;
	grap->work_rect_h[idx] = c_h;

err_exit:
	return rc;
}

int update_work_appshare(
	struct composer_info   *info, struct composer_buffer *buffer)
{
	static const int pixfmt_numplane_translate[] = {
		[CMPPIXFMT_ARGB8888] = 1,
		[CMPPIXFMT_RGB888] = 1,
		[CMPPIXFMT_RGB565] = 1,
		[CMPPIXFMT_YUV420SP] = 2,
		[CMPPIXFMT_YUV422SP] = 2,
		[CMPPIXFMT_YUV420PL] = 3
	};
	static const int fh_status_translate_app[] = {
		[0] = INFO_ID_SET_APP0,
		[1] = INFO_ID_SET_APP1,
		[2] = INFO_ID_SET_APP2
	};
	int  i;
	int  num_plane;
	unsigned long flags;
	int  rc = CMP_ERR_OK;


	if ((buffer->buf_id & BUF_ID_SET_BUFFER) == 0) {
		printk_dbg2(2, "CMP_IOCS_LAYADDR not specified.\n");
		rc = CMP_ERR_NG;
		goto err_exit;
	}

	if ((buffer->buf_id & BUF_ID_FLG_PIXFMT) == 0) {
		printk_dbg2(2, "CMP_IOCS_PIXFMT not specified.\n");
		rc = CMP_ERR_NG;
		goto err_exit;
	}

	printk_dbg2(2, "spinlock");
	spin_lock_irqsave(&buffer->buf_lock, flags);

	num_plane = pixfmt_numplane_translate[buffer->data_fmt];
	BUG_ON(num_plane < 1 || num_plane > 3);

	{
		void *app[3] = {NULL, NULL, NULL};
		int  appid[3], appoffset[3], appopen_flag;

/***************************/
/* pickup-informatin       */
/***************************/
		/* copy previous created APP-Share */
		for (i = 0; i < 3; i++) {
			if (info->id & fh_status_translate_app[i]) {
				info->id  &= ~fh_status_translate_app[i];
				app[i] = info->appinfo[i];
				info->appinfo[i] = NULL;
			}
		}

		appopen_flag = 0;
		if (buffer->addr == NULL && num_plane > 0) {
			appid[0]     = buffer->app_id;
			appoffset[0] = buffer->offset;
			appopen_flag |= 0x01;
		}
		if (buffer->addr_c0 == NULL && num_plane > 1) {
			appid[1]     = buffer->app_id_c0;
			appoffset[1] = buffer->offset_c0;
			appopen_flag |= 0x02;
		}
		if (buffer->addr_c1 == NULL && num_plane > 2) {
			appid[2]     = buffer->app_id_c1;
			appoffset[2] = buffer->offset_c1;
			appopen_flag |= 0x04;
		}

		spin_unlock_irqrestore(&buffer->buf_lock, flags);

/***************************/
/* app-share close.        */
/***************************/
		for (i = 0; i < 3; i++) {
			int app_rc = 0;
			if (app[i]) {
				sh_mobile_appmem_free(app[i]);
				/* ignore error, because recovery from error \
				   is only available by reboot-kernel.       */
			}
			if (app_rc) {
				printk_err("sh_mobile_appmem_free"
					" return by error\n");
			}
			app[i] = NULL;
		}

		info->id &= ~(INFO_ID_SET_APP0 | \
			INFO_ID_SET_APP1 | INFO_ID_SET_APP2);
/***************************/
/* app-share open.        */
/***************************/
		for (i = 0; i < 3; i++) {
			if ((appopen_flag & (1<<i)) == 0) {
				/* to next buffer */
				continue;
			}
			app[i] = sh_mobile_appmem_share(appid[i], DEV_NAME);
			if (app[i] == NULL) {
				printk_err1("sh_mobile_appmem_share"
					" return by error\n");
				rc = CMP_ERR_NG;
			}
		}

/***************************/
/* last process            */
/***************************/
		if (rc != CMP_ERR_OK) {
			/* app share close by error */
			for (i = 0; i < 3; i++) {
				int app_rc = 0;
				if (app[i]) {
					app_rc = sh_mobile_appmem_free(app[i]);
					/* recovery from error is only \
					   available by reboot-kernel. */
				}
				if (app_rc) {
					printk_err("sh_mobile_appmem_free"
						" return by error\n");
				}
				app[i] = NULL;
			}
		} else {
			/* record open result */
			printk_dbg2(2, "spinlock");
			spin_lock_irqsave(&buffer->buf_lock, flags);
			for (i = 0; i < 3; i++) {
				info->id  |= fh_status_translate_app[i];
				info->appinfo[i] = app[i];
			}
			spin_unlock_irqrestore(&buffer->buf_lock, flags);
		}
	}

err_exit:
	return rc;
}

static int update_grap_image_param(
	screen_grap_image_param * param,
	struct composer_info    *info,
	struct composer_buffer  *buf,
	struct composer_grapdata *grap,
	int    idx)
{
	static const int pixfmt_numplane_translate[] = {
		[CMPPIXFMT_ARGB8888] = 1,
		[CMPPIXFMT_RGB888] = 1,
		[CMPPIXFMT_RGB565] = 1,
		[CMPPIXFMT_YUV420SP] = 2,
		[CMPPIXFMT_YUV422SP] = 2,
		[CMPPIXFMT_YUV420PL] = 3
	};
	static const int rt_graphics_color_transleta[] = {
		[CMPPIXFMT_ARGB8888] = RT_GRAPHICS_COLOR_ARGB8888,
		[CMPPIXFMT_RGB888]   = RT_GRAPHICS_COLOR_RGB888,
		[CMPPIXFMT_RGB565]   = RT_GRAPHICS_COLOR_RGB565,
		[CMPPIXFMT_YUV420SP] = RT_GRAPHICS_COLOR_YUV420SP,
		[CMPPIXFMT_YUV422SP] = RT_GRAPHICS_COLOR_YUV422SP,
#ifdef RT_GRAPHICS_COLOR_YUV420PL
		[CMPPIXFMT_YUV420PL] = RT_GRAPHICS_COLOR_YUV420PL,
#endif
	};
	static const int rt_yuvformat_translate[][2] = {
		[CMPYUVCOLOR_BT601_COMPRESS][0] = RT_GRAPHICS_COLOR_BT601,
		[CMPYUVCOLOR_BT601_COMPRESS][1] = RT_GRAPHICS_COLOR_COMPRESSED,
		[CMPYUVCOLOR_BT601_FULLSCALE][0] = RT_GRAPHICS_COLOR_BT601,
		[CMPYUVCOLOR_BT601_FULLSCALE][1] = RT_GRAPHICS_COLOR_FULLSCALE,
		[CMPYUVCOLOR_BT709_COMPRESS][0] = RT_GRAPHICS_COLOR_BT709,
		[CMPYUVCOLOR_BT709_COMPRESS][1] = RT_GRAPHICS_COLOR_COMPRESSED,
		[CMPYUVCOLOR_BT709_FULLSCALE][0] = RT_GRAPHICS_COLOR_BT709,
		[CMPYUVCOLOR_BT709_FULLSCALE][1] = RT_GRAPHICS_COLOR_FULLSCALE,
	};
	int  rc, num_plane;
	int  offset, offset_c;

	num_plane = pixfmt_numplane_translate[buf->data_fmt];
	rc = CMP_ERR_OK;

	param->width     = grap->work_rect_w[idx];
	param->height    = grap->work_rect_h[idx];
	param->stride    = grap->work_linebyte[idx];
	param->stride_c  = grap->work_linebyte_c0[idx];
	offset  = param->stride                  * grap->work_rect_y[idx];
	offset += grap->work_pixelsize[idx]      * grap->work_rect_x[idx];
	offset_c  = param->stride_c              * grap->work_rect_y[idx];
	offset_c += grap->work_pixelsize_c0[idx] * grap->work_rect_x[idx];

	printk_dbg2(2, "offset = 0x%x (stride:%d, y:%d, x:%d)\n", offset, \
		param->stride, grap->work_rect_y[idx], grap->work_rect_x[idx]);
	printk_dbg2(2, "offset_c = 0x%x (stride:%d, y:%d, x:%d)\n", offset_c, \
		param->stride_c,
		grap->work_rect_y[idx], grap->work_rect_x[idx]);

	param->format    = rt_graphics_color_transleta[buf->data_fmt];
	switch (buf->data_fmt) {
	case CMPPIXFMT_YUV420SP:
	case CMPPIXFMT_YUV422SP:
	case CMPPIXFMT_YUV420PL:
		param->yuv_format = rt_yuvformat_translate[buf->col_space][0];
		param->yuv_range  = rt_yuvformat_translate[buf->col_space][1];
		break;
	default:
		param->yuv_format = RT_GRAPHICS_COLOR_BT601;
		param->yuv_range  = RT_GRAPHICS_COLOR_COMPRESSED;
		break;
	}

	/* set plane 1 */
	if (buf->addr) {
		param->address      = buf->addr;
		param->apmem_handle = NULL;
	} else if (info->id & INFO_ID_SET_APP0) {
		if (info->appinfo[0] == NULL) {
			printk_err2("app share 0 is not opoen\n");
			rc = CMP_ERR_NG;
			goto err_exit;
		}
		param->address      = sh_mobile_appmem_getaddress(
			info->appinfo[0], buf->offset);
		param->apmem_handle = sh_mobile_appmem_getmemoryhandle(
			info->appinfo[0]);
		if (param->address == NULL) {
			printk_err2("can not get param->address"
				" of appinfo[0]\n");
			rc = CMP_ERR_NG;
			goto err_exit;
		}
		if (param->apmem_handle == NULL) {
			printk_err2("can not get param->apmem_handle"
				" of appinfo[0]\n");
			rc = CMP_ERR_NG;
			goto err_exit;
		}
	} else {
		param->address      = NULL;
		param->apmem_handle = NULL;
		rc = CMP_ERR_NG;
		printk_err2("can not set Y/RGB plane address\n");
	}
	printk_dbg2(2, "address adjust to %p to %p\n", param->address, \
		param->address + offset);
	/* adjsut crop position */
	param->address += offset;

	param->address_c0      = NULL;
	param->apmem_handle_c0 = NULL;
	param->address_c1      = NULL;
	param->apmem_handle_c1 = NULL;
	if (num_plane <= 1) {
		/* return */
		goto finish;
	}

	/* set plane 2 */
	if (buf->addr_c0) {
		param->address_c0      = buf->addr_c0;
		param->apmem_handle_c0 = NULL;
	} else if (info->id & INFO_ID_SET_APP1) {
		if (info->appinfo[1] == NULL) {
			printk_err2("app share 1 is not opoen\n");
			rc = CMP_ERR_NG;
			goto err_exit;
		}
		param->address_c0      = sh_mobile_appmem_getaddress(
			info->appinfo[1], buf->offset);
		param->apmem_handle_c0 = sh_mobile_appmem_getmemoryhandle(
			info->appinfo[1]);
		if (param->address_c0 == NULL) {
			printk_err2("can not get param->address"
				" of appinfo[1]\n");
			rc = CMP_ERR_NG;
			goto err_exit;
		}
		if (param->apmem_handle_c0 == NULL) {
			printk_err2("can not get param->apmem_handle"
				" of appinfo[1]\n");
			rc = CMP_ERR_NG;
			goto err_exit;
		}
	} else {
		/* set error. already initialize parameteres. */
		rc = CMP_ERR_NG;
		printk_err2("can not set UV/U plane address\n");
	}
	printk_dbg2(2, "address_c0 adjust to %p to %p\n", param->address_c0, \
		param->address_c0 + offset_c);
	/* adjsut crop position */
	param->address_c0 += offset_c;

	if (num_plane <= 2) {
		/* return */
		goto finish;
	}

	/* set plane 3 */
	if (buf->addr_c1) {
		param->address_c1      = buf->addr_c1;
		param->apmem_handle_c1 = NULL;
	} else if (info->id & INFO_ID_SET_APP2) {
		if (info->appinfo[2] == NULL) {
			printk_err2("app share 2 is not opoen\n");
			rc = CMP_ERR_NG;
			goto err_exit;
		}
		param->address_c1      = sh_mobile_appmem_getaddress(
			info->appinfo[2], buf->offset);
		param->apmem_handle_c1 = sh_mobile_appmem_getmemoryhandle(
			info->appinfo[2]);
		if (param->address_c1 == NULL) {
			printk_err2("can not get param->address"
				" of appinfo[2]\n");
			rc = CMP_ERR_NG;
			goto err_exit;
		}
		if (param->apmem_handle_c1 == NULL) {
			printk_err2("can not get param->apmem_handle"
				" of appinfo[2]\n");
			rc = CMP_ERR_NG;
			goto err_exit;
		}
	} else {
		/* set error. already initialize parameteres. */
		rc = CMP_ERR_NG;
		printk_err2("can not set V plane address\n");
	}
	printk_dbg2(2, "address_c1 adjust to %p to %p\n", param->address_c1, \
		param->address_c1 + offset_c);
	/* adjsut crop position */
	param->address_c1 += offset_c;

finish:
	/* pass */
	rc = CMP_ERR_OK;
err_exit:
	return rc;
};


static int update_grap_arguments(struct composer_fh *fh)
{
	struct composer_grapdata *grap   = &fh->grap_data;
	int                      nBuffer = 0;
	struct composer_info     *info[5];
	struct composer_buffer   *buf[5];
	int                      bufidx[5];

	int                      i;
	int                      rc;

	static const int fh_status_translate[] = {
		[IDX_OUTPLANE]  = FH_STATUS_OUTPLANE,
		[IDX_MAINLAYER] = FH_STATUS_MAINLAYER,
		[IDX_CH1LAYER]  = FH_STATUS_CH1LAYER,
		[IDX_CH2LAYER]  = FH_STATUS_CH2LAYER,
		[IDX_CH3LAYER]  = FH_STATUS_CH3LAYER
	};

	/* pick layer information. */
	for (i = 0; i < ARRAY_SIZE(fh_status_translate); i++) {
		if ((fh->fh_status & fh_status_translate[i]) == 0) {
			/* not open */
			continue;
		}

		BUG_ON(nBuffer >= 5);

		info[nBuffer] = &fh->fh_info[i];
		buf[nBuffer]  = fh->fh_info[i].buffer;
		bufidx[nBuffer] = i;

		/* check buffer configuration */
		if (fh->fh_info[i].id & INFO_ID_FLG_NOTDISP) {
			/* not blending plane */
			continue;
		}
		nBuffer++;
	}

	/* this error should detect in chk_ioc_start() */
	BUG_ON(nBuffer < 2);

/* update for screen_grap_new */
	grap->_new.notify_graphics_image_conv  = notify_graphics_image_conv;
	grap->_new.notify_graphics_image_blend = notify_graphics_image_blend;
/* update for screen_grap_initialize */
	grap->_ini.mode = RT_GRAPHICS_MODE_IMAGE_BLEND;
/* update for screen_grap_quit */
	grap->_quit.mode = RT_GRAPHICS_MODE_IMAGE_BLEND;
/* update for screen_grap_image_blend */
	grap->_blend.user_data = (unsigned long) grap;

	/* create destination parameter */
	BUG_ON(bufidx[0] != IDX_OUTPLANE);
	i = 0;
	rc = update_grap_image_param(&grap->_blend.output_image,
		info[i], buf[i], grap, bufidx[i]);
	if (rc != CMP_ERR_OK) {
		printk_err2("parameter convert error in CMP_OUT_PLANE\n");
		goto err_exit;
	}

	grap->_blend.background_color = buf[i]->backcolor;

	/* create source parameter */
	grap->_blend.input_layer[0] = NULL;
	grap->_blend.input_layer[1] = NULL;
	grap->_blend.input_layer[2] = NULL;
	grap->_blend.input_layer[3] = NULL;
	for (i = 1; i < nBuffer; i++) {
		int layer_no = i-1;

		rc = update_grap_image_param(&grap->_in[layer_no].image,
			info[i], buf[i], grap, bufidx[i]);
		if (rc != CMP_ERR_OK) {
			printk_err2("parameter convert error"
				" in SRC image [%d]\n", bufidx[i]);
			goto err_exit;
		}

		grap->_blend.input_layer[layer_no] = &grap->_in[layer_no];
		grap->_in[layer_no].rect.x      = buf[i]->pos_x;
		grap->_in[layer_no].rect.y      = buf[i]->pos_y;
		grap->_in[layer_no].rect.width  = buf[i]->size_x;
		grap->_in[layer_no].rect.height = buf[i]->size_y;
		if (buf[i]->data_x == grap->_in[layer_no].rect.width &&
		   buf[i]->data_y == grap->_in[layer_no].rect.height) {
			/* no need resize */
			grap->_in[layer_no].rect.width  = 0;
			grap->_in[layer_no].rect.height = 0;
		}
		grap->_in[layer_no].alpha       = buf[i]->alpha;

		grap->_in[layer_no].rotate      = RT_GRAPHICS_ROTATE_0;
		grap->_in[layer_no].mirror      = RT_GRAPHICS_MIRROR_N;
		if (buf[i]->keycolor == CMPKEYCOLOR_OFF) {
			/* no need key color */
			grap->_in[layer_no].key_color = \
				RT_GRAPHICS_KEY_COLOR_OFF;
		} else {
			/* set key color */
			grap->_in[layer_no].key_color = buf[i]->keycolor;
		}
	}

	/* pass */
	rc = CMP_ERR_OK;

err_exit:
	return rc;
}

/******************************************
* ioctl (CMP_IOC_ISLAYEX)
******************************************/
static int  ioc_islayex(struct composer_fh *fh)
{
	/* always return 0 */
	return 0;
}

/******************************************
* ioctl (CMP_IOCG_SUPPORTPIXFMT)
******************************************/
static int  iocg_suppotpixfmt(struct composer_fh *fh, \
	struct cmp_lay_supportpixfmt *arg)
{
	int rc = -EINVAL;

	printk_dbg2(2, "arg level:%d fmt:%d\n", arg->level, arg->pixfmt);

	/* check argument */
	rc = chk_ioc_supportpixfmt(fh, arg);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

err_exit:
	return rc;
}

/******************************************
* ioctl (CMP_IOCS_OPEN)
******************************************/
static int  iocs_open(struct composer_fh *fh, \
	struct cmp_lay_open *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;

	printk_dbg2(2, "arg level:%d id:%d\n", arg->level, arg->bufferid);

	/* check argument */
	rc = chk_ioc_open(fh, arg);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* check already opened.      */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	info   = &fh->fh_info[idx];
	if (info->buffer != NULL) {
		printk_err2("level 0x%x already opened.\n", arg->level);
		rc = -EBUSY;
		goto err_exit;
	}

	/* bind buffer */
	rc = bind_buffer(fh, idx, arg->bufferid);
	if (0 == rc) {
		static const int fh_status_translate[] = {
			[CMP_OUT_PLANE]  = FH_STATUS_OUTPLANE,
			[CMP_MAIN_LAYER] = FH_STATUS_MAINLAYER,
			[CMP_CH1_LAYER]  = FH_STATUS_CH1LAYER,
			[CMP_CH2_LAYER]  = FH_STATUS_CH2LAYER,
			[CMP_CH3_LAYER]  = FH_STATUS_CH3LAYER };

		fh->fh_status |= fh_status_translate[arg->level];
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCS_CLOSE)
******************************************/
static int  iocs_close(struct composer_fh *fh, \
	struct cmp_lay_close *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;

	printk_dbg2(2, "arg level:%d\n", arg->level);

	/* check argument */
	rc = chk_ioc_close(fh, arg);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already closeed.      */
	info   = &fh->fh_info[idx];
	if (info->buffer == NULL) {
		printk_err2("level 0x%x already closed.\n", arg->level);
		rc = 0;  /* no error */
		goto err_exit;
	}

	/* release buffer */
	rc = lock_buffer(fh, idx);
	if (rc < 0) {
		printk_err2("lock_buffer interrupted.\n");
		rc = -EAGAIN;
		goto err_exit;
	}

	rc = release_buffer(fh, arg->level);
	/* unlock_buffer executed in release_buffer */

	if (0 == rc) {
		static const int fh_status_translate[] = {
			[CMP_OUT_PLANE]  = FH_STATUS_OUTPLANE,
			[CMP_MAIN_LAYER] = FH_STATUS_MAINLAYER,
			[CMP_CH1_LAYER]  = FH_STATUS_CH1LAYER,
			[CMP_CH2_LAYER]  = FH_STATUS_CH2LAYER,
			[CMP_CH3_LAYER]  = FH_STATUS_CH3LAYER };

		fh->fh_status &= ~fh_status_translate[idx];
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCS_DISP)
******************************************/
static int  iocs_disp(struct composer_fh *fh, \
	struct cmp_lay_disp *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;

	printk_dbg2(2, "arg level:%d on:%d\n", arg->level, arg->on);

	/* check argument */
	rc = chk_ioc_disp(fh, arg);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);

	/* lock buffer */
	rc = lock_buffer(fh, idx);
	if (rc < 0) {
		printk_err2("lock_buffer interrupted.\n");
		rc = -EAGAIN;
		goto err_exit;
	}
	if (0 == rc) {
		if (arg->on) {
			/* clera not blend flag. */
			info->id &= ~INFO_ID_FLG_NOTDISP;
		} else {
			/* set not blend flag. */
			info->id |=  INFO_ID_FLG_NOTDISP;
		}
		rc = unlock_buffer(fh, idx);
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCS_ALPHA)
******************************************/
static int  iocs_alpha(struct composer_fh *fh, \
	struct cmp_lay_alpha *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;
	struct composer_buffer *buffer;

	printk_dbg2(2, "arg level:%d alpha:%d\n", arg->level, arg->alpha);

	/* check argument */
	rc = chk_ioc_alpha(fh, arg, 1);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);
	buffer = info->buffer;

	/* lock buffer */
	rc = lock_buffer(fh, idx);
	if (rc < 0) {
		printk_err2("lock_buffer interrupted.\n");
		rc = -EAGAIN;
		goto err_exit;
	}
	if (0 == rc) {
		unsigned long flags;

		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buffer->buf_lock, flags);

		buffer->alpha = arg->alpha;

		spin_unlock_irqrestore(&buffer->buf_lock, flags);

		rc = unlock_buffer(fh, idx);
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCG_ALPHA)
******************************************/
static int  iocg_alpha(struct composer_fh *fh, \
	struct cmp_lay_alpha *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;
	struct composer_buffer *buffer;

	printk_dbg2(2, "arg level:%d alpha:-\n", arg->level);

	/* check argument */
	rc = chk_ioc_alpha(fh, arg, 0);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);
	buffer = info->buffer;

	/* not lock buffer, because does not modify it. */
	{
		unsigned long flags;

		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buffer->buf_lock, flags);

		arg->alpha = buffer->alpha;

		spin_unlock_irqrestore(&buffer->buf_lock, flags);
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCS_KEYCOLOR)
******************************************/
static int  iocs_keycolor(struct composer_fh *fh, \
	struct cmp_lay_keycolor *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;
	struct composer_buffer *buffer;

	printk_dbg2(2, "arg level:%d keycolor:0x%x\n",
		arg->level, arg->keycolor);

	/* check argument */
	rc = chk_ioc_keycolor(fh, arg, 1);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);
	buffer = info->buffer;

	/* lock buffer */
	rc = lock_buffer(fh, idx);
	if (rc < 0) {
		printk_err2("lock_buffer interrupted.\n");
		rc = -EAGAIN;
		goto err_exit;
	}
	if (0 == rc) {
		unsigned long flags;

		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buffer->buf_lock, flags);

		buffer->keycolor = arg->keycolor;

		spin_unlock_irqrestore(&buffer->buf_lock, flags);

		rc = unlock_buffer(fh, idx);
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCG_KEYCOLOR)
******************************************/
static int  iocg_keycolor(struct composer_fh *fh, \
	struct cmp_lay_keycolor *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;
	struct composer_buffer *buffer;

	printk_dbg2(2, "arg level:%d keycolor:-\n", arg->level);

	/* check argument */
	rc = chk_ioc_keycolor(fh, arg, 0);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);
	buffer = info->buffer;

	/* not lock buffer, because does not modify it. */
	{
		unsigned long flags;

		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buffer->buf_lock, flags);

		arg->keycolor = buffer->keycolor;

		spin_unlock_irqrestore(&buffer->buf_lock, flags);
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCS_LAYSIZE)
******************************************/
static int  iocs_laysize(struct composer_fh *fh, \
	struct cmp_lay_data_size *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;
	struct composer_buffer *buffer;

	printk_dbg2(2, "arg level:%d x:%d y:%d\n", arg->level, arg->x, arg->y);

	/* check argument */
	rc = chk_ioc_laysize(fh, arg, 1);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);
	buffer = info->buffer;

	/* lock buffer */
	rc = lock_buffer(fh, idx);
	if (rc < 0) {
		printk_err2("lock_buffer interrupted.\n");
		rc = -EAGAIN;
		goto err_exit;
	}
	if (0 == rc) {
		unsigned long flags;

		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buffer->buf_lock, flags);

		buffer->data_x = arg->x;
		buffer->data_y = arg->y;

		buffer->buf_id |= BUF_ID_FLG_LAYSIZE;

		/* determine stride */
		update_work_linebytes(&fh->grap_data, idx, buffer);
		/* determine crop   */
		update_work_imagesize(&fh->grap_data, idx, buffer);

		spin_unlock_irqrestore(&buffer->buf_lock, flags);

		rc = unlock_buffer(fh, idx);

		wakeup_allfile(buffer);
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCG_LAYSIZE)
******************************************/
static int  iocg_laysize(struct composer_fh *fh, \
	struct cmp_lay_data_size *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;
	struct composer_buffer *buffer;

	printk_dbg2(2, "arg level:%d x:- y:-\n", arg->level);

	/* check argument */
	rc = chk_ioc_laysize(fh, arg, 0);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);
	buffer = info->buffer;

	/* not lock buffer, because does not modify it. */
	{
		unsigned long flags;

		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buffer->buf_lock, flags);

		arg->x = buffer->data_x;
		arg->y = buffer->data_y;

		spin_unlock_irqrestore(&buffer->buf_lock, flags);
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCS_PIXFMT)
******************************************/
static int  iocs_pixfmt(struct composer_fh *fh, \
	struct cmp_lay_pixfmt *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;
	struct composer_buffer *buffer;

	printk_dbg2(2, "arg level:%d fmt:%d color:%d\n",
		arg->level, arg->pixfmt, arg->colorspace);

	/* check argument */
	rc = chk_ioc_pixfmt(fh, arg, 1);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);
	buffer = info->buffer;

	/* lock buffer */
	rc = lock_buffer(fh, idx);
	if (rc < 0) {
		printk_err2("lock_buffer interrupted.\n");
		rc = -EAGAIN;
		goto err_exit;
	}
	if (0 == rc) {
		unsigned long flags;

		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buffer->buf_lock, flags);
		buffer->data_fmt  = arg->pixfmt;
		buffer->col_space = arg->colorspace;
		buffer->buf_id |= BUF_ID_FLG_PIXFMT;

		/* determine stride */
		update_work_linebytes(&fh->grap_data, idx, buffer);

		spin_unlock_irqrestore(&buffer->buf_lock, flags);

		rc = unlock_buffer(fh, idx);
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCG_PIXFMT)
******************************************/
static int  iocg_pixfmt(struct composer_fh *fh, \
	struct cmp_lay_pixfmt *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;
	struct composer_buffer *buffer;

	printk_dbg2(2, "arg level:%d fmt:- col:-\n", arg->level);

	/* check argument */
	rc = chk_ioc_pixfmt(fh, arg, 0);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);
	buffer = info->buffer;

	/* not lock buffer, because does not modify it. */
	{
		unsigned long flags;

		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buffer->buf_lock, flags);

		arg->pixfmt     = buffer->data_fmt ;
		arg->colorspace = buffer->col_space;

		spin_unlock_irqrestore(&buffer->buf_lock, flags);
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCS_CROPSIZE)
******************************************/
static int  iocs_cropsize(struct composer_fh *fh, \
	struct cmp_lay_cropsize *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;
	struct composer_buffer *buffer;

	printk_dbg2(2, "arg level:%d x:%d y:%d\n", arg->level, arg->x, arg->y);

	/* check argument */
	rc = chk_ioc_cropsize(fh, arg, 1, NULL);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);
	buffer = info->buffer;

	/* lock buffer */
	rc = lock_buffer(fh, idx);
	if (rc < 0) {
		printk_err2("lock_buffer interrupted.\n");
		rc = -EAGAIN;
		goto err_exit;
	}
	if (0 == rc) {
		unsigned long flags;

		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buffer->buf_lock, flags);
		buffer->crop_size_x = arg->x;
		buffer->crop_size_y = arg->y;
		buffer->buf_id |= BUF_ID_FLG_CROPSIZE;

		/* determine crop   */
		update_work_imagesize(&fh->grap_data, idx, buffer);

		spin_unlock_irqrestore(&buffer->buf_lock, flags);

		rc = unlock_buffer(fh, idx);
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCG_CROPSIZE)
******************************************/
static int  iocg_cropsize(struct composer_fh *fh, \
	struct cmp_lay_cropsize *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;
	struct composer_buffer *buffer;

	printk_dbg2(2, "arg level:%d x:- y:-\n", arg->level);

	/* check argument */
	rc = chk_ioc_cropsize(fh, arg, 0, NULL);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);
	buffer = info->buffer;

	/* not lock buffer, because does not modify it. */
	{
		unsigned long flags;

		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buffer->buf_lock, flags);

		arg->x = buffer->crop_size_x;
		arg->y = buffer->crop_size_y;

		spin_unlock_irqrestore(&buffer->buf_lock, flags);
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCG_CROPPOS)
******************************************/
static int  iocs_croppos(struct composer_fh *fh, \
	struct cmp_lay_croppos *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;
	struct composer_buffer *buffer;

	printk_dbg2(2, "arg level:%d x:%d y:%d\n", arg->level, arg->x, arg->y);

	/* check argument */
	rc = chk_ioc_croppos(fh, arg, 1, NULL);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);
	buffer = info->buffer;

	/* lock buffer */
	rc = lock_buffer(fh, idx);
	if (rc < 0) {
		printk_err2("lock_buffer interrupted.\n");
		rc = -EAGAIN;
		goto err_exit;
	}
	if (0 == rc) {
		unsigned long flags;

		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buffer->buf_lock, flags);
		buffer->crop_pos_x = arg->x;
		buffer->crop_pos_y = arg->y;
		buffer->buf_id |= BUF_ID_FLG_CROPPOS;

		/* determine crop   */
		update_work_imagesize(&fh->grap_data, idx, buffer);

		spin_unlock_irqrestore(&buffer->buf_lock, flags);

		rc = unlock_buffer(fh, idx);
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCG_CROPPOS)
******************************************/
static int  iocg_croppos(struct composer_fh *fh, \
	struct cmp_lay_croppos *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;
	struct composer_buffer *buffer;

	printk_dbg2(2, "arg level:%d x:- y:-\n", arg->level);

	/* check argument */
	rc = chk_ioc_croppos(fh, arg, 0, NULL);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);
	buffer = info->buffer;

	/* not lock buffer, because does not modify it. */
	{
		unsigned long flags;

		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buffer->buf_lock, flags);

		arg->x = buffer->crop_pos_x;
		arg->y = buffer->crop_pos_y;

		spin_unlock_irqrestore(&buffer->buf_lock, flags);
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCS_POS)
******************************************/
static int  iocs_pos(struct composer_fh *fh, \
	struct cmp_lay_pos *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;
	struct composer_buffer *buffer;

	printk_dbg2(2, "arg level:%d x:%d y:%d\n", arg->level, arg->x, arg->y);

	/* check argument */
	rc = chk_ioc_pos(fh, arg, 1);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);
	buffer = info->buffer;

	/* lock buffer */
	rc = lock_buffer(fh, idx);
	if (rc < 0) {
		printk_err2("lock_buffer interrupted.\n");
		rc = -EAGAIN;
		goto err_exit;
	}
	if (0 == rc) {
		unsigned long flags;

		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buffer->buf_lock, flags);
		buffer->pos_x = arg->x;
		buffer->pos_y = arg->y;
		buffer->buf_id |= BUF_ID_FLG_POS;
		spin_unlock_irqrestore(&buffer->buf_lock, flags);

		rc = unlock_buffer(fh, idx);
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCG_POS)
******************************************/
static int  iocg_pos(struct composer_fh *fh, \
	struct cmp_lay_pos *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;
	struct composer_buffer *buffer;

	printk_dbg2(2, "arg level:%d x:- y:-\n", arg->level);

	/* check argument */
	rc = chk_ioc_pos(fh, arg, 0);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);
	buffer = info->buffer;

	/* not lock buffer, because does not modify it. */
	{
		unsigned long flags;

		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buffer->buf_lock, flags);

		arg->x = buffer->pos_x;
		arg->y = buffer->pos_y;

		spin_unlock_irqrestore(&buffer->buf_lock, flags);
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCS_SIZE)
******************************************/
static int  iocs_size(struct composer_fh *fh, \
	struct cmp_lay_compose_size *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;
	struct composer_buffer *buffer;

	printk_dbg2(2, "arg level:%d x:%d y:%d\n", arg->level, arg->x, arg->y);

	/* check argument */
	rc = chk_ioc_size(fh, arg, 1);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);
	buffer = info->buffer;

	/* lock buffer */
	rc = lock_buffer(fh, idx);
	if (rc < 0) {
		printk_err2("lock_buffer interrupted.\n");
		rc = -EAGAIN;
		goto err_exit;
	}
	if (0 == rc) {
		unsigned long flags;

		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buffer->buf_lock, flags);
		buffer->size_x = arg->x;
		buffer->size_y = arg->y;
		buffer->buf_id |= BUF_ID_FLG_SIZE;
		spin_unlock_irqrestore(&buffer->buf_lock, flags);

		rc = unlock_buffer(fh, idx);
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCG_SIZE)
******************************************/
static int  iocg_size(struct composer_fh *fh, \
	struct cmp_lay_compose_size *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;
	struct composer_buffer *buffer;

	printk_dbg2(2, "arg level:%d x:- y:-\n", arg->level);

	/* check argument */
	rc = chk_ioc_size(fh, arg, 0);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);
	buffer = info->buffer;

	/* not lock buffer, because does not modify it. */
	{
		unsigned long flags;

		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buffer->buf_lock, flags);

		arg->x = buffer->size_x;
		arg->y = buffer->size_y;

		spin_unlock_irqrestore(&buffer->buf_lock, flags);
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCS_VIEWLAY)
******************************************/
static int  iocs_viewlay(struct composer_fh *fh, \
	struct cmp_viewlay *arg)
{
	int rc, idx;
	struct composer_info   *info;
	struct composer_buffer *buffer;

	printk_dbg2(2, "arg level:%d data_x:%d data_y:%d " \
		"croppos_x:%d croppos_y:%d " \
		"cropsize_x:%d cropsize_y:%d " \
		"pos_x:%d pos_y:%d " \
		"compose_x:%d compose_y:%d\n" \
		, arg->level, arg->data_x, arg->data_y
		, arg->croppos_x, arg->croppos_y
		, arg->cropsize_x, arg->cropsize_y
		, arg->pos_x, arg->pos_y, arg->compose_x, arg->compose_y);

	/* check argument */
	rc = chk_ioc_viewlay(fh, arg, -1);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);
	buffer = info->buffer;

	/* check argument and update it. if necessary. */
	rc = chk_ioc_viewlay(fh, arg, idx);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* lock buffer */
	rc = lock_buffer(fh, idx);
	if (rc < 0) {
		printk_err2("lock_buffer interrupted.\n");
		rc = -EAGAIN;
		goto err_exit;
	}
	if (0 == rc) {
		unsigned long flags;
		unsigned long mask = 0x80000000;
		int           dst_not_support = 0;

		if (arg->level == CMP_OUT_PLANE ||
		   arg->level == CMP_MAIN_LAYER) {
			/* destination parameter not available */
			dst_not_support = 1;
		}

		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buffer->buf_lock, flags);

		if ((arg->data_x & mask) == 0) {
			buffer->data_x = arg->data_x;
			buffer->buf_id |= BUF_ID_FLG_LAYSIZE1;
		}
		if ((arg->data_y & mask) == 0) {
			buffer->data_y = arg->data_y;
			buffer->buf_id |= BUF_ID_FLG_LAYSIZE2;
		}
		if ((arg->cropsize_x & mask) == 0) {
			buffer->crop_size_x = arg->cropsize_x;
			buffer->buf_id |= BUF_ID_FLG_CROPSIZE1;
		}
		if ((arg->cropsize_y & mask) == 0) {
			buffer->crop_size_y = arg->cropsize_y;
			buffer->buf_id |= BUF_ID_FLG_CROPSIZE2;
		}
		if ((arg->croppos_x & mask) == 0) {
			buffer->crop_pos_x = arg->croppos_x;
			buffer->buf_id |= BUF_ID_FLG_CROPPOS1;
		}
		if ((arg->croppos_y & mask) == 0) {
			buffer->crop_pos_y = arg->croppos_y;
			buffer->buf_id |= BUF_ID_FLG_CROPPOS2;
		}
		if (dst_not_support == 0) {
			if ((arg->pos_x & mask) == 0) {
				buffer->pos_x = arg->pos_x;
				buffer->buf_id |= BUF_ID_FLG_POS1;
			}
			if ((arg->pos_y & mask) == 0) {
				buffer->pos_y = arg->pos_y;
				buffer->buf_id |= BUF_ID_FLG_POS2;
			}
			if ((arg->compose_x & mask) == 0) {
				buffer->size_x = arg->compose_x;
				buffer->buf_id |= BUF_ID_FLG_SIZE1;
			}
			if ((arg->compose_y & mask) == 0) {
				buffer->size_y = arg->compose_y;
				buffer->buf_id |= BUF_ID_FLG_SIZE2;
			}
		}

		/* determine stride */
		update_work_linebytes(&fh->grap_data, idx, buffer);
		/* determine crop   */
		update_work_imagesize(&fh->grap_data, idx, buffer);

		spin_unlock_irqrestore(&buffer->buf_lock, flags);

		rc = unlock_buffer(fh, idx);
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCS_BACKCOLOR)
******************************************/
static int  iocs_backcolor(struct composer_fh *fh, \
	struct cmp_lay_backcolor *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;
	struct composer_buffer *buffer;

	printk_dbg2(2, "arg level:%d col:0x%x\n", arg->level, arg->backcolor);

	/* check argument */
	rc = chk_ioc_backcolor(fh, arg, 1);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);
	buffer = info->buffer;

	/* lock buffer */
	rc = lock_buffer(fh, idx);
	if (rc < 0) {
		printk_err2("lock_buffer interrupted.\n");
		rc = -EAGAIN;
		goto err_exit;
	}
	if (0 == rc) {
		unsigned long flags;

		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buffer->buf_lock, flags);
		buffer->backcolor = arg->backcolor;
		spin_unlock_irqrestore(&buffer->buf_lock, flags);

		rc = unlock_buffer(fh, idx);
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCG_BACKCOLOR)
******************************************/
static int  iocg_backcolor(struct composer_fh *fh, \
	struct cmp_lay_backcolor *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;
	struct composer_buffer *buffer;

	printk_dbg2(2, "arg level:%d col:-\n", arg->level);

	/* check argument */
	rc = chk_ioc_backcolor(fh, arg, 0);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);
	buffer = info->buffer;

	/* not lock buffer, because does not modify it. */
	{
		unsigned long flags;

		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buffer->buf_lock, flags);

		arg->backcolor = buffer->backcolor;

		spin_unlock_irqrestore(&buffer->buf_lock, flags);
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCS_BACKCOLOR)
******************************************/
static int  iocs_layaddr(struct composer_fh *fh, \
	struct cmp_layaddr *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;
	struct composer_buffer *buffer;

	printk_dbg2(2, "arg level:%d " \
			"addr:%p size:0x%x id:%d offset:%d " \
			"addr2:%p size2:0x%x id2:%d offset2:%d " \
			"addr3:%p size3:0x%x id3:%d offset3:%d\n", \
			arg->level, arg->addr, arg->datasize, \
			arg->app_id, arg->offset, \
			arg->addr_c0, arg->datasize_c0, \
			arg->app_id_c0, arg->offset_c0, \
			arg->addr_c1, arg->datasize_c1, \
			arg->app_id_c1, arg->offset_c1);

	/* check argument */
	rc = chk_ioc_layaddr(fh, arg, 1);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);
	buffer = info->buffer;

	/* lock buffer */
	rc = lock_buffer(fh, idx);
	if (rc < 0) {
		printk_err2("lock_buffer interrupted.\n");
		rc = -EAGAIN;
		goto err_exit;
	}
	if (0 == rc) {
		unsigned long flags;

		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buffer->buf_lock, flags);

		buffer->addr     = arg->addr;
		buffer->app_id   = arg->app_id;
		buffer->offset   = arg->offset;
		buffer->app_addr = NULL;
		buffer->app_size = arg->datasize;

		buffer->addr_c0     = arg->addr_c0;
		buffer->app_id_c0   = arg->app_id_c0;
		buffer->offset_c0   = arg->offset_c0;
		buffer->app_addr_c0 = NULL;
		buffer->app_size_c0 = arg->datasize_c0;

		buffer->addr_c1     = arg->addr_c1;
		buffer->app_id_c1   = arg->app_id_c1;
		buffer->offset_c1   = arg->offset_c1;
		buffer->app_addr_c1 = NULL;
		buffer->app_size_c1 = arg->datasize_c1;

		buffer->buf_id |= BUF_ID_SET_BUFFER;

		spin_unlock_irqrestore(&buffer->buf_lock, flags);

		if (update_work_appshare(info, buffer)) {
			printk_dbg2(2, "app share not mapped.");
			/* app-share will be mapped CMP_IOCS_START. */

			printk_dbg2(2, "spinlock");
			spin_lock_irqsave(&buffer->buf_lock, flags);
			info->id |= INFO_ID_FLG_NEEDAPPSET;
			spin_unlock_irqrestore(&buffer->buf_lock, flags);
		}

		/* update status to change address */
		allfile_status_set(buffer, FH_STATUS_POLLADDR);

		rc = unlock_buffer(fh, idx);
	}
err_exit:
	return rc;
}
/******************************************
* ioctl (CMP_IOCG_BACKCOLOR)
******************************************/
static int  iocg_layaddr(struct composer_fh *fh, \
	struct cmp_layaddr *arg)
{
	int rc;
	int idx;
	struct composer_info   *info;
	struct composer_buffer *buffer;

	printk_dbg2(2, "arg level:%d\n", arg->level);

	/* check argument */
	rc = chk_ioc_layaddr(fh, arg, 0);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	/* convert level to internal-index */
	if (get_index_from_level(arg->level, &idx) != CMP_ERR_OK) {
		printk_err2("argument level %d invalid.\n", arg->level);
		return -EINVAL;
	}

	/* check already opened.  */
	info   = &fh->fh_info[idx];
	BUG_ON(info->buffer == NULL);
	buffer = info->buffer;

	/* not lock buffer, because does not modify it. */
	{
		unsigned long flags;


		printk_dbg2(2, "spinlock");
		spin_lock_irqsave(&buffer->buf_lock, flags);

		arg->addr   = buffer->addr;
		arg->app_id = buffer->app_id;
		arg->offset = buffer->offset;
		arg->addr_c0   = buffer->addr_c0;
		arg->app_id_c0 = buffer->app_id_c0;
		arg->offset_c0 = buffer->offset_c0;
		arg->addr_c1   = buffer->addr_c1;
		arg->app_id_c1 = buffer->app_id_c1;
		arg->offset_c1 = buffer->offset_c1;

		spin_unlock_irqrestore(&buffer->buf_lock, flags);
	}
err_exit:
	return rc;
}
static int  ioc_start(struct composer_fh *fh)
{
	int i, rc;

	/* check argument */
	rc = chk_ioc_start(fh);
	if (rc != 0) {
		/* return by error */
		goto err_exit;
	}

	if (fh->fh_status & FH_STATUS_BLENDING) {
		printk_err2("already start image blending.\n");
		rc = -EBUSY;
		goto err_exit;
	}

	/* to finish previous work. */
	flush_work(&fh->fh_wqtask);

	/* check app-shrea need to create. */
	for (i = 0; i < 5; i++) {
		struct composer_info   *info   = &fh->fh_info[i];
		struct composer_buffer *buffer = info->buffer;

		if ((info->id & INFO_ID_FLG_NEEDAPPSET) == 0) {
			/* nothing to do */
			continue;
		}

		if (update_work_appshare(info, buffer)) {
			printk_err2("app share can not open.");
			rc = -EINVAL;
			goto err_exit;
		} else {
			unsigned long flags;

			printk_dbg2(2, "spinlock");
			spin_lock_irqsave(&buffer->buf_lock, flags);

			info->id &= ~INFO_ID_FLG_NEEDAPPSET;

			spin_unlock_irqrestore(&buffer->buf_lock, flags);
		}
	}

	rc = update_grap_arguments(fh);
	if (rc != 0) {
		printk_err2("can not convert grap argument.\n");
		rc = -EINVAL;
		goto err_exit;
	}
	/* all buffer reserved to this fh */
	rc = setbusy(fh);
	if (rc != 0) {
		printk_err2("can not get buffer lock.\n");
		rc = -EAGAIN;
		goto err_exit;
	}

	fh->fh_status &= ~FH_STATUS_BLENDERR;
	fh->fh_status |= FH_STATUS_BLENDING;

	rc = queue_work(workqueue, &fh->fh_wqtask);
	if (rc) {
		/* to finish previous work. */

		flush_work(&fh->fh_wqtask);

		BUG_ON(fh->fh_status & FH_STATUS_BLENDING);

		setidle(fh);

		rc = 0;
		if (fh->fh_status & FH_STATUS_BLENDERR) {
			printk_err2("compose error found.\n");
			rc = -EINVAL;
		}
	} else {
		fh->fh_status &= ~FH_STATUS_BLENDING;
		printk_err2("unexpectly queue_work not successed..\n");
		rc = -EAGAIN;
		setidle(fh);
	}
err_exit:
	return rc;
}

static int  chk_ioc_supportpixfmt(struct composer_fh *fh, \
	struct cmp_lay_supportpixfmt *arg)
{
	int rc = -EINVAL;
	int flag;
	switch (arg->level) {
	case CMP_OUT_PLANE:
		flag = 0;
		break;
	case CMP_MAIN_LAYER:
	case CMP_CH1_LAYER:
	case CMP_CH2_LAYER:
	case CMP_CH3_LAYER:
		flag = 1;
		break;
	default:
		printk_err2("argument level 0x%x invalid.\n", arg->level);
		goto err_exit;
		break;
	}

	switch (arg->pixfmt) {
	case CMPPIXFMT_ARGB8888:
		if (flag == 0) {
			printk_err2("argument pixfmt 0x%x invalid.\n",
				arg->pixfmt);
			goto err_exit;
		}
		break;
	case CMPPIXFMT_RGB888:
	case CMPPIXFMT_RGB565:
	case CMPPIXFMT_YUV420SP:
	case CMPPIXFMT_YUV422SP:
		/* do not need check flag */
		break;
	case CMPPIXFMT_YUV420PL:
		if (flag == 0) {
			printk_err2("argument pixfmt 0x%x invalid.\n",
				arg->pixfmt);
			goto err_exit;
		}
		break;
	default:
		printk_err2("argument pixfmt 0x%x invalid.\n", arg->pixfmt);
		goto err_exit;
		break;
	}

	/* pass */
	rc = 0;
err_exit:
	return rc;
}
static int  chk_ioc_open(struct composer_fh *fh, \
	struct cmp_lay_open *arg)
{
	int rc = -EINVAL;
	switch (arg->level) {
	case CMP_OUT_PLANE:
	case CMP_MAIN_LAYER:
	case CMP_CH1_LAYER:
	case CMP_CH2_LAYER:
	case CMP_CH3_LAYER:
		break;
	default:
		printk_err2("argument level 0x%x invalid.\n", arg->level);
		goto err_exit;
		break;
	}

	if (arg->bufferid < 0 || arg->bufferid >= MAX_BUFFER) {
		printk_err2("argument level 0x%x invalid.\n", arg->level);
		goto err_exit;
	}

	/* pass */
	rc = 0;
err_exit:
	return rc;
}
static int  chk_ioc_close(struct composer_fh *fh, \
	struct cmp_lay_close *arg)
{
	int rc = -EINVAL;
	switch (arg->level) {
	case CMP_OUT_PLANE:
	case CMP_MAIN_LAYER:
	case CMP_CH1_LAYER:
	case CMP_CH2_LAYER:
	case CMP_CH3_LAYER:
		break;
	default:
		printk_err2("argument level 0x%x invalid.\n", arg->level);
		goto err_exit;
		break;
	}

	/* pass */
	rc = 0;
err_exit:
	return rc;
}
static int  chk_ioc_disp(struct composer_fh *fh, \
	struct cmp_lay_disp *arg)
{
	int rc = -EINVAL;
	switch (arg->level) {
	case CMP_CH1_LAYER:
	case CMP_CH2_LAYER:
	case CMP_CH3_LAYER:
		break;
	default:
		printk_err2("argument level 0x%x invalid.\n", arg->level);
		goto err_exit;
		break;
	}

	if (get_composer_info(fh, arg->level) == NULL) {
		printk_err2("argument level 0x%x not open.\n", arg->level);
		goto err_exit;
	}

	/* arg->on is always valid. */

	/* pass */
	rc = 0;
err_exit:
	return rc;
}
static int  chk_ioc_alpha(struct composer_fh *fh, \
	struct cmp_lay_alpha *arg, int set_flag)
{
	int rc = -EINVAL;
	switch (arg->level) {
	case CMP_CH1_LAYER:
	case CMP_CH2_LAYER:
	case CMP_CH3_LAYER:
		break;
	default:
		printk_err2("argument level 0x%x invalid.\n", arg->level);
		goto err_exit;
		break;
	}

	if (get_composer_info(fh, arg->level) == NULL) {
		printk_err2("argument level 0x%x not open.\n", arg->level);
		goto err_exit;
	}

	/* arg->alpha is always valid. */

	/* pass */
	rc = 0;
err_exit:
	return rc;
}
static int  chk_ioc_keycolor(struct composer_fh *fh, \
	struct cmp_lay_keycolor *arg, int set_flag)
{
	int rc = -EINVAL;
	struct composer_buffer *buf;

	switch (arg->level) {
	case CMP_CH1_LAYER:
	case CMP_CH2_LAYER:
	case CMP_CH3_LAYER:
		break;
	default:
		printk_err2("argument level 0x%x invalid.\n", arg->level);
		goto err_exit;
		break;
	}

	if (get_composer_info(fh, arg->level) == NULL) {
		printk_err2("argument level 0x%x not open.\n", arg->level);
		goto err_exit;
	}

	if (set_flag == 0) {
		/* skip parameter check of keycolor */
		goto pass_exit;
	}
	buf = fh->fh_info[arg->level].buffer;
	if ((buf->buf_id & BUF_ID_FLG_PIXFMT) == 0) {
		/* pixel format not specified */
		printk_dbg2(2, "pixel format not specified."
			" ignore parameter check of keycolor.\n");
	} else {
		if (buf->data_fmt == CMPPIXFMT_RGB565) {
			if (arg->keycolor > 0xffff) {
				printk_err2("argument keycolor 0x%x invalid"
					" for RGB565.\n", arg->keycolor);
				goto err_exit;
			}
		} else if (buf->data_fmt == CMPPIXFMT_RGB888) {
			if (arg->keycolor > 0xffffff) {
				printk_err2("argument keycolor 0x%x invalid"
					" for RGB888.\n", arg->keycolor);
				goto err_exit;
			}
		} else {
			/* ignore parameter */
			/* nothing to do    */
		}
	}
pass_exit:
	/* pass */
	rc = 0;
err_exit:
	return rc;
}
static int  chk_ioc_laysize(struct composer_fh *fh, \
	struct cmp_lay_data_size *arg, int set_flag)
{
	int rc = -EINVAL;
	struct composer_buffer *buf;

	switch (arg->level) {
	case CMP_OUT_PLANE:
	case CMP_MAIN_LAYER:
	case CMP_CH1_LAYER:
	case CMP_CH2_LAYER:
	case CMP_CH3_LAYER:
		break;
	default:
		printk_err2("argument level 0x%x invalid.\n", arg->level);
		goto err_exit;
		break;
	}

	if (get_composer_info(fh, arg->level) == NULL) {
		printk_err2("argument level 0x%x not open.\n", arg->level);
		goto err_exit;
	}

	if (set_flag == 0) {
		/* skip parameter check of laysize */
		goto pass_exit;
	}
	if (arg->x < 1 || arg->x > 1920) {
		printk_err2("argument x invalid. 1-1920\n");
		goto err_exit;
	}
	if (arg->y < 1 || arg->y > 1088) {
		printk_err2("argument y invalid. 1-1088\n");
		goto err_exit;
	}
	buf = fh->fh_info[arg->level].buffer;
	if ((buf->buf_id & BUF_ID_FLG_PIXFMT) == 0) {
		/* pixel format not specified */
		printk_dbg2(2, "pixel format not specified."
			" ignore parameter check of data size.\n");
	} else {
		/* width: odd size check */
		if (buf->data_fmt == CMPPIXFMT_YUV420PL ||
		   buf->data_fmt == CMPPIXFMT_YUV420SP ||
		   buf->data_fmt == CMPPIXFMT_YUV422SP) {
			if (arg->x & 1) {
				printk_err2("argument x invalid. odd\n");
				goto err_exit;
			}
		}
		/* height: odd size check */
		if (buf->data_fmt == CMPPIXFMT_YUV420PL ||
		   buf->data_fmt == CMPPIXFMT_YUV420SP) {
			if (arg->y & 1) {
				printk_err2("argument y invalid. odd\n");
				goto err_exit;
			}
		}
	}
pass_exit:
	/* pass */
	rc = 0;
err_exit:
	return rc;
}
static int  chk_ioc_pixfmt(struct composer_fh *fh, \
	struct cmp_lay_pixfmt *arg, int set_flag)
{
	int rc = -EINVAL;

	switch (arg->level) {
	case CMP_OUT_PLANE:
	case CMP_MAIN_LAYER:
	case CMP_CH1_LAYER:
	case CMP_CH2_LAYER:
	case CMP_CH3_LAYER:
		break;
	default:
		printk_err2("argument level 0x%x invalid.\n", arg->level);
		goto err_exit;
		break;
	}

	if (get_composer_info(fh, arg->level) == NULL) {
		printk_err2("argument level 0x%x not open.\n", arg->level);
		goto err_exit;
	}

	if (set_flag == 0) {
		/* skip parameter check of laysize */
		goto pass_exit;
	}
	switch (arg->pixfmt) {
	case CMPPIXFMT_ARGB8888:
	case CMPPIXFMT_RGB888:
	case CMPPIXFMT_RGB565:
	case CMPPIXFMT_YUV420SP:
	case CMPPIXFMT_YUV422SP:
	case CMPPIXFMT_YUV420PL:
		break;
	default:
		printk_err2("argument pixfmt 0x%x invalid.\n", arg->pixfmt);
		goto err_exit;
		break;
	}
	switch (arg->colorspace) {
	case CMPYUVCOLOR_BT601_COMPRESS:
	case CMPYUVCOLOR_BT601_FULLSCALE:
	case CMPYUVCOLOR_BT709_COMPRESS:
	case CMPYUVCOLOR_BT709_FULLSCALE:
		break;
	default:
		printk_err2("argument colorspace 0x%x invalid.\n",
			arg->colorspace);
		goto err_exit;
		break;
	}
pass_exit:
	/* pass */
	rc = 0;
err_exit:
	return rc;
}
static int  chk_ioc_cropsize(struct composer_fh *fh, \
	struct cmp_lay_cropsize *arg, int set_flag, \
	struct cmp_lay_data_size *laysize)
{
	int rc = -EINVAL;
	struct composer_buffer *buf;

	switch (arg->level) {
	case CMP_OUT_PLANE:
	case CMP_MAIN_LAYER:
	case CMP_CH1_LAYER:
	case CMP_CH2_LAYER:
	case CMP_CH3_LAYER:
		break;
	default:
		printk_err2("argument level 0x%x invalid.\n", arg->level);
		goto err_exit;
		break;
	}

	if (get_composer_info(fh, arg->level) == NULL) {
		printk_err2("argument level 0x%x not open.\n", arg->level);
		goto err_exit;
	}

	if (set_flag == 0) {
		/* skip parameter check of laysize */
		goto pass_exit;
	}
	if (arg->x < 1 || arg->x > 1920) {
		printk_err2("argument x invalid. 1-1920\n");
		goto err_exit;
	}
	if (arg->y < 1 || arg->y > 1088) {
		printk_err2("argument x invalid. 1-1088\n");
		goto err_exit;
	}
	buf = fh->fh_info[arg->level].buffer;
	if (laysize == NULL) {
		if ((buf->buf_id & BUF_ID_FLG_LAYSIZE) != BUF_ID_FLG_LAYSIZE) {
			/* layer size not specified */
			printk_dbg2(2, "layer size not specified."
				" ignore parameter check of crop size.\n");
		} else {
			if (arg->x >  buf->data_x) {
				printk_err2("argument x invalid."
					" over layer size\n");
				goto err_exit;
			}
			if (arg->y > buf->data_y) {
				printk_err2("argument y invalid."
					" over layer size\n");
				goto err_exit;
			}
		}
	} else {
		if (arg->x >  laysize->x) {
			printk_err2("argument x invalid. over layer size\n");
			goto err_exit;
		}
		if (arg->y > laysize->y) {
			printk_err2("argument y invalid. over layer size\n");
			goto err_exit;
		}
	}

	if ((buf->buf_id & BUF_ID_FLG_PIXFMT) == 0) {
		/* pixel format not specified */
		printk_dbg2(2, "pixel format not specified."
			" ignore parameter check of data size.\n");
	} else {
		/* width: odd size check */
		if (buf->data_fmt == CMPPIXFMT_YUV420PL ||
		   buf->data_fmt == CMPPIXFMT_YUV420SP ||
		   buf->data_fmt == CMPPIXFMT_YUV422SP) {
			if (arg->x & 1) {
				printk_err2("argument x invalid. odd\n");
				goto err_exit;
			}
		}
		/* height: odd size check */
		if (buf->data_fmt == CMPPIXFMT_YUV420PL ||
		   buf->data_fmt == CMPPIXFMT_YUV420SP) {
			if (arg->y & 1) {
				printk_err2("argument y invalid. odd\n");
				goto err_exit;
			}
		}
	}
pass_exit:
	/* pass */
	rc = 0;
err_exit:
	return rc;
}
static int  chk_ioc_croppos(struct composer_fh *fh, \
	struct cmp_lay_croppos *arg, int set_flag, \
	struct cmp_lay_data_size *laysize)
{
	int rc = -EINVAL;
	struct composer_buffer *buf;

	switch (arg->level) {
	case CMP_OUT_PLANE:
	case CMP_MAIN_LAYER:
	case CMP_CH1_LAYER:
	case CMP_CH2_LAYER:
	case CMP_CH3_LAYER:
		break;
	default:
		printk_err2("argument level 0x%x invalid.\n", arg->level);
		goto err_exit;
		break;
	}

	if (get_composer_info(fh, arg->level) == NULL) {
		printk_err2("argument level 0x%x not open.\n", arg->level);
		goto err_exit;
	}

	if (set_flag == 0) {
		/* skip parameter check of laysize */
		goto pass_exit;
	}
	if (arg->x < 0 || arg->x > 1919) {
		printk_err2("argument x invalid. 0-1919\n");
		goto err_exit;
	}
	if (arg->y < 0 || arg->y > 1087) {
		printk_err2("argument x invalid. 0-1087\n");
		goto err_exit;
	}
	buf = fh->fh_info[arg->level].buffer;

	if (laysize == NULL) {
		if ((buf->buf_id & BUF_ID_FLG_LAYSIZE) != BUF_ID_FLG_LAYSIZE) {
			/* layer size not specified */
			printk_dbg2(2, "layer size not specified."
				" ignore parameter check of crop size.\n");
		} else {
			if (arg->x >=  buf->data_x) {
				printk_err2("argument x invalid."
					" over layer size\n");
				goto err_exit;
			}
			if (arg->y >= buf->data_y) {
				printk_err2("argument y invalid."
					" over layer size\n");
				goto err_exit;
			}
		}
	} else {
		if (arg->x >=  laysize->x) {
			printk_err2("argument x invalid. over layer size\n");
			goto err_exit;
		}
		if (arg->y >= laysize->y) {
			printk_err2("argument y invalid. over layer size\n");
			goto err_exit;
		}
	}

	if ((buf->buf_id & BUF_ID_FLG_PIXFMT) == 0) {
		/* pixel format not specified */
		printk_dbg2(2, "pixel format not specified."
			" ignore parameter check of data size.\n");
	} else {
		/* width: odd size check */
		if (buf->data_fmt == CMPPIXFMT_YUV420PL ||
		   buf->data_fmt == CMPPIXFMT_YUV420SP ||
		   buf->data_fmt == CMPPIXFMT_YUV422SP) {
			if (arg->x & 1) {
				printk_err2("argument x invalid. odd\n");
				goto err_exit;
			}
		}
		/* height: odd size check */
		if (buf->data_fmt == CMPPIXFMT_YUV420PL ||
		   buf->data_fmt == CMPPIXFMT_YUV420SP) {
			if (arg->y & 1) {
				printk_err2("argument y invalid. odd\n");
				goto err_exit;
			}
		}
	}
pass_exit:
	/* pass */
	rc = 0;
err_exit:
	return rc;
}
static int  chk_ioc_pos(struct composer_fh *fh, \
	struct cmp_lay_pos *arg, int set_flag)
{
	int rc = -EINVAL;
	struct composer_buffer *buf;
	int   target_x, target_y;

	switch (arg->level) {
	case CMP_CH1_LAYER:
	case CMP_CH2_LAYER:
	case CMP_CH3_LAYER:
		break;
	default:
		printk_err2("argument level 0x%x invalid.\n", arg->level);
		goto err_exit;
		break;
	}

	if (get_composer_info(fh, arg->level) == NULL) {
		printk_err2("argument level 0x%x not open.\n", arg->level);
		goto err_exit;
	}

	if (set_flag == 0) {
		/* skip parameter check of laysize */
		goto pass_exit;
	}
	if (arg->x < 0 || arg->x > 1919) {
		printk_err2("argument x invalid. 0-1919\n");
		goto err_exit;
	}
	if (arg->y < 0 || arg->y > 1087) {
		printk_err2("argument x invalid. 0-1087\n");
		goto err_exit;
	}
	buf = fh->fh_info[IDX_OUTPLANE].buffer;

	if ((buf->buf_id & BUF_ID_FLG_CROPSIZE) == BUF_ID_FLG_CROPSIZE) {
		/* crop size specified */
		target_x = buf->crop_size_x;
		target_y = buf->crop_size_y;
	} else if ((buf->buf_id & BUF_ID_FLG_LAYSIZE) == BUF_ID_FLG_LAYSIZE) {
		/* crop size not specified */
		target_x = buf->data_x;
		target_y = buf->data_y;
	} else {
		printk_dbg2(2, "layer size not specified."
			" ignore parameter check of blend pos.\n");
		target_x = 0;
		target_y = 0;
	}

	if (target_x != 0 && target_y != 0) {
		if (arg->x >=  target_x) {
			printk_err2("argument x invalid."
				" over out plpne size\n");
			goto err_exit;
		}
		if (arg->y >= target_y) {
			printk_err2("argument y invalid."
				" over out plane size\n");
			goto err_exit;
		}
	}

	if ((buf->buf_id & BUF_ID_FLG_PIXFMT) == 0) {
		/* pixel format not specified */
		printk_dbg2(2, "pixel format not specified."
			" ignore parameter check of data size.\n");
	} else {
		/* width: odd size check */
		if (buf->data_fmt == CMPPIXFMT_YUV420PL ||
		   buf->data_fmt == CMPPIXFMT_YUV420SP ||
		   buf->data_fmt == CMPPIXFMT_YUV422SP) {
			if (arg->x & 1) {
				printk_err2("argument x invalid. odd\n");
				goto err_exit;
			}
		}
		/* height: odd size check */
		if (buf->data_fmt == CMPPIXFMT_YUV420PL ||
		   buf->data_fmt == CMPPIXFMT_YUV420SP) {
			if (arg->y & 1) {
				printk_err2("argument y invalid. odd\n");
				goto err_exit;
			}
		}
	}
pass_exit:
	/* pass */
	rc = 0;
err_exit:
	return rc;
}
static int  chk_ioc_size(struct composer_fh *fh, \
	struct cmp_lay_compose_size *arg, int set_flag)
{
	int rc = -EINVAL;
	struct composer_buffer *buf;
	int   target_x, target_y;

	switch (arg->level) {
	case CMP_CH1_LAYER:
	case CMP_CH2_LAYER:
	case CMP_CH3_LAYER:
		break;
	default:
		printk_err2("argument level 0x%x invalid.\n", arg->level);
		goto err_exit;
		break;
	}

	if (get_composer_info(fh, arg->level) == NULL) {
		printk_err2("argument level 0x%x not open.\n", arg->level);
		goto err_exit;
	}

	if (set_flag == 0) {
		/* skip parameter check of laysize */
		goto pass_exit;
	}
	if (arg->x < 1 || arg->x > 1920) {
		printk_err2("argument x invalid. 0-1919\n");
		goto err_exit;
	}
	if (arg->y < 1 || arg->y > 1088) {
		printk_err2("argument x invalid. 0-1087\n");
		goto err_exit;
	}
	buf = fh->fh_info[IDX_OUTPLANE].buffer;

	if ((buf->buf_id & BUF_ID_FLG_CROPSIZE) == BUF_ID_FLG_CROPSIZE) {
		/* crop size specified */
		target_x = buf->crop_size_x;
		target_y = buf->crop_size_y;
	} else if ((buf->buf_id & BUF_ID_FLG_LAYSIZE) == BUF_ID_FLG_LAYSIZE) {
		/* layer size not specified */
		target_x = buf->data_x;
		target_y = buf->data_y;
	} else {
		printk_dbg2(2, "layer size not specified."
			" ignore parameter check of blend pos.\n");
		target_x = 0;
		target_y = 0;
	}

	if (target_x != 0 && target_y != 0) {
		if (arg->x >  target_x) {
			printk_err2("argument x invalid."
				" over out plpne size\n");
			goto err_exit;
		}
		if (arg->y > target_y) {
			printk_err2("argument y invalid."
				" over out plane size\n");
			goto err_exit;
		}
	}

	if ((buf->buf_id & BUF_ID_FLG_PIXFMT) == 0) {
		/* pixel format not specified */
		printk_dbg2(2, "pixel format not specified."
			" ignore parameter check of data size.\n");
	} else {
		/* width: odd size check */
		if (buf->data_fmt == CMPPIXFMT_YUV420PL ||
		   buf->data_fmt == CMPPIXFMT_YUV420SP ||
		   buf->data_fmt == CMPPIXFMT_YUV422SP) {
			if (arg->x & 1) {
				printk_err2("argument x invalid. odd\n");
				goto err_exit;
			}
		}
		/* height: odd size check */
		if (buf->data_fmt == CMPPIXFMT_YUV420PL ||
		   buf->data_fmt == CMPPIXFMT_YUV420SP) {
			if (arg->y & 1) {
				printk_err2("argument y invalid. odd\n");
				goto err_exit;
			}
		}
	}
pass_exit:
	/* pass */
	rc = 0;
err_exit:
	return rc;
}

#define VIEW_MASK(X) ((X) & 0x80000000)

static int  chk_ioc_viewlay(struct composer_fh *fh, \
	struct cmp_viewlay *arg, int idx)
{
	int               rc = -EINVAL;
	int               flag;
	struct cmp_lay_data_size laysize;
	struct cmp_lay_cropsize  cropsize;
	struct cmp_lay_croppos   croppos;
	struct cmp_lay_pos       pos;
	struct cmp_lay_compose_size size;
	const struct composer_buffer *buffer;

	switch (arg->level) {
	case CMP_OUT_PLANE:
	case CMP_MAIN_LAYER:
		flag = 0x0;
		break;
	case CMP_CH1_LAYER:
	case CMP_CH2_LAYER:
	case CMP_CH3_LAYER:
		flag = 0x1;
		break;
	default:
		printk_err2("argument level 0x%x invalid.\n", arg->level);
		goto err_exit;
		break;
	}

	if (get_composer_info(fh, arg->level) == NULL) {
		printk_err2("argument level 0x%x not open.\n", arg->level);
		goto err_exit;
	}

	if (idx < 0) {
		/* skip parameter check of viewlay */
		goto pass_exit;
	}
	buffer = fh->fh_info[idx].buffer;
	if (buffer == NULL) {
		rc = 0;
		goto err_exit;
	}

	/* check configure paramter */
	{
		int    chk1, chk2, need_init_flag;

		/* set laysize needed. */

		chk1 = (VIEW_MASK(arg->data_x) == 0) ? 1 : 0;
		chk2 = (VIEW_MASK(arg->data_y) == 0) ? 1 : 0;
		need_init_flag = ((buffer->buf_id & BUF_ID_FLG_LAYSIZE) \
				!= BUF_ID_FLG_LAYSIZE) ? 1 : 0;

		if (chk1 || chk2) {
			flag |= (0x10);
			if (chk1 && chk2) {
				/* no need initialize */
			} else if (need_init_flag) {
				rc = CMP_ERR_NG_VIEWLAY1;
				printk_err2("CMP_IOCS_LAYSIZE not performed.");
				goto err_exit;
			}
		}

		/* set cropsize needed. */
		chk1 = (VIEW_MASK(arg->cropsize_x) == 0) ? 1 : 0;
		chk2 = (VIEW_MASK(arg->cropsize_y) == 0) ? 1 : 0;
		need_init_flag = ((buffer->buf_id & BUF_ID_FLG_CROPSIZE) \
				!= BUF_ID_FLG_CROPSIZE) ? 1 : 0;

		if (chk1 || chk2) {
			flag |= (0x20);
			if (chk1 && chk2) {
				/* no need initialize */
			} else if (need_init_flag) {
				rc = CMP_ERR_NG_VIEWLAY3;
				printk_err2("CMP_IOCS_CROPSIZE not performed.");
				goto err_exit;
			}

			if ((flag & 0x10) == 0 && \
				(buffer->buf_id & BUF_ID_FLG_LAYSIZE) \
				!= BUF_ID_FLG_LAYSIZE) {
				rc = CMP_ERR_NG_VIEWLAY1;
				printk_err2("CMP_IOCS_LAYSIZE not performed.");
				goto err_exit;
			}
		}

		/* set croppos needed. */
		chk1 = (VIEW_MASK(arg->croppos_x) == 0) ? 1 : 0;
		chk2 = (VIEW_MASK(arg->croppos_y) == 0) ? 1 : 0;
		need_init_flag = ((buffer->buf_id & BUF_ID_FLG_CROPPOS) \
				!= BUF_ID_FLG_CROPPOS) ? 1 : 0;
		if (chk1 || chk2) {
			flag |= (0x40);
			if (chk1 && chk2) {
				/* no need initialize */
			} else if (need_init_flag) {
				rc = CMP_ERR_NG_VIEWLAY2;
				printk_err2("CMP_IOCS_CROPPOS not performed.");
				goto err_exit;
			}
			if ((flag & 0x10) == 0 && \
				(buffer->buf_id & BUF_ID_FLG_LAYSIZE) \
				!= BUF_ID_FLG_LAYSIZE) {
				rc = CMP_ERR_NG_VIEWLAY1;
				printk_err2("CMP_IOCS_LAYSIZE not performed.");
				goto err_exit;
			}
		}

		/* set size needed. */
		chk1 = (VIEW_MASK(arg->compose_x) == 0) ? 1 : 0;
		chk2 = (VIEW_MASK(arg->compose_y) == 0) ? 1 : 0;
		need_init_flag = ((buffer->buf_id & BUF_ID_FLG_SIZE) \
				!= BUF_ID_FLG_SIZE) ? 1 : 0;

		if (chk1 || chk2) {
			flag |= (0x80);
			if (chk1 && chk2) {
				/* no need initialize */
			} else if (need_init_flag) {
				rc = CMP_ERR_NG_VIEWLAY4;
				printk_err2("CMP_IOCS_SIZE not performed.");
				goto err_exit;
			}
		}

		/* set pos needed. */
		chk1 = (VIEW_MASK(arg->pos_x) == 0) ? 1 : 0;
		chk2 = (VIEW_MASK(arg->pos_y) == 0) ? 1 : 0;
		need_init_flag = ((buffer->buf_id & BUF_ID_FLG_POS) \
				!= BUF_ID_FLG_POS) ? 1 : 0;

		if (chk1 || chk2) {
			flag |= (0x100);
			if (chk1 && chk2) {
				/* no need initialize */
			} else if (need_init_flag) {
				rc = CMP_ERR_NG_VIEWLAY5;
				printk_err2("CMP_IOCS_POS not performed.");
				goto err_exit;
			}
		}
	}


	/* create laysize parameter */
	if (flag & 0x10) {
		if (VIEW_MASK(arg->data_x)) {
			/* set argument from previous configuration. */
			arg->data_x = buffer->data_x;
		}
		if (VIEW_MASK(arg->data_y)) {
			/* set argument from previous configuration. */
			arg->data_y = buffer->data_y;
		}
		laysize.level = arg->level;
		laysize.x     = arg->data_x;
		laysize.y     = arg->data_y;
	}

	/* create croppos parameter */
	if (flag & 0x40) {
		if (VIEW_MASK(arg->croppos_x)) {
			/* set argument from previous configuration. */
			arg->croppos_x = buffer->crop_pos_x;
		}
		if (VIEW_MASK(arg->croppos_y)) {
			/* set argument from previous configuration. */
			arg->croppos_y = buffer->crop_pos_y;
		}
		croppos.level = arg->level;
		croppos.x = arg->croppos_x;
		croppos.y = arg->croppos_y;
	}

	/* create cropsize parameter */
	if (flag & 0x20) {
		if (VIEW_MASK(arg->cropsize_x)) {
			/* set argument from previous configuration. */
			arg->cropsize_x = buffer->crop_size_x;
		}
		if (VIEW_MASK(arg->cropsize_y)) {
			/* set argument from previous configuration. */
			arg->cropsize_y = buffer->crop_size_y;
		}
		cropsize.level = arg->level;
		cropsize.x = arg->cropsize_x;
		cropsize.y = arg->cropsize_y;
	}

	/* create size parameter */
	if (flag & 0x80) {
		if (VIEW_MASK(arg->compose_x)) {
			/* set argument from previous configuration. */
			arg->compose_x = buffer->size_x;
		}
		if (VIEW_MASK(arg->compose_y)) {
			/* set argument from previous configuration. */
			arg->compose_y = buffer->size_y;
		}
		size.level = arg->level;
		size.x = arg->compose_x;
		size.y = arg->compose_y;
	}

	/* create size parameter */
	if (flag & 0x100) {
		if (VIEW_MASK(arg->pos_x)) {
			/* set argument from previous configuration. */
			arg->pos_x = buffer->pos_x;
		}
		if (VIEW_MASK(arg->pos_y)) {
			/* set argument from previous configuration. */
			arg->pos_y = buffer->pos_y;
		}
		pos.level = arg->level;
		pos.x = arg->pos_x;
		pos.y = arg->pos_y;
	}

	if ((flag & 0x010) && chk_ioc_laysize(fh, &laysize, 1) != 0) {
		rc = CMP_ERR_NG_VIEWLAY1;
		goto err_exit;
	}

	if ((flag & 0x020) && chk_ioc_cropsize(fh, &cropsize, 1, &laysize)
		!= 0) {
		rc = CMP_ERR_NG_VIEWLAY2;
		goto err_exit;
	}

	if ((flag & 0x040) && chk_ioc_croppos(fh, &croppos, 1, &laysize) != 0) {
		rc = CMP_ERR_NG_VIEWLAY3;
		goto err_exit;
	}

	if ((flag & 0x081) == 0x081 && chk_ioc_size(fh, &size, 1) != 0) {
		rc = CMP_ERR_NG_VIEWLAY4;
		goto err_exit;
	}

	if ((flag & 0x101) == 0x101 && chk_ioc_pos(fh, &pos, 1) != 0) {
		rc = CMP_ERR_NG_VIEWLAY5;
		goto err_exit;
	}
pass_exit:
	/* pass */
	rc = 0;
err_exit:
	return rc;
}

#undef  VIEW_MASK

static int  chk_ioc_backcolor(struct composer_fh *fh, \
	struct cmp_lay_backcolor *arg, int set_flag)
{
	int rc = -EINVAL;

	switch (arg->level) {
	case CMP_OUT_PLANE:
		break;
	default:
		printk_err2("argument level 0x%x invalid.\n", arg->level);
		goto err_exit;
		break;
	}

	if (get_composer_info(fh, arg->level) == NULL) {
		printk_err2("argument level 0x%x not open.\n", arg->level);
		goto err_exit;
	}

	if (set_flag == 0) {
		/* skip parameter check of keycolor */
		goto pass_exit;
	}
	if (arg->backcolor > 0xffffff) {
		printk_err2("argument backcolor 0x%x invalid.\n",
			arg->backcolor);
		goto err_exit;
	}
pass_exit:
	/* pass */
	rc = 0;
err_exit:
	return rc;
}
static int  chk_ioc_layaddr(struct composer_fh *fh, \
	struct cmp_layaddr *arg, int set_flag)
{
	int rc = -EINVAL;
	struct composer_info   *info;
	struct composer_buffer *buf;
	int   target_x, target_y, target_pixfmt;

	switch (arg->level) {
	case CMP_OUT_PLANE:
	case CMP_MAIN_LAYER:
	case CMP_CH1_LAYER:
	case CMP_CH2_LAYER:
	case CMP_CH3_LAYER:
		break;
	default:
		printk_err2("argument level 0x%x invalid.\n", arg->level);
		goto err_exit;
		break;
	}

	info = get_composer_info(fh, arg->level);
	if (info == NULL) {
		printk_err2("argument level 0x%x not open.\n", arg->level);
		goto err_exit;
	}

	if (set_flag == 0) {
		/* skip parameter check of laysize */
		goto pass_exit;
	}
	buf = info->buffer;

	if ((buf->buf_id & BUF_ID_FLG_LAYSIZE) == BUF_ID_FLG_LAYSIZE) {
		/* layer size not specified */
		target_x = buf->data_x;
		target_y = buf->data_y;
	} else {
		printk_dbg2(2, "layer size not specified."
			" ignore parameter check of set adr.\n");
		target_x = 0;
		target_y = 0;
	}

	if ((buf->buf_id & BUF_ID_FLG_PIXFMT)) {
		/* layer size not specified */
		target_pixfmt = buf->data_fmt;
	} else {
		printk_dbg2(2, "pixel format not specified."
			" ignore parameter check of set adr.\n");
		target_pixfmt = -1;
	}

	if (target_x != 0 && target_y != 0 && target_pixfmt > 0) {
		int min_y_size = 0, min_c_size = 0, min_c1_size = 0;
		switch (target_pixfmt) {
		case CMPPIXFMT_ARGB8888:
			min_y_size = target_x * target_y * 4;
			min_c_size = 0;
			break;
		case CMPPIXFMT_RGB888:
			min_y_size = target_x * target_y * 3;
			min_c_size = 0;
			break;
		case CMPPIXFMT_RGB565:
			min_y_size = target_x * target_y * 2;
			min_c_size = 0;
			break;
		case CMPPIXFMT_YUV420SP:
			min_y_size = target_x * target_y;
			min_c_size = target_x * target_y / 2;
			break;
		case CMPPIXFMT_YUV422SP:
			min_y_size = target_x * target_y;
			min_c_size = target_x * target_y;
			break;
		case CMPPIXFMT_YUV420PL:
			min_y_size = target_x * target_y;
			min_c_size = target_x * target_y / 4;
			min_c1_size = min_c_size;
			break;
		}

		if (arg->datasize <  min_y_size) {
			printk_err2("argument datasize is too small"
				" for %dx%d fmt:%d\n",
				target_x, target_y, target_pixfmt);
			goto err_exit;
		}
		if (min_c_size > 0 &&  arg->datasize_c0 <  min_c_size) {
			printk_err2("argument datasize_c0 is too small"
				" for %dx%d fmt:%d\n",
				target_x, target_y, target_pixfmt);
			goto err_exit;
		}
		if (min_c1_size > 0 && arg->datasize_c1 <  min_c1_size) {
			printk_err2("argument datasize_c1 is too small"
				" for %dx%d fmt:%d\n",
				target_x, target_y, target_pixfmt);
			goto err_exit;
		}
	}
pass_exit:
	/* pass */
	rc = 0;
err_exit:
	return rc;
}
static int  chk_ioc_start(struct composer_fh *fh)
{
	int i;
	int rc = 0;

	static const int fh_status_translate[] = {
		[CMP_OUT_PLANE]  = FH_STATUS_OUTPLANE,
		[CMP_MAIN_LAYER] = FH_STATUS_MAINLAYER,
		[CMP_CH1_LAYER]  = FH_STATUS_CH1LAYER,
		[CMP_CH2_LAYER]  = FH_STATUS_CH2LAYER,
		[CMP_CH3_LAYER]  = FH_STATUS_CH3LAYER
	};

	static const int buffer_mask[] = {
		[CMP_OUT_PLANE]  = BUF_ID_SET_BUFFER | \
			BUF_ID_FLG_PIXFMT | BUF_ID_FLG_LAYSIZE,
		[CMP_MAIN_LAYER] = BUF_ID_SET_BUFFER | \
			BUF_ID_FLG_PIXFMT | BUF_ID_FLG_LAYSIZE,
		[CMP_CH1_LAYER] = BUF_ID_SET_BUFFER | BUF_ID_FLG_PIXFMT \
			| BUF_ID_FLG_LAYSIZE | BUF_ID_FLG_POS | BUF_ID_FLG_SIZE,
		[CMP_CH2_LAYER] = BUF_ID_SET_BUFFER | BUF_ID_FLG_PIXFMT \
			| BUF_ID_FLG_LAYSIZE | BUF_ID_FLG_POS | BUF_ID_FLG_SIZE,
		[CMP_CH3_LAYER] = BUF_ID_SET_BUFFER | BUF_ID_FLG_PIXFMT \
			| BUF_ID_FLG_LAYSIZE | BUF_ID_FLG_POS | BUF_ID_FLG_SIZE
	};

	static unsigned int src_pos_x[5],  src_pos_y[5];
	static unsigned int src_size_w[5], src_size_h[5];
	int    resize_idx;

	/* check open */
	if ((fh->fh_status & FH_STATUS_OUTPLANE) == 0) {
		printk_dbg2(2, "does not open CMP_OUT_PLANE\n");
		rc = CMP_ERR_NG_OPEN_OUT;
		goto err_exit;
	}

	if ((fh->fh_status & FH_STATUS_MAINLAYER) == 0) {
		printk_dbg2(2, "does not open CMP_MAIN_LAYER\n");
		rc = CMP_ERR_NG_OPEN_MAIN;
		goto err_exit;
	}

	/* check parameter */
	for (i = 0; i < 5; i++) {
		unsigned int crop_x, crop_y, crop_w, crop_h;
		struct composer_info   *info;
		struct composer_buffer *buf;
		if ((fh->fh_status & fh_status_translate[i]) == 0) {
			/* not open */
			continue;
		}
		info = &fh->fh_info[i];
		buf = fh->fh_info[i].buffer;

		/* check buffer configuration */
		if (info->id & INFO_ID_FLG_NOTDISP) {
			/* not blending plane */
			continue;
		}

		/* check necessary configuration */
		if ((buf->buf_id & buffer_mask[i]) != buffer_mask[i]) {
			rc = CMP_ERR_NG_INSUFFICIENT_L0 - i;
			printk_err2("configuration not enough idx:%d 0x%x"
				" require:0x%x\n", i,
				buf->buf_id, buffer_mask[i]);
			goto err_exit;
		}

		/* set default image size */
		crop_x = 0;           crop_y = 0;
		crop_w = buf->data_x; crop_h = buf->data_y;

		if ((buf->buf_id & BUF_ID_FLG_CROPPOS)
			== BUF_ID_FLG_CROPPOS) {
			crop_x = buf->crop_pos_x;
			crop_y = buf->crop_pos_y;
		}
		if ((buf->buf_id & BUF_ID_FLG_CROPSIZE)
			== BUF_ID_FLG_CROPSIZE) {
			crop_w = buf->crop_size_x;
			crop_h = buf->crop_size_y;
		}

		/* check crop parameter */
		if (crop_x          >= buf->data_x || \
		     crop_w          >  buf->data_x || \
		     crop_x + crop_w >  buf->data_x || \
		     crop_y          >= buf->data_y || \
		     crop_h          >  buf->data_y || \
		     crop_y + crop_h >  buf->data_y) {
			rc = CMP_ERR_NG_LAYERPOS_L0 - i;
			printk_err2("src area invalid. " \
				"(%d, %d)-(%d, %d), valid:(0, 0)-(%d, %d)\n",
					crop_x, crop_y, crop_x + crop_w,
					crop_y + crop_h,
					buf->data_x, buf->data_y);
			goto err_exit;
		}

		/* record src image size */
		src_pos_x[i] = crop_x;
		src_pos_y[i] = crop_y;
		src_size_w[i] = crop_w;
		src_size_h[i] = crop_h;
		printk_dbg2(2, "%d: src_pos(%d, %d), src_size(%d, %d)\n",
			i, src_pos_x[i], src_pos_y[i],
			src_size_w[i], src_size_h[i]);
	}

	if (src_size_w[IDX_OUTPLANE] != src_size_w[IDX_MAINLAYER] ||
	    src_size_h[IDX_OUTPLANE] != src_size_h[IDX_MAINLAYER]) {
		printk_err2("OUT and Main layer size invalid. "
			"out(%d, %d) main(%d, %d)\n", \
			src_size_w[IDX_OUTPLANE], src_size_h[IDX_OUTPLANE], \
			src_size_w[IDX_MAINLAYER], src_size_h[IDX_MAINLAYER]);
		rc = CMP_ERR_NG_LAYERPOS_L1;
		goto err_exit;
	}

	/* check destination size */
	resize_idx = -1;
	for (i = 0; i < 5; i++) {
		unsigned int dst_x, dst_y, dst_w, dst_h;

		struct composer_info   *info;
		struct composer_buffer *buf;

		if ((fh->fh_status & fh_status_translate[i]) == 0) {
			/* not open */
			continue;
		}
		info = &fh->fh_info[i];
		buf = fh->fh_info[i].buffer;

		/* check buffer configuration */
		if (info->id & INFO_ID_FLG_NOTDISP) {
			/* not blending plane */
			continue;
		}
		/* set default image size */
		if ((buf->buf_id & (BUF_ID_FLG_POS | BUF_ID_FLG_SIZE))
			!= (BUF_ID_FLG_POS | BUF_ID_FLG_SIZE)) {
			/* destination parameter not supported. */
			continue;
		}

		dst_x = buf->pos_x;  dst_y = buf->pos_y;
		dst_w = buf->size_x; dst_h = buf->size_y;

		printk_dbg2(2, "%d: dst_pos(%d, %d), dst_size(%d, %d)\n",
			i, dst_x, dst_y, dst_w, dst_h);

		if (dst_x         >= src_size_w[IDX_MAINLAYER] || \
		     dst_w         >  src_size_w[IDX_MAINLAYER] || \
		     dst_x + dst_w >  src_size_w[IDX_MAINLAYER] || \
		     dst_y         >= src_size_h[IDX_MAINLAYER] || \
		     dst_h         >  src_size_h[IDX_MAINLAYER] || \
		     dst_y + dst_h >  src_size_h[IDX_MAINLAYER]) {
			rc = CMP_ERR_NG_LAYERPOS_L0 - i;
			printk_err2("dst area invalid. " \
				"(%d, %d)-(%d, %d), valid:(0, 0)-(%d, %d)\n",
					dst_x, dst_y, dst_x + dst_w,
					dst_y + dst_h,
					src_size_w[IDX_MAINLAYER],
					src_size_h[IDX_MAINLAYER]);
			goto err_exit;
		}

		if (src_size_w[i] != dst_w || src_size_h[i] != dst_h) {
			if (resize_idx < 0) {
				/* found resize channel */
				resize_idx = i;
			} else {
				printk_err2("already request resize at %d\n",
					resize_idx);
				rc = CMP_ERR_NG_NORESOURCE;
				goto err_exit;
			}

			if (dst_w < 16 || dst_h < 16 ||
			   src_size_w[i] < 4 || src_size_h[i] < 4 ||
			   dst_w           > src_size_w[i]  * 16 ||
			   src_size_w[i]    > dst_w         * 16 ||
			   dst_h            > src_size_h[i] * 16 ||
			   src_size_h[i]    > dst_h         * 16) {
				printk_err2("not supported scaler"
					" (%d, %d)->(%d, %d)\n",
					src_size_w[i], src_size_h[i],
					dst_w, dst_h);
				rc = CMP_ERR_NG_RESIZE_L0-i;
				goto err_exit;
			}
		}
	}

	/* pass */
	rc = 0;
err_exit:
	return rc;
}


static void notify_graphics_image_conv(int result, unsigned long user_data)
{
	/* currently not implemented. */
	printk_err1("callback unexpected.");
}
static void notify_graphics_image_blend(int result, unsigned long user_data)
{
	struct composer_grapdata *grap = (struct composer_grapdata *)user_data;

#if _ERR_DBG
	if (debug >= 1) {
		struct list_head *list;
		int    match = 0;

		printk_dbg2(2, "result:0x%x user_data:0x%lx\n ",
			result, user_data);

		list_for_each(list, &file_top)
		{
			struct composer_fh *fh;

			fh = list_entry(list, struct composer_fh, fh_filelist);
			printk_dbg2(2, "list of valid user_data: %p\n", \
				&fh->grap_data);

			if (user_data == (unsigned long)&fh->grap_data) {
				match = 1;
				break;
			}
		}
		if (!match) {
			printk_err("user_data 0x%lx unexpected."
				" ignore callback\n", user_data);
			return;
		}
	}
#endif

	if (result != SMAP_LIB_GRAPHICS_OK) {
		printk_err1("notify_graphics_image_blend result:%d\n", result);
		grap->status = 3;
	} else {
		grap->status = 1;
	}
	wake_up_interruptible_all(&grap->wait_notify);
}

#if _LOG_DBG >= 2
static void dump_screen_grap_initialize(screen_grap_initialize *arg)
{
	printk_dbg2(1, "screen_grap_initialize\n");
	printk_dbg2(1, "  handle:%p mode:%ld\n", arg->handle, arg->mode);
}

static void dump_address_data_to_string(
	unsigned char *mes, const unsigned char *ptr)
{
	*mes = 0;
	sprintf(mes, "%02x %02x %02x %02x %02x %02x %02x %02x :"
		"%02x %02x %02x %02x %02x %02x %02x %02x",
		ptr[0], ptr[1], ptr[2],  ptr[3],
		ptr[4], ptr[5], ptr[6],  ptr[7],
		ptr[8], ptr[9], ptr[10], ptr[11],
		ptr[12], ptr[13], ptr[14], ptr[15]);
}


static void dump_screen_grap_image_param(
	screen_grap_image_param * arg, char *name)
{
	printk_dbg2(1, " screen_grap_image_param[%s]\n", name);
	printk_dbg2(1, "  width:%d height:%d stride:%d stride_c:%d "
		"format:%d yuv_format:%d yuv_range:%d "
		"address:(%p, %p) address_c0(%p, %p) address_c1(%p, %p)\n",
		arg->width, arg->height, arg->stride, arg->stride_c,
		arg->format, arg->format, arg->yuv_range,
		arg->address, arg->apmem_handle,
		arg->address_c0, arg->apmem_handle_c0,
		arg->address_c1, arg->apmem_handle_c1);
	if (debug <= 2) {
		/* only report RT-API argument */
		return;
	}
	if (arg->apmem_handle && arg->address) {
		int i;
		char msg[256];
		const unsigned char *ptr = arg->address;
		printk_dbg2(2, "data in address");
		for (i = 0; i < 256; i += 16) {
			dump_address_data_to_string(&msg[0], ptr+i);
			printk_dbg2(1, "  %s\n", msg);
		}
	}
	if (arg->apmem_handle_c0 && arg->address_c0) {
		int i;
		char msg[256];
		const unsigned char *ptr = arg->address_c0;
		printk_dbg2(1, "data in address_c0\n");
		for (i = 0; i < 256; i += 16) {
			dump_address_data_to_string(&msg[0], ptr+i);
			printk_dbg2(1, "  %s\n", msg);
		}
	}
	if (arg->apmem_handle_c1 && arg->address_c1) {
		int i;
		char msg[256];
		const unsigned char *ptr = arg->address_c1;
		printk_dbg2(1, "data in address_c1\n");
		for (i = 0; i < 256; i += 16) {
			dump_address_data_to_string(&msg[0], ptr+i);
			printk_dbg2(1, "  %s\n", msg);
		}
	}
}

static void dump_screen_grap_layer(screen_grap_layer *arg, char *_name)
{
	static char name[256];

	if (arg == NULL) {
		/* layer not opened */
		return;
	}
	printk_dbg2(1, " screen_grap_layer[%s]\n", _name);
	sprintf(name, "%s.image", _name);
	dump_screen_grap_image_param(&arg->image, name);
	printk_dbg2(1, " %s.rect(x:%d y:%d width:%d height:%d) "
		"alpha:%d rotate:%d mirror:%d key_color:0x%x\n",
		name,
		arg->rect.x, arg->rect.y, arg->rect.width, arg->rect.height,
		arg->alpha, arg->rotate, arg->mirror, arg->key_color);
}

static void dump_screen_grap_image_blend(screen_grap_image_blend *arg)
{
	printk_dbg2(1, "screen_grap_image_blend\n");

	printk_dbg2(1,
		"  handle:%p input_layer:(%p %p %p %p) "
		"background_color:0x%lx user_data:0x%lx\n",
		arg->handle, arg->input_layer[0], arg->input_layer[1],
		arg->input_layer[2], arg->input_layer[3],
		arg->background_color, arg->user_data);
	dump_screen_grap_image_param(&arg->output_image, "output_image");
	dump_screen_grap_layer(arg->input_layer[0], "input_layer0");
	dump_screen_grap_layer(arg->input_layer[1], "input_layer1");
	dump_screen_grap_layer(arg->input_layer[2], "input_layer2");
	dump_screen_grap_layer(arg->input_layer[3], "input_layer3");
}

static void dump_screen_grap_quit(screen_grap_quit *arg)
{
	printk_dbg2(1, "screen_grap_quit\n");
	printk_dbg2(1, "  handle:%p mode:%ld\n", arg->handle, arg->mode);
}

static void dump_screen_grap_delete(screen_grap_delete *arg)
{
	printk_dbg2(1, "screen_grap_delete\n");
	printk_dbg2(1, "  handle:%p\n", arg->handle);
}



#endif

static void work_runblend(struct work_struct *work)
{
	unsigned long _ptr = (unsigned long)work \
		- offsetof(struct composer_fh, fh_wqtask);
	struct composer_fh *fh = (struct composer_fh *) _ptr;
	int  rc = CMP_ERR_OK;

	static const int fh_status_translate[] = {
		[IDX_OUTPLANE]  = FH_STATUS_OUTPLANE,
		[IDX_MAINLAYER] = FH_STATUS_MAINLAYER,
		[IDX_CH1LAYER]  = FH_STATUS_CH1LAYER,
		[IDX_CH2LAYER]  = FH_STATUS_CH2LAYER,
		[IDX_CH3LAYER]  = FH_STATUS_CH3LAYER
	};

	void *handle;
	int  lc;

	fh->grap_data.status  = 0;
	handle = screen_graphics_new(&fh->grap_data._new);
	if (handle == NULL) {
		printk_err1("screen_graphics_new return error.\n");
		rc = CMP_ERR_NG;
		goto finish;
	}
	fh->grap_data.handle        = handle;
	fh->grap_data._ini.handle   = handle;
	fh->grap_data._blend.handle = handle;
	fh->grap_data._quit.handle  = handle;
	fh->grap_data._del.handle   = handle;

#if _LOG_DBG >= 2
	dump_screen_grap_initialize(&fh->grap_data._ini);
	dump_screen_grap_image_blend(&fh->grap_data._blend);
	dump_screen_grap_quit(&fh->grap_data._quit);
	dump_screen_grap_delete(&fh->grap_data._del);
#endif
	rc = screen_graphics_initialize(&fh->grap_data._ini);
	if (rc != SMAP_LIB_GRAPHICS_OK) {
		printk_err1("screen_graphics_initialize return by %d.\n", rc);
		rc = CMP_ERR_NG;
		goto finish2;
	}

	rc = screen_graphics_image_blend(&fh->grap_data._blend);
	if (rc != SMAP_LIB_GRAPHICS_OK) {
		printk_err1("screen_graphics_image_blend return by %d.\n", rc);
		rc = CMP_ERR_NG;
		goto finish3;
	}

	lc = 0;
	while (fh->grap_data.status == 0) {
		rc = wait_event_interruptible_timeout(
			fh->grap_data.wait_notify, \
			fh->grap_data.status != 0, msecs_to_jiffies(500));
		if (rc <= 0) {
			if (lc < 3 || (lc&15) == 15) {
				/* report error */
				printk_err1("waiting. callback:%d %d\n",
					rc, lc);
			}
			if (lc > 31 && kthread_should_stop()) {
				printk_err1("detect terminate request.\n");
				break;
			}
		}
		lc++;
	}

	if (fh->grap_data.status != 1) {
		printk_err1("callback result is error.\n");
		rc = CMP_ERR_NG;
	}

finish3:
	rc = screen_graphics_quit(&fh->grap_data._quit);
	if (rc != SMAP_LIB_GRAPHICS_OK) {
		printk_err1("screen_graphics_image_quit return by %d.\n", rc);
		rc = CMP_ERR_NG;
	}
finish2:
	screen_graphics_delete(&fh->grap_data._del);
finish:
	fh->grap_data.handle        = NULL;
	fh->grap_data._ini.handle   = NULL;
	fh->grap_data._blend.handle = NULL;
	fh->grap_data._quit.handle  = NULL;
	fh->grap_data._del.handle   = NULL;

	printk_dbg2(1, "work_runblend results rc:%d\n", rc);

#if _LOG_DBG >= 2
	dump_screen_grap_image_param(&fh->grap_data._blend.output_image,
		"output_image");
#endif

	/* image blending finished. */
	if (rc) {
		fh->fh_status |= FH_STATUS_BLENDERR;
	} else {
		/* no error. set flag to complete blending. */
		int i;
		for (i = 0; i < ARRAY_SIZE(fh_status_translate); i++) {
			struct composer_info   *info;
			struct composer_buffer *buf;

			if ((fh->fh_status & fh_status_translate[i]) == 0) {
				/* not open */
				continue;
			}
			info = &fh->fh_info[i];
			buf = fh->fh_info[i].buffer;

			/* check buffer configuration */
			if (info->id & INFO_ID_FLG_NOTDISP) {
				/* not blending plane */
				continue;
			}

			/* update status, to set poll blend */
			allfile_status_set(buf, FH_STATUS_POLLBLEND);
		}

		for (i = 0; i < ARRAY_SIZE(fh_status_translate); i++) {
			struct composer_info   *info;
			struct composer_buffer *buf;

			if ((fh->fh_status & fh_status_translate[i]) == 0) {
				/* not open */
				continue;
			}
			info = &fh->fh_info[i];
			buf = fh->fh_info[i].buffer;

			/* check buffer configuration */
			if (info->id & INFO_ID_FLG_NOTDISP) {
				/* not blending plane */
				continue;
			}

			/* wakeup all file handle used thisbufer. */
			wakeup_allfile(buf);
		}
	}
	fh->fh_status &= ~FH_STATUS_BLENDING;

	return;
}
static int  setbusy(struct composer_fh *fh)
{
	int rc, i;
	int lock_flag;
	int try_count;
	struct composer_buffer *buffer[5];
	struct composer_info   *info[5];
	int                    nBuffer;

	static const int fh_status_translate[] = {
		[IDX_OUTPLANE]  = FH_STATUS_OUTPLANE,
		[IDX_MAINLAYER] = FH_STATUS_MAINLAYER,
		[IDX_CH1LAYER]  = FH_STATUS_CH1LAYER,
		[IDX_CH2LAYER]  = FH_STATUS_CH2LAYER,
		[IDX_CH3LAYER]  = FH_STATUS_CH3LAYER
	};

	printk_dbg2(2, "fh->fh_status: 0x%x\n", fh->fh_status);

	memset(buffer, 0, sizeof(buffer));
	memset(info,   0, sizeof(info));

	/* pickup-buffer */
	nBuffer = 0;
	for (i = 0; i < ARRAY_SIZE(fh_status_translate); i++) {
		if ((fh->fh_status & fh_status_translate[i]) == 0) {
			/* not open */
			continue;
		}
		/* check buffer configuration */
		if (fh->fh_info[i].id & INFO_ID_FLG_NOTDISP) {
			/* not blending plane */
			continue;
		}

		info[nBuffer]   = &fh->fh_info[i];
		buffer[nBuffer] = fh->fh_info[i].buffer;

		BUG_ON(buffer[nBuffer] == NULL);
		nBuffer++;
	}
	printk_dbg2(2, "num of buffer to lock: %d\n", nBuffer);


	try_count = 0;
retry:
	if (try_count > 5) {
		rc = -EAGAIN;
		printk_err("retry count reach limit. give up setbusy");
		goto err_exit;
	}

	lock_flag = 0;
	rc = 0;
	/* try lock */
	for (i = 0; i < nBuffer; i++) {
		if (down_trylock(&buffer[i]->buf_sem)) {
			/* set return code */
			rc = -EAGAIN;
		} else {
			/* set flag, buffer locked. */
			lock_flag |= (1<<i);
		}
	}
	printk_dbg2(2, "lock result: %d\n", rc);

	if (rc != 0) {
		/* unlock buffer */
		for (i = 0; i < nBuffer; i++) {
			if (lock_flag & (1<<i)) {
				/* try lock success.*/
				up(&buffer[i]->buf_sem);
			}
		}
		/* wait event to enable unlock */
		for (i = 0; i < nBuffer; i++) {
			if (lock_flag & (1<<i)) {
				/* try lock success.*/
				continue;
			}

			do {
				rc = wait_event_interruptible_timeout(
					fh->fh_wait,
					buffer[i]->buf_sem.count > 0,
					1*HZ);
			} while (rc == 0);

			if (rc < 0) {
				printk_err2("buffer lock interrupted.\n");
				return rc;
			}
		}

		/* may be previous busy buffer free */
		try_count++;
		goto retry;
	}
	for (i = 0; i < nBuffer; i++) {
		/* set flag to this buffer use to blending. */
		info[i]->id |= INFO_ID_FLG_BLENDING;
	}
	printk_dbg2(2, "lock result: %d\n", rc);

#if _LOG_DBG >= 2
	if (2 < debug) {
		int semaphore_count[5];
		int buffer_id[5];

		memset(semaphore_count, 0, sizeof(semaphore_count));
		memset(buffer_id,      0, sizeof(buffer_id));

		for (i = 0; i < nBuffer; i++) {
			semaphore_count[i] = buffer[i]->buf_sem.count;
			buffer_id[i]       = info[i]->id;
		}
		printk_dbg(2, "lock_info: %d(0x%x) %d(0x%x) " \
			"%d(0x%x) %d(0x%x) %d(0x%x)\n",       \
			semaphore_count[0], buffer_id[0],     \
			semaphore_count[1], buffer_id[1],     \
			semaphore_count[2], buffer_id[2],     \
			semaphore_count[3], buffer_id[3],     \
			semaphore_count[4], buffer_id[4]);
	}
#endif
err_exit:
	return rc;
}
static int setidle(struct composer_fh *fh)
{
	int rc, i;

	static const int fh_status_translate[] = {
		[IDX_OUTPLANE]  = FH_STATUS_OUTPLANE,
		[IDX_MAINLAYER] = FH_STATUS_MAINLAYER,
		[IDX_CH1LAYER]  = FH_STATUS_CH1LAYER,
		[IDX_CH2LAYER]  = FH_STATUS_CH2LAYER,
		[IDX_CH3LAYER]  = FH_STATUS_CH3LAYER
	};

	printk_dbg2(2, "fh->fh_status: 0x%x\n", fh->fh_status);

	/* pickup-buffer */
	for (i = 0; i < ARRAY_SIZE(fh_status_translate); i++) {
		if ((fh->fh_status & fh_status_translate[i]) == 0) {
			/* not open */
			continue;
		}
		/* check buffer configuration */
		if (fh->fh_info[i].id & INFO_ID_FLG_NOTDISP) {
			/* not blending plane */
			continue;
		}

		if (fh->fh_info[i].id & INFO_ID_FLG_BLENDING) {
			fh->fh_info[i].id &= ~INFO_ID_FLG_BLENDING;
			unlock_buffer(fh, i);
		}
	}
	/* pass */
	rc = 0;
	return rc;
}


/******************************************************/
/* file operation entry function                      */
/******************************************************/
static ssize_t core_read(struct file *filp, char __user *buf, \
		size_t sz, loff_t *off)
{
	int rc = -EIO;
#if INTERNAL_DEBUG >= 1
	rc = -EINVAL; /* not implemented. */
#endif
	return rc;
}

static unsigned int core_poll(struct file *filep, \
		struct poll_table_struct *poll)
{
	int rc;
	struct composer_fh     *fh;

	fh = (struct composer_fh *)filep->private_data;

	printk_dbg2(2, "down");
	down(&fh->fh_sem);

	printk_dbg2(2, "current file:%p fh_status:0x%x\n", fh, fh->fh_status);

	if (!(fh->fh_status & (FH_STATUS_POLLBLEND | FH_STATUS_POLLADDR))) {
		/* wait event */
		poll_wait(filep, &fh->fh_wait, poll);
	}

	rc = 0;
	if (fh->fh_status & FH_STATUS_POLLBLEND) {
		rc |= POLLIN | POLLRDNORM;
		fh->fh_status &= ~FH_STATUS_POLLBLEND;
	}
	if (fh->fh_status & FH_STATUS_POLLADDR) {
		rc |= POLLOUT | POLLRDNORM;
		fh->fh_status &= ~FH_STATUS_POLLADDR;
	}
	printk_dbg2(2, "poll result file:%p fh_status:0x%x rc:0x%x\n",
		fh, fh->fh_status, rc);

	up(&fh->fh_sem);
	return rc;
}

static int core_ioctl(struct inode *inode, struct file *filep, \
		unsigned int cmd, unsigned long arg)
{
	int rc  = -EINVAL;
	int dir = _IOC_DIR(cmd);
	int sz  = _IOC_SIZE(cmd);
	struct composer_fh *fh;
	static unsigned long _arg_area[64];
	void   *parg;

/********************/
/* ProLoge of IOCTL */
/********************/
	if (sz != 0 && ((dir & (_IOC_WRITE|_IOC_READ)) != 0)) {
		if (sz >= sizeof(_arg_area)) {
			printk_err2("ioctl argument size too large\n");
			goto err_exit;
		}
	}

	if (sz != 0 && (dir & _IOC_WRITE) != 0) {
		parg = &_arg_area[0];
		printk_dbg2(2, "copy_from_user\n");
		if (copy_from_user(parg, (void __user *)arg, sz)) {
			printk_err2("fail in copy_from_user\n");
			goto err_exit;
		}
	}

	fh = (struct composer_fh *)filep->private_data;
	parg = &_arg_area[0];

	printk_dbg2(2, "down");
	down(&fh->fh_sem);

	switch (cmd) {
	case CMP_IOC_ISLAYEX:
		rc = ioc_islayex(fh);
		break;
	case CMP_IOCG_SUPPORTPIXFMT:
		rc = iocg_suppotpixfmt(fh, parg);
		break;
	case CMP_IOCS_OPEN:
		rc = iocs_open(fh, parg);
		break;
	case CMP_IOCS_CLOSE:
		rc = iocs_close(fh, parg);
		break;
	case CMP_IOCS_DISP:
		rc = iocs_disp(fh, parg);
		break;
	case CMP_IOCS_ALPHA:
		rc = iocs_alpha(fh, parg);
		break;
	case CMP_IOCG_ALPHA:
		rc = iocg_alpha(fh, parg);
		break;
	case CMP_IOCS_KEYCOLOR:
		rc = iocs_keycolor(fh, parg);
		break;
	case CMP_IOCG_KEYCOLOR:
		rc = iocg_keycolor(fh, parg);
		break;
	case CMP_IOCS_LAYSIZE:
		rc = iocs_laysize(fh, parg);
		break;
	case CMP_IOCG_LAYSIZE:
		rc = iocg_laysize(fh, parg);
		break;
	case CMP_IOCS_PIXFMT:
		rc = iocs_pixfmt(fh, parg);
		break;
	case CMP_IOCG_PIXFMT:
		rc = iocg_pixfmt(fh, parg);
		break;
	case CMP_IOCS_CROPSIZE:
		rc = iocs_cropsize(fh, parg);
		break;
	case CMP_IOCG_CROPSIZE:
		rc = iocg_cropsize(fh, parg);
		break;
	case CMP_IOCS_CROPPOS:
		rc = iocs_croppos(fh, parg);
		break;
	case CMP_IOCG_CROPPOS:
		rc = iocg_croppos(fh, parg);
		break;
	case CMP_IOCS_POS:
		rc = iocs_pos(fh, parg);
		break;
	case CMP_IOCG_POS:
		rc = iocg_pos(fh, parg);
		break;
	case CMP_IOCS_SIZE:
		rc = iocs_size(fh, parg);
		break;
	case CMP_IOCG_SIZE:
		rc = iocg_size(fh, parg);
		break;
	case CMP_IOCS_VIEWLAY:
		rc = iocs_viewlay(fh, parg);
		break;
	case CMP_IOCS_BACKCOLOR:
		rc = iocs_backcolor(fh, parg);
		break;
	case CMP_IOCG_BACKCOLOR:
		rc = iocg_backcolor(fh, parg);
		break;
	case CMP_IOCS_LAYADDR:
		rc = iocs_layaddr(fh, parg);
		break;
	case CMP_IOCG_LAYADDR:
		rc = iocg_layaddr(fh, parg);
		break;
	case CMP_IOC_START:
		rc = ioc_start(fh);
		break;
	default:
		printk_err2("invalid cmd 0x%x\n", cmd);
	}

	up(&fh->fh_sem);

/********************/
/* EpiLoge of IOCTL */
/********************/
	if (rc == 0 && sz != 0 && (dir & _IOC_READ) != 0) {
		parg = &_arg_area[0];
		printk_dbg2(2, "copy_to_user\n");
		if (copy_to_user((void __user *)arg, parg, sz)) {
			printk_err2("fail in copy_to_user\n");
			rc = -EINVAL;
		}
	}
err_exit:

	return rc;
}

static int core_open(struct inode *inode, struct file *filep)
{
	int rc = 0;
	unsigned long flags;
	struct composer_fh *private_fh;

	printk_dbg2(2, "spinlock");
	spin_lock_irqsave(&irqlock, flags);
	if (num_open >= MAX_OPEN) {
		/* set return code */
		rc = -ENODEV;
	} else {
		/* increment open count */
		num_open++;
	}
	spin_unlock_irqrestore(&irqlock, flags);

	if (rc) {
		/* return by error */
		goto err_exit;
	}

	/* allocate per-filehandle data */
	private_fh = allocate_device();
	if (private_fh == NULL) {
		rc = -ENOMEM;

		spin_lock_irqsave(&irqlock, flags);
		num_open--;
		spin_unlock_irqrestore(&irqlock, flags);
	} else {
		filep->private_data = private_fh;
	}
err_exit:
	return rc;
}

static int core_release(struct inode *inode, struct file *filep)
{
	struct composer_fh *fh;
	unsigned long flags;

	fh = (struct composer_fh *)filep->private_data;

	free_device(fh);

	printk_dbg2(2, "spinlock");
	spin_lock_irqsave(&irqlock, flags);
	num_open--;
	spin_unlock_irqrestore(&irqlock, flags);

	return 0;
}

/******************************************************/
/* module initialize function                         */
/******************************************************/
static int __init sh_mobile_composer_init(void)
{
	int ret = 0;
	int i;

	/* initialize globals */
	if (debug >= 3) {
		/* set app share library. to debug mode */
		sh_mobile_appmem_debugmode(1);
	}

	memset(&local_buffer, 0, sizeof(local_buffer));
	spin_lock_init(&irqlock);
	sema_init(&sem, 1);
	num_open = 0;
	INIT_LIST_HEAD(&file_top);
	INIT_LIST_HEAD(&appshare_top);

	for (i = 0; i < MAX_BUFFER; i++) {
		struct composer_buffer   *fh_buffer = &local_buffer[i];
		initialize_bufferinformation(fh_buffer);
	}

	/* create workqueue */
	workqueue = create_singlethread_workqueue("composer");
	if (workqueue == NULL) {
		printk_err("fail to create_singlethread_workqueue");
		ret = -ENOMEM;
		goto err_exit;
	}

	/* regist device */
	ret = misc_register(&composer_device);
	if (ret) {
		printk_err("fail to misc_register (MISC_DYNAMIC_MINOR)\n");
		return ret;
	}
	return 0;

err_exit:
	if (workqueue) {
		destroy_workqueue(workqueue);
		workqueue = NULL;
	}
	return ret;
}

static void __exit sh_mobile_composer_release(void)
{
	if (num_open > 0)
		printk_err("there is 'not close device'.\n");

	/* unregist device */
	misc_deregister(&composer_device);

	/* release all resources */
	/* destroy workqueue */
	if (workqueue) {
		destroy_workqueue(workqueue);
		workqueue = NULL;
	}

	return;
}

module_param(debug, int, 0644);
MODULE_PARM_DESC(debug, "SHMobile Composer debug level");
MODULE_LICENSE("GPL");
module_init(sh_mobile_composer_init);
module_exit(sh_mobile_composer_release);
