/*
 * Function        : helper functions for rt-api shared buffers
 *
 * Based on videobuf-dma-contig.c,
 * Copyright (C) 2011 Renesas Electronics Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 */

/*
 * helper functions for physically contiguous capture buffers
 *
 * The functions support hardware lacking scatter gather support
 * (i.e. the buffers must be linear in physical memory)
 *
 * Copyright (c) 2008 Magnus Damm
 *
 * Based on videobuf-vmalloc.c,
 * (c) 2007 Mauro Carvalho Chehab, <mchehab@infradead.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2
 */

#include <linux/videodev2.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/mm.h>
#include <linux/pagemap.h>
#include <linux/dma-mapping.h>
#include <linux/semaphore.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <media/videobuf-dma-contig.h>
#include <linux/delay.h>

#define _LCK_DBG   0
#define _MEM_DBG   0 /* debug memory */         /* 0: disable  1: enable */
#define _OVL_DBG   0 /* debug overlay */        /* 0: disable  1: enable */
#define _ERR_DBG   0
#define _ERR_RTAPI 1


#define _NOTIFY_FB       0 /* notify Overlay state to FB */
#define _UPDATE_FBPARAM  0 /* set choromakey and alpha   */

#include "sh_mobile_overlay.h"
#include <video/sh_mobile_lcdc.h>

#undef  DEV_NAME
#define DEV_NAME "v4l2buf-mfi"
#if RTAPI_MEMORY_INTERFACE
#include <media/sh_mobile_appmem.h>
#include <rtapi/system_memory.h>
#include <rtapi/screen_display.h>
#include <rtapi/screen_overlay.h>
#endif

/* place holder to handle debug interface */
#define TRACE_CLR
#define TRACE_ENTER(id)
#define TRACE_LEAVE(id)
#define TRACE_LOG(id)
#define TRACE_LOG1(id, arg)

#define printk_err(fmt, arg...) \
	do { \
		printk(KERN_ERR DEV_NAME ": %s: " fmt, __func__, ## arg); \
	} while (0)

#define printk_dbg(level, fmt, arg...) \
	do { \
		if (level > 0) \
			printk(KERN_INFO DEV_NAME ": %s: " fmt, \
				__func__, ## arg); \
	} while (0)

#define dbg_lock(lock, fmt, arg...) \
	do { \
		if (_LCK_DBG > 0) \
			if (spin_is_locked(lock)) \
				printk(KERN_DEBUG DEV_NAME ": %s: " fmt, \
				 __func__, ## arg); \
	} while (0)

struct videobuf_mfi_memory {
	u32 magic;
	void *vaddr;
	dma_addr_t dma_handle;
	unsigned long size;
	int is_userptr;
#if RTAPI_MEMORY_INTERFACE
	unsigned long offset;
	void     *apmem_handle;
#endif
};

#define MAGIC_DC_MEM 0x0733ac61
#define MAGIC_CHECK(is, mag)						    \
	if (unlikely((is) != (mag)))	{				    \
		printk_err("magic mismatch: %x expected %x\n", (is), (mag));\
		BUG();							    \
	}


/* #include "debug_trace.h" */

#if RTAPI_MEMORY_INTERFACE

static DECLARE_MUTEX(sem_lock);
static wait_queue_head_t    wait_id[2];
static int                  ovl_state[2];
static int                  result_code[2];
static int                  ovl_outmode[2];

#define OVL_STATE_READY    0x0001
#define OVL_STATE_FRMREADY 0x0002
#define OVL_STATE_ERROR    0x0100
#define OVL_STATE_INIT     0x0200
#define OVL_STATE_FRAMEBUF 0x0400
#define OVL_STATE_FRAMEERR 0x0800

static void function_notify_overlay(
	int result, screen_ovl_notify_return_data *notify_data, int data_size)
{
	TRACE_ENTER(FUNC_OVL_CALLBACK);
	printk_dbg(_OVL_DBG,     \
		"function_notify_overlay:%d\n", notify_data->command_type);
	switch (notify_data->command_type) {
	case RT_OVERLAY_NOTIFY_RETURN:
		TRACE_LOG(FUNC_OVL_CALLBACK);
		if (notify_data->overlay_id == RT_OVERLAY_ID1) {
			result_code[0] = result;
			ovl_state[0]  |= OVL_STATE_READY;
			wake_up_interruptible(&wait_id[0]);
		} else if (notify_data->overlay_id == RT_OVERLAY_ID2) {
			result_code[1] = result;
			ovl_state[1]  |= OVL_STATE_READY;
			wake_up_interruptible(&wait_id[1]);
		} else if (notify_data->overlay_id == RT_OVERLAY_ID1A) {
			result_code[0] = result;
			ovl_state[0]  |= OVL_STATE_FRMREADY;
			wake_up_interruptible(&wait_id[0]);
		} else if (notify_data->overlay_id == RT_OVERLAY_ID2A) {
			result_code[1] = result;
			ovl_state[1]  |= OVL_STATE_FRMREADY;
			wake_up_interruptible(&wait_id[1]);
		}
		break;
	case RT_OVERLAY_NOTIFY_RETURN_BUFFER:
		TRACE_LOG(FUNC_OVL_CALLBACK);
		if (notify_data->overlay_id == RT_OVERLAY_ID1 ||
			notify_data->overlay_id == RT_OVERLAY_ID2) {
			struct sh_mobile_v4l2_videobuffer *emvb = \
				(struct sh_mobile_v4l2_videobuffer *) \
				notify_data->return_buffer.user_data;
			struct videobuf_queue *vbq     = emvb->vbq;
			struct sh_mobile_v4l2_device *dev = vbq->priv_data;

			if (dev->mDev.callback) {
				/* generate callback */
				dev->mDev.callback(emvb);
			}
		}
		break;
	case RT_OVERLAY_NOTIFY_ERROR_INFO:
		TRACE_LOG(FUNC_OVL_CALLBACK);
		if (notify_data->overlay_id == RT_OVERLAY_ID1 ||
			notify_data->overlay_id == RT_OVERLAY_ID1A) {
			result_code[0] = result;
			ovl_state[0]  |= OVL_STATE_ERROR;
			wake_up_interruptible(&wait_id[0]);
		} else if (notify_data->overlay_id == RT_OVERLAY_ID2 ||
			notify_data->overlay_id == RT_OVERLAY_ID2A) {
			result_code[1] = result;
			ovl_state[1]  |= OVL_STATE_ERROR;
			wake_up_interruptible(&wait_id[1]);
		}
		break;
	}
	TRACE_LEAVE(FUNC_OVL_CALLBACK);
}
#endif


static int check_available_usermem(struct sh_mobile_v4l2_device *dev)
{
	if (dev == NULL) {
		printk_err("invalid argument.\n");
		return 0;
	}
	if (dev->workbuf.memory == V4L2_WORKBUF_MEMORY_PHYSICAL) {
		/* physical memory */
		return 1;
	} else if (dev->workbuf.memory == V4L2_WORKBUF_MEMORY_APPSHARE) {
		/* app share memory */
		if (dev->workbuf.apmem_addr && dev->workbuf.apmem_id) {
			/* memory useable */
			return 1;
		}
	}
	printk_dbg(_ERR_DBG, "private buffer not registered.\n");
	return 0;
};
#if RTAPI_MEMORY_INTERFACE
static int check_available_overlay(struct sh_mobile_v4l2_device *dev)
{
	if (!check_available_usermem(dev))
		return 0;
	if (dev->mDev.handle_overlay == NULL) {
		printk_dbg(_ERR_DBG, "Overlay not available.\n");
		return 0;
	}
	/* overlay useable */
	return 1;
};
#endif

/**
 * videobuf_mfi_user_put() - reset pointer to user space buffer
 * @mem: per-buffer private videobuf-dma-contig data
 *
 * This function resets the user space pointer
 */
static void videobuf_mfi_user_put(struct sh_mobile_v4l2_device *dev,
    struct videobuf_mfi_memory *mem)
{
#if RTAPI_MEMORY_INTERFACE
	if (mem->is_userptr) {
		unsigned long flags;
		spin_lock_irqsave(&dev->vbq_lock, flags);
		dev->mMem.mRefCount--;
		spin_unlock_irqrestore(&dev->vbq_lock, flags);
		printk_dbg(_MEM_DBG,                                    \
			"memory reference counts:%d\n", dev->mMem.mRefCount);
		mem->vaddr = NULL;
	}
	mem->offset       = 0;
	mem->apmem_handle = NULL;
#endif
	mem->is_userptr = 0;
	mem->dma_handle = 0;
	mem->size = 0;
}

/**
 * videobuf_mfi_user_get() - setup user space memory pointer
 * @mem: per-buffer private videobuf-dma-contig data
 * @vb: video buffer to map
 *
 * This function validates and sets up a pointer to user space memory.
 * Only physically contiguous pfn-mapped memory is accepted.
 *
 * Returns 0 if successful.
 */
static int videobuf_mfi_user_get(struct sh_mobile_v4l2_device *dev,
	struct videobuf_mfi_memory *mem, struct videobuf_buffer *vb)
{
	int ret = -EINVAL;
#if RTAPI_MEMORY_INTERFACE
	unsigned long user_address = vb->baddr;
	unsigned long flags;
	void *handle = NULL;

	if (dev->workbuf.memory == V4L2_WORKBUF_MEMORY_PHYSICAL) {
		/* physical memory */
		mem->size         =  vb->bsize;
		mem->dma_handle   =  0;
		mem->is_userptr   =  1;
		mem->offset       =  0;
		mem->apmem_handle =  NULL;
		mem->vaddr        =  (void *)user_address;
	} else {
		/* app share memory */
		system_mem_ap_share_mem  smem;
		unsigned long offset;
		unsigned long endoffset;
		int rtapi_ret;
		void *apmem_handle;

		/* calculate offset */
		if (user_address < dev->workbuf.apmem_addr) {
			printk_dbg(_ERR_DBG, "can not access user memory.\n");
			goto error;
		}

		offset    = user_address - dev->workbuf.apmem_addr;
		endoffset = offset + vb->bsize - 1;

		/* get handle */
		handle = system_memory_info_new();
		apmem_handle = sh_mobile_appmem_getmemoryhandle(  \
			dev->mMem.appmem_handle);

		/* confirm valid range (allocatable?) */
		printk_dbg(_MEM_DBG,                                     \
			"handle:%p mem handle:%p offset:%lx\n",          \
			handle, apmem_handle, endoffset);

		smem.handle       = handle;
		smem.apmem_handle = apmem_handle;
		smem.apmem_offset = endoffset;
		smem.apmem_apaddr = 0;
		rtapi_ret = system_memory_ap_share_mem(&smem);
		if (rtapi_ret != 0) {
			printk_dbg(_ERR_DBG, "end offset not accessible.\n");
			goto error;
		}

		/* mapping shared memory */
		printk_dbg(_MEM_DBG,                                     \
			"handle:%p mem handle:%p offset:%lx\n",              \
			handle, apmem_handle, offset);

		smem.handle       = handle;
		smem.apmem_handle = apmem_handle;
		smem.apmem_offset = offset;
		smem.apmem_apaddr = 0;
		rtapi_ret = system_memory_ap_share_mem(&smem);
		if (rtapi_ret != 0) {
			printk_dbg(_ERR_DBG, "can not access user memory.\n");
			goto error;
		}

		printk_dbg(_MEM_DBG, "memory allocation success:%x\n", \
			smem.apmem_apaddr);
		mem->size         =  vb->bsize;
		mem->dma_handle   =  0;
		mem->is_userptr   =  1;
		mem->offset       =  offset;
		mem->apmem_handle =  apmem_handle;
		mem->vaddr        =  (void *)smem.apmem_apaddr;
	}

	printk_dbg(_MEM_DBG, "i:%d   adr:0x%p  offset:0x%lx\n", \
		vb->i, mem->vaddr, mem->offset);

	ret = 0;
	dbg_lock(&dev->vbq_lock, "vbq_lock");
	spin_lock_irqsave(&dev->vbq_lock, flags);
	dev->mMem.mRefCount++;
	spin_unlock_irqrestore(&dev->vbq_lock, flags);
	printk_dbg(_MEM_DBG, "memory reference counts:%d\n",    \
	    dev->mMem.mRefCount);

error:
	if (handle != NULL) {
		system_mem_info_delete  del;
		del.handle = handle;
		system_memory_info_delete(&del);
		handle = NULL;
	}
#endif
	return ret;
}

static struct videobuf_buffer *__videobuf_alloc(size_t size)
{
	struct videobuf_mfi_memory *mem;
	struct videobuf_buffer *vb;

	vb = kzalloc(size + sizeof(*mem), GFP_KERNEL);
	if (vb) {
		mem = vb->priv = ((char *)vb) + size;
		mem->magic = MAGIC_DC_MEM;
	}

	return vb;
}

static void *__videobuf_to_vaddr(struct videobuf_buffer *buf)
{
	struct videobuf_mfi_memory *mem = buf->priv;

	BUG_ON(!mem);
	MAGIC_CHECK(mem->magic, MAGIC_DC_MEM);
	return mem->vaddr;
}

static int __videobuf_iolock(struct videobuf_queue *q,
			     struct videobuf_buffer *vb,
			     struct v4l2_framebuffer *fbuf)
{
	struct videobuf_mfi_memory *mem = vb->priv;

	BUG_ON(!mem);
	MAGIC_CHECK(mem->magic, MAGIC_DC_MEM);

	switch (vb->memory) {
	case V4L2_MEMORY_MMAP:
		printk_dbg(_MEM_DBG, "%s memory method MMAP\n", __func__);
		return -EINVAL;
		break;
	case V4L2_MEMORY_USERPTR:
		printk_dbg(_MEM_DBG, "%s memory method USERPTR\n", __func__);

		/* handle pointer from user space */
		if (vb->baddr) {
			if (check_available_usermem(q->priv_data)) {
				return videobuf_mfi_user_get(           \
					q->priv_data, mem, vb);
			} else {
				printk_dbg(_ERR_DBG,  \
					"user memory is not useable.\n");
				return -EINVAL;
			}
		}

		/* not support */
		printk_dbg(_ERR_DBG, "0 not supported.\n");
		return -ENOMEM;

	case V4L2_MEMORY_OVERLAY:
	default:
		printk_dbg(_MEM_DBG, "%s memory method OVERLAY/unknown\n",\
			__func__);
		return -EINVAL;
	}

	return 0;
}

static int __videobuf_mmap_mapper(struct videobuf_queue *q,
				  struct videobuf_buffer *buf,
				  struct vm_area_struct *vma)
{
	return -ENOMEM;
}

static struct videobuf_qtype_ops qops = {
	.magic        = MAGIC_QTYPE_OPS,

	.alloc        = __videobuf_alloc,
	.iolock       = __videobuf_iolock,
	.mmap_mapper  = __videobuf_mmap_mapper,
	.vaddr        = __videobuf_to_vaddr,
};

/********************/
/* Memory Interface */
/********************/
void videobuf_queue_sh_mobile_interface_init(struct sh_mobile_v4l2_device *dev)
{
/*allocate interface*/
#if RTAPI_MEMORY_INTERFACE
	printk_dbg(_MEM_DBG, "create memory interface\n");
	dev->mMem.mRefCount = 0;
	dev->mMem.type      = 0;

	dev->mDev.handle_overlay  = NULL;
	dev->mDev.overlay_id  = 0;
	dev->mDev.overlay_id2 = 0;
	dev->mDev.output_mode = 0;
#endif
}

void videobuf_queue_sh_mobile_interface_release( \
	struct sh_mobile_v4l2_device *dev)
{
/*release interface*/
#if RTAPI_MEMORY_INTERFACE

	if (dev->mDev.handle_overlay) {
		screen_ovl_delete del;
		printk_dbg(_OVL_DBG, "free display\n");
		del.handle = dev->mDev.handle_overlay;
		screen_overlay_delete(&del);
		dev->mDev.handle_overlay = NULL;

	}

	if (dev->mMem.appmem_handle) {
		printk_dbg(_MEM_DBG, "free memory\n");

		if (sh_mobile_appmem_free(dev->mMem.appmem_handle)) {
			/* output error message */
			printk_err("can not free memory\n");
		}
		dev->mMem.appmem_handle = NULL;
	}
#endif
}

int  videobuf_queue_sh_mobile_interface_config( \
	struct sh_mobile_v4l2_device *dev)
{
#if !RTAPI_MEMORY_INTERFACE
	return -EINVAL;
#else
	unsigned long flags;
	struct appmem_handle *mMemory;
	int  Count;
	int  ret = 0;

	spin_lock_irqsave(&dev->vbq_lock, flags);
	mMemory = dev->mMem.appmem_handle;
	Count =   dev->mMem.mRefCount;
	spin_unlock_irqrestore(&dev->vbq_lock, flags);

	/* check now using */
	if (Count) {
		printk_dbg(_MEM_DBG, "memory %p now using count:%d\n",\
			mMemory, Count);
		return -EBUSY;
	}

	/* free memory */
	if (mMemory) {
		if (sh_mobile_appmem_free(mMemory)) {
			/* output error message */
			printk_err("can not free memory\n");
		}
		mMemory = NULL;
		dev->mMem.appmem_handle = NULL;
	}
	dev->mMem.type = 0;

	if (dev->workbuf.memory == V4L2_WORKBUF_MEMORY_PHYSICAL) {
		/* physical memory */
		dev->mMem.type = 1;
		printk_dbg(_MEM_DBG, "memory type physical\n");
	} else if (dev->workbuf.memory == V4L2_WORKBUF_MEMORY_APPSHARE &&
		dev->workbuf.apmem_addr) {
		/* next allocate memory */

		printk_dbg(_MEM_DBG, "id:%d\n", dev->workbuf.apmem_id);

		mMemory = sh_mobile_appmem_share( \
			dev->workbuf.apmem_id, "v4l2 overlay");

		if (mMemory == NULL) {
			printk_err("can not allocate memory\n");
			ret = -EINVAL;
		} else {
			dev->mMem.appmem_handle = mMemory;
			dev->mMem.type = 2;
			printk_dbg(_MEM_DBG, "memory type app share\n");
		}
	}

	return ret;
#endif
}

#if _ERR_RTAPI || _OVL_DBG
static void debuglog_screen_ovl_initialize(
	int type, int rc, screen_ovl_initialize * param
) {
	if (type == 1) {
		/* log return code. */
		printk_dbg(1, "screen_overlay_initialize error: %d\n", rc);
	}

	printk_dbg(1, \
		" handle: %p " \
		"overlay_id:%d output_mode:%d " \
		"image_width:%d image_height:%d " \
		"image_stride:%d image_format:%d " \
		"yuv_format:%d yuv_range:%d "
		"\n",
		param->handle,
		param->overlay_id,   param->output_mode,
		param->image_width,  param->image_height,
		param->image_stride, param->image_format,
		param->yuv_format,   param->yuv_range);
}

static void debuglog_screen_ovl_set_param(
	int type, int rc, screen_ovl_set_param * param
) {
	if (type == 1) {
		/* log return code. */
		printk_dbg(1, "screen_overlay_set_parameters " \
			"error: %d\n", rc);
	}

	printk_dbg(1, \
		" handle: %p overlay_id:%d " \
		" src_crop(x:%d y:%d width:%d height:%d) " \
		" dst(width:%d height:%d) " \
		"rotate:%d mirror %d " \
		"blend_pos(x:%d y:%d) alpha:%d\n",
		param->handle,      param->overlay_id,
		param->src_crop.x,  param->src_crop.y,
		param->src_crop.width,  param->src_crop.height,
		param->dst_width,       param->dst_height,
		param->rotate,      param->mirror,
		param->blend_pos_x, param->blend_pos_y,
		param->alpha);
}

#endif

#if RTAPI_MEMORY_INTERFACE
static int create_ovl_initialize(screen_ovl_initialize *param, \
	struct v4l2_pix_format  *pix)
{
	int min_stride[3];
	int ret = 0;
	int yuv_flag = 0;

	param->image_width  = pix->width;
	param->image_height = pix->height;
	min_stride[0] = min_stride[1] = min_stride[2] = 0;

	switch (pix->pixelformat) {
	case V4L2_PIX_FMT_YUV420:
		param->image_format = RT_OVERLAY_COLOR_YUV420_PLANAR;
		min_stride[0] = pix->width;
		min_stride[1] = pix->width / 2;
		min_stride[2] = pix->width / 2;
		yuv_flag = 1;
		break;
	case V4L2_PIX_FMT_NV12:
		param->image_format = RT_OVERLAY_COLOR_YUV420_SEMIPLANAR;
		min_stride[0] = pix->width;
		min_stride[1] = pix->width;
		yuv_flag = 1;
		break;
	case V4L2_PIX_FMT_NV16:
		param->image_format = RT_OVERLAY_COLOR_YUV422_SEMIPLANAR;
		min_stride[0] = pix->width;
		min_stride[1] = pix->width;
		yuv_flag = 1;
		break;
	case V4L2_PIX_FMT_UYVY:
		param->image_format = RT_OVERLAY_COLOR_YUV422_PACKED;
		min_stride[0] = pix->width * 2;
		yuv_flag = 1;
		break;
	case V4L2_PIX_FMT_RGB565:
		param->image_format = RT_OVERLAY_COLOR_RGB565;
		min_stride[0] = pix->width * 2;
		break;
	case V4L2_PIX_FMT_RGB24:
		param->image_format = RT_OVERLAY_COLOR_RGB888;
		min_stride[0] = pix->width * 3;
		break;
	case V4L2_PIX_FMT_RGB32:
		param->image_format = RT_OVERLAY_COLOR_ARGB8888;
		min_stride[0] = pix->width * 4;
		break;
	default:
		printk_dbg(_ERR_DBG, "invalid pixelformat.\n");
		ret = -EINVAL;
		goto err_exit;
	}
	param->image_stride = min_stride[0];

	if (yuv_flag) {
		/* YUV format */
		if (param->overlay_id != RT_OVERLAY_ID1 &&
			param->overlay_id != RT_OVERLAY_ID2) {
			printk_dbg(_ERR_DBG, "invalid pixelformat.\n");
			ret = -EINVAL;
			goto err_exit;
		}
		switch (pix->colorspace) {
		case V4L2_COLORSPACE_BT709_COMPRESS:
			param->yuv_format  = RT_OVERLAY_COLOR_BT709;
			param->yuv_range   = RT_OVERLAY_COLOR_COMPRESSED;
			break;
		case V4L2_COLORSPACE_BT709_FULLSCALE:
			param->yuv_format  = RT_OVERLAY_COLOR_BT709;
			param->yuv_range   = RT_OVERLAY_COLOR_FULLSCALL;
			break;
		case V4L2_COLORSPACE_BT601_COMPRESS:
			param->yuv_format  = RT_OVERLAY_COLOR_BT601;
			param->yuv_range   = RT_OVERLAY_COLOR_COMPRESSED;
			break;
		case V4L2_COLORSPACE_BT601_FULLSCALE:
			param->yuv_format  = RT_OVERLAY_COLOR_BT601;
			param->yuv_range   = RT_OVERLAY_COLOR_FULLSCALL;
			break;
		default:
			printk_dbg(_ERR_DBG, "invalid colorspace.\n");
			ret = -EINVAL;
			goto err_exit;
		}
	} else {
		int width;
		/* RGB format */
		if (param->overlay_id != RT_OVERLAY_ID1A &&
			param->overlay_id != RT_OVERLAY_ID2A) {
			printk_dbg(_ERR_DBG, "invalid pixelformat.\n");
			ret = -EINVAL;
			goto err_exit;
		}
		if (pix->bytesperline != 0) {
			switch (pix->pixelformat) {
			case V4L2_PIX_FMT_RGB565:
				width = pix->bytesperline / 2;
				break;
			case V4L2_PIX_FMT_RGB24:
				width = pix->bytesperline / 3;
				break;
			case V4L2_PIX_FMT_RGB32:
				width = pix->bytesperline / 4;
				break;
			default:
				width = param->image_width;
				break;
			}
			param->image_width = width;
		}
		param->yuv_format  = RT_OVERLAY_COLOR_BT601;
		param->yuv_range   = RT_OVERLAY_COLOR_COMPRESSED;
	}
err_exit:
	return ret;
}

static int create_ovl_parameter(screen_ovl_set_param *param, \
	struct v4l2_crop        *crop,                       \
	struct v4l2_window      *win,                        \
	int                     rotate,                      \
	int                     mirror)
{
	param->src_crop.x  = crop->c.left;
	param->src_crop.y  = crop->c.top;
	param->src_crop.width  = crop->c.width;
	param->src_crop.height = crop->c.height;
	switch (rotate) {
	default:
		printk_dbg(_ERR_DBG, "invalid rotate parameter.\n");
		return -EINVAL;
	case 0:
	case 2:
		param->dst_width   = win->w.width;
		param->dst_height  = win->w.height;
		break;
	case 1:
	case 3:
		param->dst_width   = win->w.height;
		param->dst_height  = win->w.width;
		break;
	}
	switch (rotate) {
	default:
		printk_dbg(_ERR_DBG, "invalid rotate parameter.\n");
		return -EINVAL;
	case 0:
		param->rotate  = RT_OVERLAY_ROTATE_0;
		break;
	case 1:
		param->rotate  = RT_OVERLAY_ROTATE_90;
		break;
	case 2:
		param->rotate  = RT_OVERLAY_ROTATE_180;
		break;
	case 3:
		param->rotate  = RT_OVERLAY_ROTATE_270;
		break;
	}
	switch (mirror) {
	default:
		printk_dbg(_ERR_DBG, "invalid mirror parameter.\n");
		return -EINVAL;
	case V4L2_MIRROR_NONE:
		param->mirror  = RT_OVERLAY_MIRROR_N;  /* no mirror */
		break;
	case V4L2_MIRROR_HFLIP:
		param->mirror  = RT_OVERLAY_MIRROR_H;  /* horizontal mirror */
		break;
	case V4L2_MIRROR_VFLIP:
		param->mirror  = RT_OVERLAY_MIRROR_V;  /* vertical mirror */
		break;
	}
	param->blend_pos_x      = win->w.left;
	param->blend_pos_y      = win->w.top;
	param->alpha            = 255;
	param->background_color = 0x0; /* constant */
	param->key_color        = RT_OVERLAY_KEY_COLOR_OFF;
	return 0;
}
#endif

int  videobuf_queue_sh_mobile_interface_createoverlay(              \
	struct sh_mobile_v4l2_device *dev, int type, void (*callback)(void *))
{
#if !RTAPI_MEMORY_INTERFACE
	return -EINVAL;
#else
	void *handle = NULL;
	int  ret = 0;
	int  output_mode = -1, overlay_id, overlay_id2, index = 0;
	TRACE_ENTER(FUNC_OVL_CREATE);

	if (dev->mDev.handle_overlay != NULL) {
		/* already create overlay */
		printk_dbg(_ERR_DBG, "already overlay created.\n");
		TRACE_LEAVE(FUNC_OVL_CREATE);
		return -EBUSY;
	}

	switch (type) {
	case 1:
		overlay_id = RT_OVERLAY_ID1;
		overlay_id2 = RT_OVERLAY_ID1A;
		index      = 0;
		break;
	case 2:
		overlay_id = RT_OVERLAY_ID2;
		overlay_id2 = RT_OVERLAY_ID2A;
		index      = 1;
		break;
	default:
		printk_dbg(_ERR_DBG, "type type.\n");
		ret = -EINVAL;
		goto err_exit;
	}

	if ((ovl_state[index] & OVL_STATE_INIT) == 0) {
		init_waitqueue_head(&wait_id[index]);
		ovl_state[index]   = OVL_STATE_INIT;
		result_code[index] = 0;
		ovl_outmode[index] = 0;
	}

	switch (dev->output) {
	case V4L2_OUTPUT_LCD1:
		output_mode = RT_OVERLAY_LCD1;
		break;
	case V4L2_OUTPUT_OUTSIDE:
		output_mode = RT_OVERLAY_SDENC;
		break;
	default:
		printk_dbg(_ERR_DBG, "invalid output.\n");
		ret = -EINVAL;
		goto err_exit;
	}

#if _NOTIFY_FB
	/* turn OFF LCD refresh mode. */
	if (output_mode == RT_OVERLAY_LCD1) {
		int display_mode;

		if (output_mode == RT_OVERLAY_SDENC)
			display_mode = RT_DISPLAY_SDENC;
		else
			display_mode = RT_DISPLAY_LCD1;

		if (sh_mobile_lcdc_refresh(RT_DISPLAY_REFRESH_OFF,  \
			display_mode)) {
			printk_dbg(_ERR_DBG, "can not change refresh mode.\n");
			ret = -EINVAL;
			goto err_exit;
		}
	}
#endif /* _NOTIFY_FB */
	/* create Overlay */
	{
		screen_ovl_new ovl;
		ovl.notify_overlay = function_notify_overlay;
		handle = screen_overlay_new(&ovl);
		if (handle == NULL) {
			printk_dbg(_ERR_DBG, "create_handle error.\n");
			ret = -EINVAL;
			goto err_exit;
		}
	}

	/* initialize overlay */
	{
		screen_ovl_initialize  ini;
		int                    rc;

		memset(&ini, 0, sizeof(ini)); /* all member will be 0 */
		ini.handle       = handle;
		ini.overlay_id   = overlay_id;
		ini.output_mode  = output_mode;

		rc = create_ovl_initialize(&ini, &dev->pix);
		if (rc != 0) {
			ret = -EINVAL;
			goto err_exit;
		}

		TRACE_LOG(FUNC_OVL_CREATE);
#if _OVL_DBG
		debuglog_screen_ovl_initialize(0, 0, &ini);
#endif
		rc = screen_overlay_initialize(&ini);
		printk_dbg(_OVL_DBG, \
			"screen_overlay_initialize result:%d\n", rc);
		if (rc) {
#if _ERR_RTAPI
			debuglog_screen_ovl_initialize(1, rc, &ini);
#endif
			printk_dbg(_ERR_DBG, \
				"screen_overlay_initialize failed.\n");
			ret = -EINVAL;
			goto err_exit;
		}
	}
	if ((dev->status & STAT_FBUFON) &&
		(dev->status & STAT_CONFIG_SFRMBUF)) {
		screen_ovl_initialize  ini;
		int                    rc;

		memset(&ini, 0, sizeof(ini)); /* all member will be 0 */
		ini.handle       = handle;
		ini.overlay_id   = overlay_id2;
		ini.output_mode  = output_mode;

		rc = create_ovl_initialize(&ini, &dev->fbuf.fmt);
		if (rc != 0) {
			ret = -EINVAL;
			goto err_exit;
		}

		TRACE_LOG(FUNC_OVL_CREATE);
#if _OVL_DBG
		debuglog_screen_ovl_initialize(0, 0, &ini);
#endif
		rc = screen_overlay_initialize(&ini);
		printk_dbg(_OVL_DBG, \
			"screen_overlay_initialize result:%d\n", rc);
		if (rc) {
#if _ERR_RTAPI
			debuglog_screen_ovl_initialize(1, rc, &ini);
#endif
			printk_dbg(_ERR_DBG, \
				"screen_overlay_initialize failed.\n");
			/* continue processing,
			   without framebuffer handling. */
			ovl_state[index] |= OVL_STATE_FRAMEERR;
		} else {
			/* set flag of using OVERLAY_1A, OVERLAY_2A */
			ovl_state[index] |= OVL_STATE_FRAMEBUF;
		}
	}

	/* overlay create successed. */
	dev->mDev.handle_overlay = handle;
	dev->mDev.overlay_id     = overlay_id;
	dev->mDev.overlay_id2    = overlay_id2;
	dev->mDev.output_mode    = output_mode;
	dev->mDev.callback       = callback;
	dev->mDev.state          = 1;

	memset(&dev->mDev.prev_param,  0xff, sizeof(dev->mDev.prev_param));
	memset(&dev->mDev.prev_param2, 0xff, sizeof(dev->mDev.prev_param2));
	dev->mDev.prev_chromakey = -1;
	dev->mDev.prev_globalalpha = -1;

	/* set overlay output mode */
	printk_dbg(_OVL_DBG, "lock semaphore\n");
	down(&sem_lock);
	ovl_outmode[index] = output_mode;
	up(&sem_lock);

	/* continue overlay initialize sequence.
	   When a problem was detected, internal state changed to ERROR
	   to re-create OVERLAY at later process. */

	/* set parameter */
	{
		screen_ovl_set_param  param;
		int           retval;

		memset(&param, 0, sizeof(param));
		param.handle      = handle;
		param.overlay_id  = overlay_id;
		retval = create_ovl_parameter(&param, &dev->crop, &dev->win, \
			dev->rotate, dev->mirror);
		if (retval) {
			ret = retval;
			goto err_exit2;
		}

		TRACE_LOG(FUNC_OVL_CREATE);
#if _OVL_DBG
		debuglog_screen_ovl_set_param(0, 0, &param);
#endif
		retval = screen_overlay_set_parameters(&param);
		printk_dbg(_OVL_DBG, \
			"screen_overlay_set_parameters result:%d\n", retval);
		if (retval == SMAP_LIB_OVERLAY_OK) {
			/* backup previous setting */
			dev->mDev.prev_param = param;
		} else {
#if _ERR_RTAPI
			debuglog_screen_ovl_set_param(1, retval, &param);
#endif
			printk_dbg(_ERR_DBG, \
				"screen_overlay_set_parameters failed.\n");
			ret = -EINVAL;
			goto err_exit2;
		}
	}
	if ((ovl_state[index] & OVL_STATE_FRAMEBUF) &&
		(ovl_state[index] & OVL_STATE_FRAMEERR) == 0) {
		screen_ovl_set_param  param;
		struct v4l2_crop      crop;
		int           retval;

		memset(&param, 0, sizeof(param));
		param.handle      = handle;
		param.overlay_id  = overlay_id2;

		memset(&crop,  0, sizeof(crop));
		crop.c.left   = 0;
		crop.c.top    = 0;
		crop.c.width  = dev->fbuf.fmt.width;
		crop.c.height = dev->fbuf.fmt.height;
		retval = create_ovl_parameter(&param, &crop, &dev->fbuf_win, \
			0 /* no rotate */, V4L2_MIRROR_NONE);

		if (dev->fbuf.flags & V4L2_FBUF_FLAG_GLOBAL_ALPHA) {
			/* updae global alpha */
			param.alpha = dev->win.global_alpha;
		}
		if (dev->fbuf.flags & V4L2_FBUF_FLAG_CHROMAKEY) {
			/* updae chromakey */
			param.key_color = dev->win.chromakey;
		}

		if (retval) {
			ret = retval;
			ovl_state[index] |= OVL_STATE_FRAMEERR;
		} else {

			TRACE_LOG(FUNC_OVL_CREATE);
#if _OVL_DBG
			debuglog_screen_ovl_set_param(0, 0, &param);
#endif
			retval = screen_overlay_set_parameters(&param);
			printk_dbg(_OVL_DBG, \
				"screen_overlay_set_parameters"
				" result:%d\n", retval);
			if (retval == SMAP_LIB_OVERLAY_OK) {
				/* backup previous setting */
				dev->mDev.prev_param2 = param;
			} else {
#if _ERR_RTAPI
				debuglog_screen_ovl_set_param(1,
					retval, &param);
#endif
				printk_dbg(_ERR_DBG, \
					"screen_overlay_set_parameters"
					" failed.\n");
				ovl_state[index] |= OVL_STATE_FRAMEERR;
			}
		}
	}

#if _UPDATE_FBPARAM
	/* set FB parameter */
		int display_mode;

		if (output_mode == RT_OVERLAY_SDENC)
			display_mode = RT_DISPLAY_SDENC;
		else
			display_mode = RT_DISPLAY_LCD1;

		printk_dbg(_OVL_DBG, "set_chromakey\n");
		if (sh_mobile_lcdc_keyclr_set(dev->win.chromakey, \
				display_mode) == 0) {
			/* record current config */
			dev->mDev.prev_chromakey = dev->win.chromakey;
		}

		printk_dbg(_OVL_DBG, "set_global_alpha\n");
		if (sh_mobile_lcdc_alpha_set(dev->win.global_alpha, \
			display_mode) == 0) {
			/* record current config */
			dev->mDev.prev_globalalpha = dev->win.global_alpha;
		}
	}
#endif

	/* start notify */
	{
		screen_ovl_notify ovl;
		unsigned int  _mask;
		int           retval;

		/*initialize state */
		_mask = OVL_STATE_READY | OVL_STATE_ERROR;
		ovl_state[index]  &= ~_mask;
		result_code[index] = SMAP_LIB_OVERLAY_NG;

		TRACE_LOG(FUNC_OVL_CREATE);
		ovl.handle     = handle;
		ovl.overlay_id = overlay_id;
		retval = screen_overlay_start_notify(&ovl);
		if (retval != SMAP_LIB_DISPLAY_OK) {
			printk_dbg(_ERR_DBG, \
				"screen_overlay_start_notify failed.\n");
			ret = -EINVAL;
			goto err_exit2;
		}

		printk_dbg(_OVL_DBG, "wait start notify\n");
		retval = wait_event_interruptible_timeout(wait_id[index],     \
				((ovl_state[index] & _mask) != 0),  \
				(2 * HZ));
		printk_dbg(_OVL_DBG, "wait start result: 0x%x state: 0x%x\n", \
				result_code[index], ovl_state[index]);
		if ((ovl_state[index] & OVL_STATE_READY) == 0) {
			printk_dbg(_ERR_DBG, \
				"screen_overlay_start_notify failed.\n");
			ret = -EINVAL;
			goto err_exit2;
		}
	}
	if ((ovl_state[index] & OVL_STATE_FRAMEBUF) &&
		(ovl_state[index] & OVL_STATE_FRAMEERR) == 0) {
		screen_ovl_notify ovl;
		unsigned int  _mask;
		int           retval;

		/*initialize state */
		_mask = OVL_STATE_FRMREADY | OVL_STATE_ERROR;
		ovl_state[index]  &= ~_mask;
		result_code[index] = SMAP_LIB_OVERLAY_NG;

		TRACE_LOG(FUNC_OVL_CREATE);
		ovl.handle     = handle;
		ovl.overlay_id = overlay_id2;
		retval = screen_overlay_start_notify(&ovl);
		if (retval != SMAP_LIB_DISPLAY_OK) {
			printk_dbg(_ERR_DBG, \
				"screen_overlay_start_notify failed.\n");
			ovl_state[index] |= OVL_STATE_FRAMEERR;
		}

		printk_dbg(_OVL_DBG, "wait start notify\n");
		retval = wait_event_interruptible_timeout(wait_id[index],     \
				((ovl_state[index] & _mask) != 0),  \
				(2 * HZ));
		printk_dbg(_OVL_DBG, "wait start result: 0x%x state: 0x%x\n", \
				result_code[index], ovl_state[index]);
		if ((ovl_state[index] & OVL_STATE_FRMREADY) == 0) {
			printk_dbg(_ERR_DBG, \
				"screen_overlay_start_notify failed.\n");
			ovl_state[index] |= OVL_STATE_FRAMEERR;
		}
	}
	TRACE_LEAVE(FUNC_OVL_CREATE);
	return 0;

err_exit2:
	/* current state change to ERROR */
	ovl_state[index] |= OVL_STATE_ERROR;
	TRACE_LEAVE(FUNC_OVL_CREATE);
	return 0;

err_exit:
	if (handle) {
		screen_ovl_delete del;
		del.handle = handle;
		screen_overlay_delete(&del);
	}
#if _NOTIFY_FB
	/* turn ON LCD refresh mode. */
	if (output_mode == RT_OVERLAY_LCD1 &&
		ovl_outmode[0] != RT_OVERLAY_LCD1 &&
		ovl_outmode[1] != RT_OVERLAY_LCD1) {
		/* turn on refresh mode. */
		sh_mobile_lcdc_refresh(RT_DISPLAY_REFRESH_ON, RT_DISPLAY_LCD1);
	}
#endif /* _NOTIFY_FB */
	TRACE_LEAVE(FUNC_OVL_CREATE);
	return -1;
#endif
}

int  videobuf_queue_sh_mobile_interface_destroyoverlay(             \
	struct sh_mobile_v4l2_device *dev)
{
#if !RTAPI_MEMORY_INTERFACE
	return -EINVAL;
#else
	screen_ovl_quit quit;
	int             ret = 0;
	int             index;
	TRACE_ENTER(FUNC_OVL_DESTROY);

	if (dev->mDev.handle_overlay == NULL) {
		/* overlay not created. */
		TRACE_LEAVE(FUNC_OVL_DESTROY);
		return -EBUSY;
	}

	if (dev->mDev.overlay_id == RT_OVERLAY_ID1) {
		/* overlay ID1 */
		index = 0;
	} else if (dev->mDev.overlay_id == RT_OVERLAY_ID2) {
		/* overlay ID2 */
		index = 1;
	} else {
		printk_dbg(_ERR_DBG, "unknown overlay id.\n");
		TRACE_LEAVE(FUNC_OVL_DESTROY);
		return -EINVAL;
	}

	quit.handle     = dev->mDev.handle_overlay;
	quit.overlay_id = dev->mDev.overlay_id;
	printk_dbg(_OVL_DBG, \
		"quit handle: %p id:%d "
		"\n",
		quit.handle,      quit.overlay_id);
	ret = screen_overlay_quit(&quit);
	if (ret != SMAP_LIB_OVERLAY_OK) {
		printk_dbg(_ERR_RTAPI, "screen_overlay_quit " \
			"error: %d\n", ret);
		printk_dbg(_ERR_RTAPI, " handle:%p overlay_id:%d\n",\
			quit.handle, quit.overlay_id);
		printk_dbg(_ERR_DBG, "destroy overlay failed.\n");
		ret = -1;
	}

	if (ovl_state[index] & OVL_STATE_FRAMEBUF) {
		int ret_ovl;

		quit.handle     = dev->mDev.handle_overlay;
		quit.overlay_id = dev->mDev.overlay_id2;
		printk_dbg(_OVL_DBG, \
			"quit handle: %p id:%d "
			"\n",
			quit.handle,      quit.overlay_id);

		ret_ovl = screen_overlay_quit(&quit);
		if (ret_ovl != SMAP_LIB_OVERLAY_OK) {
			printk_dbg(_ERR_RTAPI, "screen_overlay_quit " \
				"error: %d\n", ret);
			printk_dbg(_ERR_RTAPI, " handle:%p overlay_id:%d\n",\
				quit.handle, quit.overlay_id);
			printk_dbg(_ERR_DBG, "destroy overlay failed.\n");
			ret = -1;
		}
		ovl_state[index] &= ~(OVL_STATE_FRAMEBUF | \
			OVL_STATE_FRMREADY | OVL_STATE_FRAMEERR);
	}

	/* free handle */
	if (dev->mDev.handle_overlay) {
		screen_ovl_delete del;
		del.handle = dev->mDev.handle_overlay;
		screen_overlay_delete(&del);

		/* reset state */
		ovl_state[index]  = 0;
		ovl_outmode[index] = 0;

		dev->mDev.handle_overlay = NULL;
		dev->mDev.overlay_id     = 0;
		dev->mDev.overlay_id2    = 0;
		dev->mDev.output_mode    = 0;
		dev->mDev.callback       = 0;
		dev->mDev.state          = 0;
#if _NOTIFY_FB
		/* turn ON LCD refresh mode. */
		printk_dbg(_OVL_DBG, "lock semaphore\n");
		down(&sem_lock);
		if ((ovl_outmode[0] != RT_DISPLAY_LCD1 &&
		     ovl_outmode[1] != RT_DISPLAY_LCD1)) {
			/* currently not use overlay */
			if (sh_mobile_lcdc_refresh(RT_DISPLAY_REFRESH_ON, \
				RT_DISPLAY_LCD1)) {
				printk_dbg(_ERR_DBG, \
					"can not change refresh mode.\n");
				/* continue process to initialize variables. */
			}
		}
		up(&sem_lock);
#endif /* _NOTIFY_FB */
	}
	TRACE_LEAVE(FUNC_OVL_DESTROY);
	return 0;
#endif
}

int  videobuf_queue_sh_mobile_interface_postimage(
	struct sh_mobile_v4l2_videobuffer *buf,
	struct sh_mobile_v4l2_device *dev)
{
	int ret = 0;
#if RTAPI_MEMORY_INTERFACE
	screen_ovl_write_image  write;
	screen_ovl_set_param    param;
	struct videobuf_mfi_memory *mem;
	int                     index;
	TRACE_ENTER(FUNC_OVL_POSTIMAGE);

	if (!check_available_overlay(dev)) {
		printk_dbg(_ERR_DBG, "can not handle overlay.\n");
		TRACE_LEAVE(FUNC_OVL_POSTIMAGE);
		return -EINVAL;
	}

	if (dev->mDev.overlay_id == RT_OVERLAY_ID1) {
		/* overlay ID1 */
		index = 0;
	} else if (dev->mDev.overlay_id == RT_OVERLAY_ID2) {
		/* overlay ID2 */
		index = 1;
	} else {
		printk_dbg(_ERR_DBG, "unknown overlay id.\n");
		TRACE_LEAVE(FUNC_OVL_POSTIMAGE);
		return -EINVAL;
	}
	if (ovl_state[index] & OVL_STATE_ERROR) {
		printk_dbg(_ERR_DBG, "createo verlay error.\n");
		TRACE_LEAVE(FUNC_OVL_POSTIMAGE);
		return -ERESTARTSYS;
	}
	if (!(ovl_state[index] & OVL_STATE_READY)) {
		printk_dbg(_ERR_DBG, "no valid device.\n");
		TRACE_LEAVE(FUNC_OVL_POSTIMAGE);
		return -EINVAL;
	}

	mem = (struct videobuf_mfi_memory *)buf->vb.priv;

	/****************************************/
	/* handle frame buffer                  */
	/****************************************/

	if ((ovl_state[index] & OVL_STATE_FRMREADY) == 0 ||
		buf->fbuf.base == NULL) {
		/* skip write image */
		goto skip_framebuf;
	}

	write.handle        = dev->mDev.handle_overlay;
	write.overlay_id    = dev->mDev.overlay_id2;
	write.image_handle  = NULL;
	write.user_data     = (unsigned long) buf;
	write.image_address = buf->fbuf.base;
	write.image_size    = buf->fbuf.fmt.sizeimage;

	{
		struct v4l2_crop  crop;

		memset(&param, 0, sizeof(param));
		param.handle      = dev->mDev.handle_overlay;
		param.overlay_id  = dev->mDev.overlay_id2;

		memset(&crop,  0, sizeof(crop));
		crop.c.left   = 0;
		crop.c.top    = 0;
		crop.c.width  = buf->fbuf.fmt.width;
		crop.c.height = buf->fbuf.fmt.height;
		ret = create_ovl_parameter(&param, &crop, &buf->fbuf_win, \
			0 /* no rotate */, V4L2_MIRROR_NONE);

		if (buf->fbuf.flags & V4L2_FBUF_FLAG_GLOBAL_ALPHA) {
			/* updae global alpha */
			param.alpha = buf->win.global_alpha;
		}
		if (buf->fbuf.flags & V4L2_FBUF_FLAG_CHROMAKEY) {
			/* updae chromakey */
			param.key_color = buf->win.chromakey;
		}

		if (ret) {
			/* set error flag */
			ovl_state[index] |= OVL_STATE_FRAMEERR;
		} else if (memcmp(&param, &dev->mDev.prev_param2, \
			sizeof(param)) != 0) {

			TRACE_LOG(FUNC_OVL_POSTIMAGE);
#if _OVL_DBG
			debuglog_screen_ovl_set_param(0, 0, &param);
#endif
			ret = screen_overlay_set_parameters(&param);
			printk_dbg(_OVL_DBG, \
				"screen_overlay_set_parameters"
				" result:%d\n", ret);
			if (ret == SMAP_LIB_OVERLAY_OK) {
				/* backup previous setting */
				dev->mDev.prev_param2 = param;
			} else {
#if _ERR_RTAPI
				debuglog_screen_ovl_set_param(1,
					ret, &param);
#endif
				printk_dbg(_ERR_DBG, \
					"screen_overlay_set_parameters"
					" failed.\n");
				ovl_state[index] |= OVL_STATE_FRAMEERR;
				goto skip_framebuf;
			}
		}
	}

	TRACE_LOG(FUNC_OVL_POSTIMAGE);
	printk_dbg(_OVL_DBG, \
	"write handle:%p id:%d " \
	"memhandle:%p userdata:%lx " \
	"image_address:%p  image_size:0x%lx" \
	"\n",
		write.handle, write.overlay_id,
		write.image_handle, write.user_data,
		write.image_address, write.image_size);
	ret = screen_overlay_write_image(&write);
	if (ret != SMAP_LIB_OVERLAY_OK) {
		printk_dbg(_ERR_RTAPI, "screen_overlay_write_image "\
			"error: %d\n", ret);
		printk_dbg(_ERR_RTAPI, \
			" handle:%p overlay_id:%d " \
			"image_handle:%p user_data:%lx " \
			"image_address:%p  image_size:0x%lx" \
			"\n",
			write.handle, write.overlay_id,
			write.image_handle, write.user_data,
			write.image_address, write.image_size);
		printk_dbg(_ERR_DBG, "[error] write_image\n");
		ovl_state[index] |= OVL_STATE_FRAMEERR;
		goto skip_framebuf;
	}

skip_framebuf:
	/****************************************/
	/* handle overlay                       */
	/****************************************/

	memset(&write, 0, sizeof(write));
	write.handle        = dev->mDev.handle_overlay;
	write.overlay_id    = dev->mDev.overlay_id;
	write.image_handle  = sh_mobile_appmem_getmemoryhandle(
		dev->mMem.appmem_handle);
	write.user_data     = (unsigned long) buf;
	write.image_address = mem->vaddr;
	write.image_size    = mem->size;

	memset(&param, 0, sizeof(param));
	param.handle       = dev->mDev.handle_overlay;
	param.overlay_id   = dev->mDev.overlay_id;
	if (create_ovl_parameter(&param, &buf->crop, &buf->win, \
		buf->rotate, buf->mirror) != 0) {
		/* report error */
		return -1;
	}

	if (memcmp(&param, &dev->mDev.prev_param, sizeof(param)) != 0) {
		TRACE_LOG(FUNC_OVL_POSTIMAGE);
#if _OVL_DBG
		debuglog_screen_ovl_set_param(0, 0, &param);
#endif
		ret = screen_overlay_set_parameters(&param);
		if (ret ==  SMAP_LIB_OVERLAY_OK) {
			/* record current config */
			dev->mDev.prev_param = param;
		} else {
#if _ERR_RTAPI
			debuglog_screen_ovl_set_param(1, ret, &param);
#endif
			printk_dbg(_ERR_DBG, "[error] set_parameters\n");
			return -1;
		}
	}

#if _UPDATE_FBPARAM
	/* set FB parameter */
	if (ovl_state[index] & OVL_STATE_FRAMEBUF) {
		/* framebuffer under FB driver not used. */
		/* nothing to do.                        */
	} else if (dev->mDev.output_mode == RT_OVERLAY_LCD1) {
		int display_mode;

		if (dev->mDev.output_mode == RT_OVERLAY_SDENC)
			display_mode = RT_DISPLAY_SDENC;
		else
			display_mode = RT_DISPLAY_LCD1;

		if (buf->win.chromakey != dev->mDev.prev_chromakey) {
			printk_dbg(_OVL_DBG, "set_chromakey\n");
			if (sh_mobile_lcdc_keyclr_set(buf->win.chromakey, \
				display_mode) == 0) {
				/* record current config */
				dev->mDev.prev_chromakey = buf->win.chromakey;
			}
		}

		if (buf->win.global_alpha != dev->mDev.prev_globalalpha) {
			printk_dbg(_OVL_DBG, "set_global_alpha\n");
			if (sh_mobile_lcdc_alpha_set( \
				buf->win.global_alpha, display_mode) == 0) {
				/* record current config */
				dev->mDev.prev_globalalpha = \
					buf->win.global_alpha;
			}
		}
	}
#endif

	TRACE_LOG(FUNC_OVL_POSTIMAGE);
	printk_dbg(_OVL_DBG, \
	"write handle:%p id:%d " \
	"memhandle:%p userdata:%lx " \
	"image_address:%p  image_size:0x%lx" \
	"\n",
		write.handle, write.overlay_id,
		write.image_handle, write.user_data,
		write.image_address, write.image_size);
	ret = screen_overlay_write_image(&write);
	if (ret != SMAP_LIB_OVERLAY_OK) {
		printk_dbg(_ERR_RTAPI, "screen_overlay_write_image "\
			"error: %d\n", ret);
		printk_dbg(_ERR_RTAPI, \
			" handle:%p overlay_id:%d " \
			"image_handle:%p user_data:%lx " \
			"image_address:%p  image_size:0x%lx" \
			"\n",
			write.handle, write.overlay_id,
			write.image_handle, write.user_data,
			write.image_address, write.image_size);
		printk_dbg(_ERR_DBG, "[error] write_image\n");
		return -1;
	}
	TRACE_LEAVE(FUNC_OVL_POSTIMAGE);
#endif
	return ret;
}


void videobuf_queue_sh_mobilebuf_init(struct videobuf_queue *q,
				    const struct videobuf_queue_ops *ops,
				    struct device *dev,
				    spinlock_t *irqlock,
				    enum v4l2_buf_type type,
				    enum v4l2_field field,
				    unsigned int msize,
				    void *priv)
{
	videobuf_queue_core_init(q, ops, dev, irqlock, type, field, msize,
				 priv, &qops);
}

void videobuf_sh_mobilebuf_free(struct videobuf_queue *q,
			      struct videobuf_buffer *buf)
{
	struct videobuf_mfi_memory *mem = buf->priv;

	/* mmapped memory can't be freed here, otherwise mmapped region
	   would be released, while still needed. In this case, the memory
	   release should happen inside videobuf_vm_close().
	   So, it should free memory only if the memory were allocated for
	   read() operation.
	 */
	if (buf->memory != V4L2_MEMORY_USERPTR)
		return;

	if (!mem)
		return;

	MAGIC_CHECK(mem->magic, MAGIC_DC_MEM);

	/* handle user space pointer case */
	if (buf->baddr) {
		printk_dbg(_MEM_DBG, "release user memory\n");
		videobuf_mfi_user_put(q->priv_data, mem);
		return;
	}

	mem->vaddr = NULL;
}

