/*
 * Fifo-attached Serial Interface (FSI) support
 *
 * Copyright (C) 2011 Renesas Electronics Corporation.
 *
 * Copyright (C) 2009 Renesas Solutions Corp.
 * Kuninori Morimoto <morimoto.kuninori@renesas.com>
 *
 * Based on ssi.c
 * Copyright (c) 2007 Manuel Lauss <mano@roarinelk.homelinux.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/delay.h>
#include <linux/list.h>
#include <linux/pm_runtime.h>
#include <linux/io.h>
#include <linux/slab.h>
#include <linux/dmaengine.h>
#include <linux/sh_dma.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/initval.h>
#include <sound/soc.h>
#include <sound/pcm_params.h>
#include <sound/sh_fsi.h>
#include <asm/atomic.h>

/* register offset */
#define DO_FMT		0x0000
#define DOFF_CTL	0x0004
#define DOFF_ST		0x0008
#define DI_FMT		0x000C
#define DIFF_CTL	0x0010
#define DIFF_ST		0x0014
#define CKG1		0x0018	/* ACK_MD */
#define CKG2		0x001C	/* ACK_RV */
#define DIDT		0x0020
#define DODT		0x0024
#define MUTE_ST		0x0028
#define OUT_DMAC	0x002C
#define OUT_SEL		0x0030
#define IN_DMAC		0x0038
#define REG_END		IN_DMAC

#define A_MST_CTLR	0x0180
#define B_MST_CTLR	0x01A0
#define CPU_INT_ST	0x01F4
#define CPU_IEMSK	0x01F8
#define CPU_IMSK	0x01FC
#define INT_ST		0x0200
#define IEMSK		0x0204
#define IMSK		0x0208
#define MUTE		0x020C
#define CLK_RST		0x0210
#define SOFT_RST	0x0214
#define FIFO_SZ		0x0218
#define SWAP_SEL	0x0228
#define MREG_START	A_MST_CTLR
#define MREG_END	SWAP_SEL

/* DO_FMT */
/* DI_FMT */
#define CR_FMT(param) ((param) << 4)
# define CR_MONO	0x0
# define CR_MONO_D	0x1
# define CR_PCM		0x2
# define CR_I2S		0x3
# define CR_TDM		0x4
# define CR_TDM_D	0x5
# define CR_SPDIF	0x00000130

/* DOFF_CTL */
/* DIFF_CTL */
#define IRQ_HALF	0x00100000
#define IRQ_ONE		0x00200000
#define FIFO_CLR	0x00000001

/* DOFF_ST */
#define ERR_OVER	0x00000010
#define ERR_UNDER	0x00000001
#define ST_ERR		(ERR_OVER | ERR_UNDER)

/* CKG1 */
#define ACKMD_MASK	0x00007000
#define BPFMD_MASK	0x00000700

/* A/B MST_CTLR */
#define BP		(1 << 4) /* Fix the signal of Biphase output */
#define SE		(1 << 0) /* Fix the master clock */

/* CLK_RST */
#define B_CLK		0x00000010
#define A_CLK		0x00000001

/* INT_ST */
#define INT_B_IN	(1 << 12)
#define INT_B_OUT	(1 << 8)
#define INT_A_IN	(1 << 4)
#define INT_A_OUT	(1 << 0)



/* SOFT_RST */
#define PBSR		(1 << 12) /* Port B Software Reset */
#define PASR		(1 <<  8) /* Port A Software Reset */
#define IR		(1 <<  4) /* Interrupt Reset */
#define FSISR		(1 <<  0) /* Software Reset */

/* FIFO_SZ */
#define OUT_SZ_MASK	0x7
#define BO_SZ_SHIFT	8
#define AO_SZ_SHIFT	0

/* SWAP SELECT */
#define NO_SWAP		0x00
#define BYTE_SWAP	0x01
#define WORD_SWAP	0x02
#define BYTEWORD_SWAP	0x03

#define FSI_RATES SNDRV_PCM_RATE_8000_96000

#define FSI_FMTS (SNDRV_PCM_FMTBIT_S24_LE | SNDRV_PCM_FMTBIT_S16_LE)


#define PLAY_STREAM(substream) (substream->stream == SNDRV_PCM_STREAM_PLAYBACK)
#define CAPT_STREAM(substream) (substream->stream == SNDRV_PCM_STREAM_CAPTURE)


/* Get struct fsi_priv from struct snd_pcm_substream */
#define GET_CPU_DAI(substream) \
	(((struct snd_soc_pcm_runtime *)substream->private_data)->dai->cpu_dai)

/************************************************************************


	struct


************************************************************************/
struct fsi_audio_stream {
	struct sh_dmae_slave *dma_slave;
	struct dma_chan *chan;
	dma_cookie_t cookie;
	struct dma_async_tx_descriptor *desc;
	enum dma_data_direction direction;
	u32 dma_size[2];
	int head, tail;
	unsigned int dma_pos;

	spinlock_t dma_lock;	/* for locking in DMA operations */

	struct work_struct trigger_work;
	struct snd_pcm_substream *stream;

	/* for alsa */
	int active:1;	/* we are using this stream for transfer now */
	int power:1;
	unsigned int old_offset;
	snd_pcm_uframes_t output_ptr;
	unsigned long period_byte;
	int last_flag;

	int period_len;
	int buffer_len;
	int periods;
};

struct fsi_priv {
	void __iomem *base;
	struct fsi_master *master;
	int ch;
	struct fsi_pcm_resource {
		char	*name;
		u_int	base;
		void __iomem	*vadr;
		int	irq;
		struct sh_fsi_plat_data *plat;
	} src;
	u_int8_t in_use;		/* PCM in use flag. */
	struct fsi_audio_stream s[2];

	int fifo_max;
	int chan;
	u32 mst_ctrl;
};

struct fsi_regs {
	int ver;
	u32 int_st;
	u32 iemsk;
	u32 imsk;
};

struct fsi_master {
	void __iomem *base;
	int irq;
	int irq_enable;		/* enable/free irq flag */
	struct fsi_priv fsia;
	struct fsi_priv fsib;
	struct fsi_regs *regs;
	struct sh_fsi_platform_info *info;	/* info have dma slave */
	spinlock_t lock;	/* enable/free irq lock */
};

static bool filter(struct dma_chan *chan, void *param)
{
	chan->private = param;
	return true;
}

static void fsi_trigger_work(struct work_struct *work);
static int fsi_dma_start(struct fsi_audio_stream *s, dma_addr_t buff, u32 size);


/************************************************************************


	basic read write function


************************************************************************/
static void __fsi_reg_write(u32 reg, u32 data)
{
	/* valid data area is 24bit */
	data &= 0x00ffffff;

	__raw_writel(data, reg);
}

static u32 __fsi_reg_read(u32 reg)
{
	return __raw_readl(reg);
}

static void __fsi_reg_mask_set(u32 reg, u32 mask, u32 data)
{
	u32 val = __fsi_reg_read(reg);

	val &= ~mask;
	val |= data & mask;

	__fsi_reg_write(reg, val);
}

static void fsi_reg_write(struct fsi_priv *fsi, u32 reg, u32 data)
{
	if (reg > REG_END)
		return;

	__fsi_reg_write((u32)(fsi->base + reg), data);
}

static u32 fsi_reg_read(struct fsi_priv *fsi, u32 reg)
{
	if (reg > REG_END)
		return 0;

	return __fsi_reg_read((u32)(fsi->base + reg));
}

static void fsi_reg_mask_set(struct fsi_priv *fsi, u32 reg, u32 mask, u32 data)
{
	if (reg > REG_END)
		return;

	__fsi_reg_mask_set((u32)(fsi->base + reg), mask, data);
}

static void fsi_master_write(struct fsi_master *master, u32 reg, u32 data)
{
	unsigned long flags;

	if ((reg < MREG_START) ||
		(reg > MREG_END))
		return;

	spin_lock_irqsave(&master->lock, flags);
	__fsi_reg_write((u32)(master->base + reg), data);
	spin_unlock_irqrestore(&master->lock, flags);
}

static u32 fsi_master_read(struct fsi_master *master, u32 reg)
{
	u32 ret;
	unsigned long flags;

	if ((reg < MREG_START) ||
		(reg > MREG_END))
		return 0;

	spin_lock_irqsave(&master->lock, flags);
	ret = __fsi_reg_read((u32)(master->base + reg));
	spin_unlock_irqrestore(&master->lock, flags);

	return ret;
}

static void fsi_master_mask_set(struct fsi_master *master,
				u32 reg, u32 mask, u32 data)
{
	unsigned long flags;

	if ((reg < MREG_START) || (reg > MREG_END))
		return;

	spin_lock_irqsave(&master->lock, flags);
	__fsi_reg_mask_set((u32)(master->base + reg), mask, data);
	spin_unlock_irqrestore(&master->lock, flags);
}

/************************************************************************


	basic function


************************************************************************/
static struct fsi_master *fsi_get_master(struct fsi_priv *fsi)
{
	return fsi->master;
}

static int fsi_is_port_a(struct fsi_priv *fsi)
{
	return fsi->master->base == fsi->base;
}

static struct snd_soc_dai *fsi_get_dai(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai_link *machine = rtd->dai;

	return machine->cpu_dai;
}

static struct fsi_priv *fsi_get_priv(struct snd_pcm_substream *substream)
{
	struct snd_soc_dai *dai = fsi_get_dai(substream);

	return dai->private_data;
}

static u32 fsi_get_info_flags(struct fsi_priv *fsi)
{
	int is_porta = fsi_is_port_a(fsi);
	struct fsi_master *master = fsi_get_master(fsi);

	return is_porta ? master->info->porta_flags :
		master->info->portb_flags;
}

static u32 fsi_get_info_flags2(struct fsi_priv *fsi)
{
	int is_porta = fsi_is_port_a(fsi);
	struct fsi_master *master = fsi_get_master(fsi);

	return is_porta ? master->info->porta_flags2 :
		master->info->portb_flags2;
}

static int fsi_is_master_mode(struct fsi_priv *fsi, int is_play)
{
	u32 mode;
	u32 flags = fsi_get_info_flags(fsi);

	mode = is_play ? SH_FSI_OUT_SLAVE_MODE : SH_FSI_IN_SLAVE_MODE;

	/* return
	 * 1 : master mode
	 * 0 : slave mode
	 */

	return (mode & flags) != mode;
}

static u32 fsi_port_ab_io_bit(struct fsi_priv *fsi, int is_play)
{
	int is_porta = fsi_is_port_a(fsi);
	u32 data;

	if (is_porta)
		data = is_play ? (1 << 0) : (1 << 4);
	else
		data = is_play ? (1 << 8) : (1 << 12);

	return data;
}

static void fsi_stream_push(struct fsi_priv *fsi,
				struct snd_pcm_substream *substream,
				u32 buffer_len,
				u32 period_len)
{
	struct fsi_audio_stream *s;

	s = &fsi->s[substream->stream];
	s->stream = substream;
	s->buffer_len = buffer_len;
	s->period_len = period_len;
	s->periods = 0;
}

static void fsi_stream_pop(struct fsi_priv *fsi,
				struct snd_pcm_substream *substream)
{
	struct fsi_audio_stream *s;

	s = &fsi->s[substream->stream];
	substream = s->stream;
	s->stream = NULL;
	s->buffer_len = 0;
	s->period_len = 0;
	s->periods = 0;
}


/************************************************************************


		irq function


************************************************************************/
static void fsi_irq_enable(struct fsi_priv *fsi, int is_play)
{
	u32 data = fsi_port_ab_io_bit(fsi, is_play);
	struct fsi_master *master = fsi_get_master(fsi);

	fsi_master_mask_set(master, master->regs->imsk, data, data);
	fsi_master_mask_set(master, master->regs->iemsk, data, data);
}

static void fsi_irq_disable(struct fsi_priv *fsi, int is_play)
{
	u32 data = fsi_port_ab_io_bit(fsi, is_play);
	struct fsi_master *master = fsi_get_master(fsi);

	fsi_master_mask_set(master, master->regs->imsk, data, 0);
	fsi_master_mask_set(master, master->regs->iemsk, data, 0);
}

static u32 fsi_irq_get_status(struct fsi_master *master)
{
	return fsi_master_read(master, master->regs->int_st);
}

static void fsi_irq_clear_all_status(struct fsi_master *master)
{
	fsi_master_write(master, master->regs->int_st, 0x0000000);
}

static void fsi_irq_clear_status(struct fsi_priv *fsi)
{
	u32 data = 0;
	struct fsi_master *master = fsi_get_master(fsi);

	data |= fsi_port_ab_io_bit(fsi, 0);
	data |= fsi_port_ab_io_bit(fsi, 1);

	/* clear interrupt factor */
	fsi_master_mask_set(master, master->regs->int_st, data, 0);
}

/************************************************************************


		SPDIF master clock function

These functions are used later FSI2
************************************************************************/
static void fsi_spdif_clk_ctrl(struct fsi_priv *fsi, int enable)
{
	struct fsi_master *master = fsi_get_master(fsi);
	u32 val = BP | SE;

	if (master->regs->ver < 2) {
		pr_err("fsi: register access err (%s)\n", __func__);
		return;
	}

	if (enable)
		fsi_master_mask_set(master, fsi->mst_ctrl, val, val);
	else
		fsi_master_mask_set(master, fsi->mst_ctrl, val, 0);
}

/************************************************************************


	ctrl function


************************************************************************/
static void fsi_clk_ctrl(struct fsi_priv *fsi, int enable)
{
	u32 val = fsi_is_port_a(fsi) ? (1 << 0) : (1 << 4);
	struct fsi_master *master = fsi_get_master(fsi);

	if (enable)
		fsi_master_mask_set(master, CLK_RST, val, val);
	else
		fsi_master_mask_set(master, CLK_RST, val, 0);
}

static void fsi_fifo_init(struct fsi_priv *fsi,
				int is_play,
				struct snd_soc_dai *dai)
{
	struct fsi_master *master = fsi_get_master(fsi);
	u32 ctrl, shift, i;

	/* get on-chip RAM capacity */
	shift = fsi_master_read(master, FIFO_SZ);
	shift >>= fsi_is_port_a(fsi) ? AO_SZ_SHIFT : BO_SZ_SHIFT;
	shift &= OUT_SZ_MASK;
	fsi->fifo_max = 256 << shift;
	dev_dbg(dai->dev, "fifo = %d words\n", fsi->fifo_max);

	/*
	 * The maximum number of sample data varies depending
	 * on the number of channels selected for the format.
	 *
	 * FIFOs are used in 4-channel units in 3-channel mode
	 * and in 8-channel units in 5- to 7-channel mode
	 * meaning that more FIFOs than the required size of DPRAM
	 * are used.
	 *
	 * ex) if 256 words of DP-RAM is connected
	 * 1 channel:  256 (256 x 1 = 256)
	 * 2 channels: 128 (128 x 2 = 256)
	 * 3 channels:  64 ( 64 x 3 = 192)
	 * 4 channels:  64 ( 64 x 4 = 256)
	 * 5 channels:  32 ( 32 x 5 = 160)
	 * 6 channels:  32 ( 32 x 6 = 192)
	 * 7 channels:  32 ( 32 x 7 = 224)
	 * 8 channels:  32 ( 32 x 8 = 256)
	 */
	for (i = 1; i < fsi->chan; i <<= 1)
		fsi->fifo_max >>= 1;
	dev_dbg(dai->dev, "%d channel %d store\n", fsi->chan, fsi->fifo_max);

	ctrl = is_play ? DOFF_CTL : DIFF_CTL;

	/* set interrupt generation factor */
	fsi_reg_write(fsi, ctrl, IRQ_HALF);

	/* clear FIFO */
	fsi_reg_mask_set(fsi, ctrl, FIFO_CLR, FIFO_CLR);
}

static void fsi_soft_all_reset(struct fsi_master *master)
{
	/* port AB reset */
	fsi_master_mask_set(master, SOFT_RST, PASR | PBSR, 0);
	mdelay(10);

	/* soft reset */
	fsi_master_mask_set(master, SOFT_RST, FSISR, 0);
	fsi_master_mask_set(master, SOFT_RST, FSISR, FSISR);
	mdelay(10);
}

static int audio_dma_writable_flag(struct fsi_audio_stream *s)
{
	struct dma_tx_state txstate;
	enum dma_status status;
	int ret;

	status = s->chan->device->device_tx_status(s->chan, s->cookie,
							&txstate);

	if (status == DMA_IN_PROGRESS)
		ret = txstate.residue & TCRB_USED ? 1 : 0; /* case repeat */
	else
		ret = (status != DMA_SUCCESS);

	/* 0: writable */
	/* 1: busy */

	return ret;
}

/* Gets the period-byte. */
static void audio_get_period_byte(struct fsi_audio_stream *s)
{
	struct snd_soc_dai *dai = GET_CPU_DAI(s->stream);
	struct snd_pcm_substream *substream = s->stream;
	struct snd_pcm_runtime *runtime = substream->runtime;
	snd_pcm_uframes_t frames;

	frames = runtime->buffer_size - s->output_ptr % runtime->buffer_size;

	if (PLAY_STREAM(substream)) {
		if ((runtime->control->appl_ptr >= s->output_ptr) &&
			((s->output_ptr + frames) > runtime->control->appl_ptr))
			frames = runtime->control->appl_ptr -
							s->output_ptr;
	} else {
		snd_pcm_uframes_t data_size;

		if (s->output_ptr < runtime->control->appl_ptr)
			data_size = (runtime->boundary -
					runtime->control->appl_ptr) +
					s->output_ptr;
		else
			data_size = s->output_ptr - runtime->control->appl_ptr;

		if (runtime->buffer_size < (frames + data_size))
			frames = runtime->buffer_size - data_size;
	}

	if ((runtime->period_size * 2) > frames)
		s->period_byte = frames_to_bytes(runtime, frames);
	else
		s->period_byte = snd_pcm_lib_period_bytes(substream);

	dev_dbg(dai->dev, "ptr %lu, s->period_byte %lu\n",
			s->output_ptr, runtime->control->appl_ptr);
}



static void fsi_process_dma(struct fsi_audio_stream *s)
{
	struct snd_pcm_runtime *runtime;
	unsigned int dma_size;
	unsigned int offset;
	dma_addr_t dma_ptr;
	int ret;
	struct fsi_priv *fsi;

	/* need this check */
	if (s->stream == NULL)
		return;

	runtime = s->stream->runtime;
	fsi = fsi_get_priv(s->stream);

	/* must be set here - only valid for running streams,
	   not for forced_clock dma fills */
	while (s->active && (audio_dma_writable_flag(s) == 0)) {
		if (s->old_offset) {
			s->output_ptr = s->old_offset - 1;
			s->old_offset = 0;
		}

		audio_get_period_byte(s);
		if (s->period_byte == 0) {
			return;
			/* new data not yet transfered
			 from user area. or data finished. */
		}
		dma_size = s->period_byte;

		offset = frames_to_bytes(runtime,
				s->output_ptr % runtime->buffer_size);
		dma_ptr = runtime->dma_addr + offset;

		if (PLAY_STREAM(s->stream))
			fsi_irq_enable(fsi, 1);
		else
			fsi_irq_enable(fsi, 0);

		ret = fsi_dma_start(s, dma_ptr, dma_size);
		if (ret)
			goto end;

		s->output_ptr += bytes_to_frames(runtime, dma_size);
		if (s->output_ptr >= runtime->boundary)
			s->output_ptr -= runtime->boundary;
	}
end:
	return;
}

static void fsi_dma_callback(void *arg)
{
	struct fsi_audio_stream *s = arg;
	struct snd_pcm_substream *substream = s->stream;
	struct snd_pcm_runtime *runtime;

	if (s->stream == NULL)
		return;

	runtime = substream->runtime;

	if (!s->last_flag) {
		s->dma_pos += s->dma_size[s->tail];
		s->tail = !s->tail;
		if (s->dma_pos >=
			frames_to_bytes(runtime, runtime->buffer_size))
			s->dma_pos = 0;
	}

	/*
	 * If we are getting a callback for an active stream then we inform
	 * the PCM middle layer we've finished a period
	 */
	if (s->active)
		/* maybe clear "output_ptr = 0" in dma_stop() */
		snd_pcm_period_elapsed(s->stream);
	else
		return;

	if (PLAY_STREAM(substream) &&
		(runtime->control->appl_ptr == s->output_ptr)) {
		;
	} else
		fsi_process_dma(s);

	return;
}

static int fsi_dma_start(struct fsi_audio_stream *s, dma_addr_t buff, u32 size)
{
	dma_cookie_t cookie;
	struct dma_async_tx_descriptor *desc;
	struct scatterlist sg;
	struct snd_soc_dai *dai = GET_CPU_DAI(s->stream);

	sg_init_table(&sg, 1);
	sg_set_page(&sg, pfn_to_page(PFN_DOWN(buff)),
			size, offset_in_page(buff));
	sg_dma_address(&sg) = buff;

	desc = s->chan->device->device_prep_slave_sg(s->chan,
		&sg, 1, s->direction, DMA_PREP_INTERRUPT | DMA_CTRL_ACK);
	if (!desc) {
		dev_err(dai->dev,
			"%s : Failed to allocate a dma descriptor\n",
			__func__);
		return -ENOMEM;
	}

	desc->callback = fsi_dma_callback;
	desc->callback_param = s;
	cookie = desc->tx_submit(desc);
	if (cookie < 0) {
		dev_err(dai->dev,
			"%s : Failed to submit a dma transfer\n",
			__func__);
		return cookie;
	}
	s->dma_size[s->head] = size;
	s->head = !s->head;
	s->desc = desc;
	s->cookie = cookie;

	dma_async_issue_pending(s->chan);

	return 0;
}




static u_int audio_get_dma_pos(struct fsi_audio_stream *s)
{
	struct snd_pcm_runtime *runtime = s->stream->runtime;

	return bytes_to_frames(runtime, s->dma_pos);
}

/*
 * this stops the dma and clears the dma ptrs
 */
static void fsi_dma_end(struct fsi_audio_stream *s)
{
	unsigned long flags;

	spin_lock_irqsave(&s->dma_lock, flags);
	s->output_ptr = 0;
	s->dma_pos = 0;
	s->last_flag = 0;
	s->old_offset = 0;
	spin_unlock_irqrestore(&s->dma_lock, flags);

	return;
}


/* playback interrupt */
static int fsi_data_push(struct fsi_priv *fsi, int startup)
{
	struct snd_pcm_runtime *runtime;
	struct snd_pcm_substream *substream = NULL;
	struct fsi_audio_stream *s;
	u32 status;
	int over_period;

	if (!fsi ||
		!fsi->s[0].stream	||
		!fsi->s[0].stream->runtime) {
		printk(KERN_INFO "%s(%d):NULL\n", __func__, __LINE__);
		return -EINVAL;
	}
	over_period	= 0;
	substream	= fsi->s[0].stream;
	runtime		= substream->runtime;
	s = &fsi->s[0];

	status = fsi_reg_read(fsi, DOFF_ST);
	if (!startup) {
		struct snd_soc_dai *dai = fsi_get_dai(substream);

		if (status & ERR_OVER)
			dev_err(dai->dev, "%s over run\n", __func__);
		if (status & ERR_UNDER)
			dev_err(dai->dev, "under run\n");
	}
	fsi_reg_write(fsi, DOFF_ST, 0);

	if (over_period)
		snd_pcm_period_elapsed(substream);

	return 0;
}

static int fsi_data_pop(struct fsi_priv *fsi, int startup)
{
	struct snd_pcm_runtime *runtime;
	struct snd_pcm_substream *substream = NULL;
	struct fsi_audio_stream *s;
	u32 status;
	int over_period;

	if (!fsi			||
		!fsi->s[1].stream	||
		!fsi->s[1].stream->runtime)
		return -EINVAL;

	over_period	= 0;
	substream	= fsi->s[1].stream;
	runtime		= substream->runtime;
	s = &fsi->s[1];

	status = fsi_reg_read(fsi, DIFF_ST);
	if (!startup) {
		struct snd_soc_dai *dai = fsi_get_dai(substream);

		if (status & ERR_OVER)
			dev_err(dai->dev, "%s over run\n", __func__);
		if (status & ERR_UNDER)
			dev_err(dai->dev, "under run\n");
	}
	fsi_reg_write(fsi, DIFF_ST, 0);

	if (over_period)
		snd_pcm_period_elapsed(substream);

	return 0;
}

static irqreturn_t fsi_interrupt(int irq, void *data)
{
	struct fsi_master *master = data;
	u32 int_st = fsi_irq_get_status(master);

	/* clear irq status */
	fsi_master_mask_set(master, SOFT_RST, IR, 0);
	fsi_master_mask_set(master, SOFT_RST, IR, IR);

	if (int_st & INT_A_OUT)
		fsi_data_push(&master->fsia, 0);
	if (int_st & INT_B_OUT)
		fsi_data_push(&master->fsib, 0);
	if (int_st & INT_A_IN)
		fsi_data_pop(&master->fsia, 0);
	if (int_st & INT_B_IN)
		fsi_data_pop(&master->fsib, 0);

	fsi_irq_clear_all_status(master);

	return IRQ_HANDLED;
}

/************************************************************************


		dai ops


************************************************************************/
static int fsi_dai_startup(struct snd_pcm_substream *substream,
				struct snd_soc_dai *dai)
{
	struct fsi_priv *fsi = fsi_get_priv(substream);
	const char *msg;
	u32 flags = fsi_get_info_flags(fsi);
	u32 flags2 = fsi_get_info_flags2(fsi);
	struct fsi_master *master = fsi_get_master(fsi);
	int is_porta = fsi_is_port_a(fsi);
	u32 fmt;
	u32 reg;
	u32 data;
	u32 mask;
	int is_play = (substream->stream == SNDRV_PCM_STREAM_PLAYBACK);
	int is_master;
	int ret = 0;
	struct fsi_audio_stream *s;

	struct snd_pcm_runtime *runtime;

	runtime = substream->runtime;

	pm_runtime_get_sync(dai->dev);

	fsi_master_write(fsi->master, SWAP_SEL, WORD_SWAP);

	if (is_play)
		fsi_reg_write(fsi, OUT_DMAC, 0x21);	/* DMA on */
	else
		fsi_reg_write(fsi, IN_DMAC, 0x21);	/* DMA on */

	/* Set Master, LR clock, bit clock */
	mask = is_play ? (1 << 0) : (1 << 4);
	is_master = fsi_is_master_mode(fsi, is_play);
	if (is_master) {
		data = mask;
		mask |= SH_FSI_LRCLK_VALID_MASK | SH_FSI_BITCLK_VALID_MASK;
		data |= (flags2 & (SH_FSI_LRCLK_VALID_MASK | SH_FSI_BITCLK_VALID_MASK));
	} else
		data = 0;

	fsi_reg_mask_set(fsi, CKG1, mask, data);

	/* clock inversion (CKG2) */
	data = 0;
	if (SH_FSI_LRM_INV & flags)
		data |= 1 << 12;
	if (SH_FSI_BRM_INV & flags)
		data |= 1 << 8;
	if (SH_FSI_LRS_INV & flags)
		data |= 1 << 4;
	if (SH_FSI_BRS_INV & flags)
		data |= 1 << 0;

	fsi_reg_write(fsi, CKG2, data);

	/* do fmt, di fmt */
	data = 0;
	reg = is_play ? DO_FMT : DI_FMT;
	fmt = is_play ? SH_FSI_GET_OFMT(flags) : SH_FSI_GET_IFMT(flags);
	switch (fmt) {
	case SH_FSI_FMT_MONO:
		msg = "MONO";
		data = CR_FMT(CR_MONO);
		fsi->chan = 1;
		break;
	case SH_FSI_FMT_MONO_DELAY:
		msg = "MONO Delay";
		data = CR_FMT(CR_MONO_D);
		fsi->chan = 1;
		break;
	case SH_FSI_FMT_PCM:
		msg = "PCM";
		data = CR_FMT(CR_PCM);
		fsi->chan = 2;
		break;
	case SH_FSI_FMT_I2S:
		msg = "I2S";
		data = CR_FMT(CR_I2S);
		fsi->chan = 2;
		break;
	case SH_FSI_FMT_TDM:
		msg = "TDM";
		fsi->chan = is_play ?
			SH_FSI_GET_CH_O(flags) : SH_FSI_GET_CH_I(flags);
		data = CR_FMT(CR_TDM) | (fsi->chan - 1);
		break;
	case SH_FSI_FMT_TDM_DELAY:
		msg = "TDM Delay";
		fsi->chan = is_play ?
			SH_FSI_GET_CH_O(flags) : SH_FSI_GET_CH_I(flags);
		data = CR_FMT(CR_TDM_D) | (fsi->chan - 1);
		break;
	case SH_FSI_FMT_SPDIF:
		if (master->regs->ver < 2) {
			dev_err(dai->dev, "This FSI can not use SPDIF\n");
			return -EINVAL;
		}
		data = CR_SPDIF;
		fsi->chan = 2;
		fsi_spdif_clk_ctrl(fsi, 1);
		fsi_reg_mask_set(fsi, OUT_SEL, 0x0010, 0x0010);
		break;
	default:
		msg = "-";
		dev_err(dai->dev, "unknown format.\n");
		return -EINVAL;
	}

	fsi_reg_write(fsi, reg, data);

	/*
	 * clear clk reset if master mode
	 */
	if (is_master)
		fsi_clk_ctrl(fsi, 1);

	/* irq clear */
	fsi_irq_disable(fsi, is_play);
	fsi_irq_clear_status(fsi);

	/* fifo init */
	fsi_fifo_init(fsi, is_play, dai);

	/* dma_slave is NULL by kzalloc, only one time. */
	s = &fsi->s[SNDRV_PCM_STREAM_PLAYBACK];
	if (s->dma_slave == NULL) {
		if (is_porta)
			s->dma_slave = &fsi->master->info->dma->dma_porta_tx;
		else
			s->dma_slave = &fsi->master->info->dma->dma_portb_tx;
		s->direction = DMA_TO_DEVICE;
		s->chan = NULL;
		spin_lock_init(&s->dma_lock);
		INIT_WORK(&s->trigger_work, fsi_trigger_work);
	}
	s = &fsi->s[SNDRV_PCM_STREAM_CAPTURE];
	if (s->dma_slave == NULL) {
		s->dma_slave = &fsi->master->info->dma->dma_porta_rx;
		s->direction = DMA_FROM_DEVICE;
		s->chan = NULL;
		spin_lock_init(&s->dma_lock);
		INIT_WORK(&s->trigger_work, fsi_trigger_work);
	}

	return ret;
}

static void fsi_dai_shutdown(struct snd_pcm_substream *substream,
				struct snd_soc_dai *dai)
{
	struct fsi_priv *fsi = fsi_get_priv(substream);
	int is_play = substream->stream == SNDRV_PCM_STREAM_PLAYBACK;

	fsi_irq_disable(fsi, is_play);
	pm_runtime_put_sync(dai->dev);
}


static int fsi_dai_trigger(struct snd_pcm_substream *substream, int cmd,
				struct snd_soc_dai *dai)
{
	struct fsi_priv *fsi = fsi_get_priv(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	int is_play = substream->stream == SNDRV_PCM_STREAM_PLAYBACK;
	int ret = 0;

	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
		fsi_stream_push(fsi, substream,
				frames_to_bytes(runtime, runtime->buffer_size),
				frames_to_bytes(runtime, runtime->period_size));
		break;
	case SNDRV_PCM_TRIGGER_STOP:
		fsi_irq_disable(fsi, is_play);
		fsi_stream_pop(fsi, substream);
		break;
	}

	return ret;
}

static int fsi_dai_hw_params(struct snd_pcm_substream *substream,
			     struct snd_pcm_hw_params *params,
			     struct snd_soc_dai *dai)
{
	struct fsi_priv *fsi = fsi_get_priv(substream);
	struct fsi_master *master = fsi_get_master(fsi);
	int (*set_rate)(int is_porta, int rate) = master->info->set_rate;
	int fsi_ver = master->regs->ver;
	int is_play = (substream->stream == SNDRV_PCM_STREAM_PLAYBACK);
	int ret;

	/* if slave mode, set_rate is not needed */
	if (!fsi_is_master_mode(fsi, is_play))
		return 0;

	/* it is error if no set_rate */
	if (!set_rate)
		return -EIO;

	ret = set_rate(fsi_is_port_a(fsi), params_rate(params));
	if (ret > 0) {
		u32 data = 0;

		switch (ret & SH_FSI_ACKMD_MASK) {
		default:
			/* FALL THROUGH */
		case SH_FSI_ACKMD_512:
			data |= (0x0 << 12);
			break;
		case SH_FSI_ACKMD_256:
			data |= (0x1 << 12);
			break;
		case SH_FSI_ACKMD_128:
			data |= (0x2 << 12);
			break;
		case SH_FSI_ACKMD_64:
			data |= (0x3 << 12);
			break;
		case SH_FSI_ACKMD_32:
			if (fsi_ver < 2)
				dev_err(dai->dev, "unsupported ACKMD\n");
			else
				data |= (0x4 << 12);
			break;
		}

		switch (ret & SH_FSI_BPFMD_MASK) {
		default:
			/* FALL THROUGH */
		case SH_FSI_BPFMD_32:
			data |= (0x0 << 8);
			break;
		case SH_FSI_BPFMD_64:
			data |= (0x1 << 8);
			break;
		case SH_FSI_BPFMD_128:
			data |= (0x2 << 8);
			break;
		case SH_FSI_BPFMD_256:
			data |= (0x3 << 8);
			break;
		case SH_FSI_BPFMD_512:
			data |= (0x4 << 8);
			break;
		case SH_FSI_BPFMD_16:
			if (fsi_ver < 2)
				dev_err(dai->dev, "unsupported ACKMD\n");
			else
				data |= (0x7 << 8);
			break;
		}

		fsi_reg_mask_set(fsi, CKG1, (ACKMD_MASK | BPFMD_MASK) , data);
		udelay(10);
		fsi_clk_ctrl(fsi, 1);
		ret = 0;
	}
	return ret;

}

static struct snd_soc_dai_ops fsi_dai_ops = {
	.startup	= fsi_dai_startup,
	.shutdown	= fsi_dai_shutdown,
	.trigger	= fsi_dai_trigger,
	.hw_params	= fsi_dai_hw_params,
};

/************************************************************************


		pcm ops


************************************************************************/
static struct snd_pcm_hardware fsi_pcm_hardware = {
	.info =		SNDRV_PCM_INFO_INTERLEAVED	|
			SNDRV_PCM_INFO_MMAP		|
			SNDRV_PCM_INFO_MMAP_VALID	|
			SNDRV_PCM_INFO_PAUSE,
	.formats		= FSI_FMTS,
	.rates			= FSI_RATES,
	.rate_min		= 8000,
	.rate_max		= 192000,
	.channels_min		= 1,
	.channels_max		= 2,
	.buffer_bytes_max	= 64 * 1024,
	.period_bytes_min	= 32,
	.period_bytes_max	= 8192,
	.periods_min		= 1,
	.periods_max		= 32,
	.fifo_size		= 256,
};

static int fsi_pcm_open(struct snd_pcm_substream *substream)
{
	struct snd_pcm_runtime *runtime = substream->runtime;
	struct fsi_priv *fsi = fsi_get_priv(substream);
	int ret = 0;
	unsigned long flags;

	snd_soc_set_runtime_hwparams(substream, &fsi_pcm_hardware);

	ret = snd_pcm_hw_constraint_integer(runtime,
					SNDRV_PCM_HW_PARAM_PERIODS);

	spin_lock_irqsave(&fsi->master->lock, flags);
	if ((fsi->master->irq_enable)++ == 0)
		enable_irq(fsi->src.irq);
	spin_unlock_irqrestore(&fsi->master->lock, flags);

	return ret;
}

static int fsi_hw_params(struct snd_pcm_substream *substream,
			 struct snd_pcm_hw_params *hw_params)
{
	return snd_pcm_lib_malloc_pages(substream,
					params_buffer_bytes(hw_params));
}

static int fsi_hw_free(struct snd_pcm_substream *substream)
{
	return snd_pcm_lib_free_pages(substream);
}

static snd_pcm_uframes_t fsi_pointer(struct snd_pcm_substream *substream)
{
	struct fsi_priv *fsi = fsi_get_priv(substream);
	struct fsi_audio_stream *s = &fsi->s[substream->stream];

	return audio_get_dma_pos(s);
}

static inline void fsi_stop_dma(struct fsi_audio_stream *s)
{
	if (s->chan != NULL) {
		/* DMA stop */
		s->chan->device->device_control(s->chan, DMA_TERMINATE_ALL, 0);

		/* DMA release */
		dma_release_channel(s->chan);
		s->chan = NULL;
	}
}

static int fsi_pcm_close(struct snd_pcm_substream *substream)
{
	struct fsi_priv *fsi = fsi_get_priv(substream);
	struct fsi_audio_stream *s = &fsi->s[substream->stream];
	unsigned long flags;

	spin_lock_irqsave(&fsi->master->lock, flags);
	if (--(fsi->master->irq_enable) == 0)
		disable_irq(fsi->src.irq);
	spin_unlock_irqrestore(&fsi->master->lock, flags);

	if (cancel_work_sync(&s->trigger_work))
		fsi_stop_dma(s);

	s->stream = NULL;

	return 0;
}

static int fsi_pcm_prepare(struct snd_pcm_substream *substream)
{
	struct fsi_priv *fsi = fsi_get_priv(substream);
	struct fsi_audio_stream *s = &fsi->s[substream->stream];

	s->output_ptr = 0;
	s->period_byte = 0;
	s->dma_pos = 0;
	s->head = 0;
	s->tail = 0;
	s->last_flag = 0;

	return 0;
}

static int fsi_pcm_trigger(struct snd_pcm_substream *substream, int cmd)
{
	struct fsi_priv *fsi = fsi_get_priv(substream);
	struct fsi_audio_stream *s = &fsi->s[substream->stream];
	int err = 0;

	switch (cmd) {
	case SNDRV_PCM_TRIGGER_PAUSE_RELEASE:
	case SNDRV_PCM_TRIGGER_RESUME:
	case SNDRV_PCM_TRIGGER_START:
		s->active = 1;
		break;

	case SNDRV_PCM_TRIGGER_STOP:
		/* requested stream shutdown */
		s->active = 0;
		fsi_dma_end(s);
		break;

	case SNDRV_PCM_TRIGGER_SUSPEND:
		s->active = 0;
		s->old_offset = audio_get_dma_pos(s) + 1;
		break;

	case SNDRV_PCM_TRIGGER_PAUSE_PUSH:
		s->active = 0;
		break;

	default:
		return -EINVAL;
	}

	schedule_work(&s->trigger_work);

	return err;
}

static int audio_dma_status(struct fsi_audio_stream *s)
{
	struct dma_tx_state txstate;
	enum dma_status status;

	status = s->chan->device->device_tx_status(s->chan, s->cookie,
							&txstate);

	return status == DMA_IN_PROGRESS;
}

static inline int fsi_start_dma(struct fsi_audio_stream *s)
{
	if (s->chan == NULL) {
		/* get DMA */
		dma_cap_mask_t mask;
		dma_cap_zero(mask);
		dma_cap_set(DMA_SLAVE, mask);

		s->chan = dma_request_channel(mask, filter, s->dma_slave);
		if (!s->chan) {
			printk(KERN_ERR
				"%s : unable to get audio dma\n",
				__func__);
			return -EBUSY;
		}
	} else if (audio_dma_status(s)) {
		/* DMA stop */
		s->chan->device->device_control(s->chan, DMA_TERMINATE_ALL, 0);
	}

	fsi_process_dma(s);

	return 0;
}

static void fsi_trigger_work(struct work_struct *work)
{
	struct fsi_audio_stream *s =
		container_of(work, struct fsi_audio_stream, trigger_work);

	if (s->active)
		fsi_start_dma(s);
	else
		fsi_stop_dma(s);
	return;
}

static struct snd_pcm_ops fsi_pcm_ops = {
	.open		= fsi_pcm_open,
	.close		= fsi_pcm_close,
	.ioctl		= snd_pcm_lib_ioctl,
	.hw_params	= fsi_hw_params,
	.hw_free	= fsi_hw_free,
	.prepare	= fsi_pcm_prepare,
	.trigger	= fsi_pcm_trigger,
	.pointer	= fsi_pointer,
};

/************************************************************************


		snd_soc_platform


************************************************************************/
#define PREALLOC_BUFFER		(32 * 1024)
#define PREALLOC_BUFFER_MAX	(32 * 1024)

static void fsi_pcm_free(struct snd_pcm *pcm)
{
	snd_pcm_lib_preallocate_free_for_all(pcm);
}

static int fsi_pcm_new(struct snd_card *card,
			struct snd_soc_dai *dai,
			struct snd_pcm *pcm)
{
	int ret = 0;
	struct snd_soc_pcm_runtime *rtd = pcm->private_data;
	struct snd_soc_dai *cpu_dai = rtd->dai->cpu_dai;
	struct fsi_priv *fsi = cpu_dai->private_data;

	if (fsi->in_use) {
		ret = -EBUSY;
		goto out;
	}

	card->dev->coherent_dma_mask = 0xffffffff;
	ret = snd_pcm_lib_preallocate_pages_for_all(pcm,
		SNDRV_DMA_TYPE_DEV, card->dev, 128 * 1024, 128 * 1024);

	fsi->in_use = 1;
out:
	return ret;
}

/************************************************************************


		alsa struct


************************************************************************/
struct snd_soc_dai fsi_soc_dai[] = {
	{
		.name			= "FSIA",
		.id			= 0,
		.playback = {
			.rates		= FSI_RATES,
			.formats	= FSI_FMTS,
			.channels_min	= 1,
			.channels_max	= 8,
		},
		.capture = {
			.rates		= FSI_RATES,
			.formats	= FSI_FMTS,
			.channels_min	= 1,
			.channels_max	= 8,
		},
		.ops = &fsi_dai_ops,
	},
	{
		.name			= "FSIB",
		.id			= 1,
		.playback = {
			.rates		= FSI_RATES,
			.formats	= FSI_FMTS,
			.channels_min	= 1,
			.channels_max	= 8,
		},
		.capture = {
			.rates		= FSI_RATES,
			.formats	= FSI_FMTS,
			.channels_min	= 1,
			.channels_max	= 8,
		},
		.ops = &fsi_dai_ops,
	},
};
EXPORT_SYMBOL_GPL(fsi_soc_dai);

struct snd_soc_platform fsi_soc_platform = {
	.name		= "fsi-pcm",
	.pcm_ops	= &fsi_pcm_ops,
	.pcm_new	= fsi_pcm_new,
	.pcm_free	= fsi_pcm_free,
};
EXPORT_SYMBOL_GPL(fsi_soc_platform);

/************************************************************************


		platform function


************************************************************************/
static int fsi_probe(struct platform_device *pdev)
{
	struct fsi_master *master;
	const struct platform_device_id	*id_entry;
	struct resource *res;
	unsigned int irq;
	int ret;

	id_entry = pdev->id_entry;
	if (!id_entry) {
		dev_err(&pdev->dev, "unknown fsi device\n");
		return -ENODEV;
	}
	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	irq = platform_get_irq(pdev, 0);
	if (!res || (int)irq <= 0) {
		dev_err(&pdev->dev, "Not enough FSI platform resources.\n");
		ret = -ENODEV;
		goto exit;
	}

	master = kzalloc(sizeof(*master), GFP_KERNEL);
	if (!master) {
		dev_err(&pdev->dev, "Could not allocate master\n");
		ret = -ENOMEM;
		goto exit;
	}

	master->base = ioremap_nocache(res->start, resource_size(res));
	if (!master->base) {
		ret = -ENXIO;
		dev_err(&pdev->dev, "Unable to ioremap FSI registers.\n");
		goto exit_kfree;
	}

	master->irq		= irq;
	master->irq_enable	= 0;	/* disable */
	master->info		= pdev->dev.platform_data;
	master->regs		= (struct fsi_regs *)id_entry->driver_data;
	spin_lock_init(&master->lock);

	master->fsia.base	= master->base;
	master->fsia.master	= master;
	master->fsia.mst_ctrl	= A_MST_CTLR;

	master->fsib.base	= master->base + 0x40;
	master->fsib.master	= master;
	master->fsib.mst_ctrl	= B_MST_CTLR;

	pm_runtime_enable(&pdev->dev);
	pm_runtime_resume(&pdev->dev);

	/* FSIA */
	fsi_soc_dai[0].dev		= &pdev->dev;
	fsi_soc_dai[0].private_data	= &master->fsia;
	/* FSIB */
	fsi_soc_dai[1].dev		= &pdev->dev;
	fsi_soc_dai[1].private_data	= &master->fsib;

	fsi_soft_all_reset(master);

	ret = request_irq(irq, &fsi_interrupt, IRQF_DISABLED,
				id_entry->name, master);
	if (ret) {
		dev_err(&pdev->dev, "irq request err\n");
		goto exit_iounmap;
	}
	disable_irq(irq);

	ret = snd_soc_register_platform(&fsi_soc_platform);
	if (ret < 0) {
		dev_err(&pdev->dev, "cannot snd soc register\n");
		goto exit_free_irq;
	}

	ret = snd_soc_register_dais(fsi_soc_dai, ARRAY_SIZE(fsi_soc_dai));
	if (ret < 0) {
		dev_err(&pdev->dev, "can't register dais.\n");
		goto exit_free_irq;
	}

	return 0;


exit_free_irq:
	free_irq(irq, master);
exit_iounmap:
	iounmap(master->base);
	pm_runtime_disable(&pdev->dev);
exit_kfree:
	kfree(master);
	master = NULL;
exit:
	return ret;
}

static int fsi_remove(struct platform_device *pdev)
{
	struct fsi_master *master;

	master = fsi_get_master(fsi_soc_dai[0].private_data);

	snd_soc_unregister_dais(fsi_soc_dai, ARRAY_SIZE(fsi_soc_dai));
	snd_soc_unregister_platform(&fsi_soc_platform);

	pm_runtime_disable(&pdev->dev);

	free_irq(master->irq, master);

	iounmap(master->base);
	kfree(master);

	fsi_soc_dai[0].dev		= NULL;
	fsi_soc_dai[0].private_data	= NULL;
	fsi_soc_dai[1].dev		= NULL;
	fsi_soc_dai[1].private_data	= NULL;

	return 0;
}

static int fsi_runtime_nop(struct device *dev)
{
	/* Runtime PM callback shared between ->runtime_suspend()
	 * and ->runtime_resume(). Simply returns success.
	 *
	 * This driver re-initializes all registers after
	 * pm_runtime_get_sync() anyway so there is no need
	 * to save and restore registers here.
	 */
	return 0;
}

static struct dev_pm_ops fsi_pm_ops = {
	.runtime_suspend	= fsi_runtime_nop,
	.runtime_resume		= fsi_runtime_nop,
};

static struct fsi_regs fsi_regs = {
	.ver	= 1,
	.int_st	= INT_ST,
	.iemsk	= IEMSK,
	.imsk	= IMSK,
};

static struct fsi_regs fsi2_regs = {
	.ver	= 2,
	.int_st	= CPU_INT_ST,
	.iemsk	= CPU_IEMSK,
	.imsk	= CPU_IMSK,
};

static struct platform_device_id fsi_id_table[] = {
	{ "sh_fsi",	(kernel_ulong_t)&fsi_regs },
	{ "sh_fsi2",	(kernel_ulong_t)&fsi2_regs },
};

static struct platform_driver fsi_driver = {
	.driver		= {
		.name	= "sh_fsi",
		.pm	= &fsi_pm_ops,
	},
	.probe		= fsi_probe,
	.remove		= fsi_remove,
	.id_table	= fsi_id_table,
};

static int __init fsi_mobile_init(void)
{
	return platform_driver_register(&fsi_driver);
}

static void __exit fsi_mobile_exit(void)
{
	platform_driver_unregister(&fsi_driver);
}
module_init(fsi_mobile_init);
module_exit(fsi_mobile_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("SuperH onchip FSI audio driver");
MODULE_AUTHOR("Renesas Electronics");
