/*
 *  File Name       : RSPF0_FMSIDRV_API.c
 *  Function        : SPU OPEN SOUND I/F
 *  Release Version : Ver 1.00
 *  Release Date    : 2011/11/30
 *
 *  Copyright (C) 2011 Renesas Electronics Corporation
 *
 *  This program is free software;you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by Free
 *  Softwere Foundation; either version 2 of License, or (at your option) any
 *  later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A
 *  PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program;
 *  If not, write to the Free Software Foundation, Inc., 51 Franklin St,
 *  Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
#include "../include/RSPF0_FMSIDRV_API.h"
#include "RSPF0_FMSIDRV_REG.h"


enum {
	RSPF0_FMSIDRV_PORT_CLOSE = 0,
	RSPF0_FMSIDRV_PORT_OPEN = 1
};
enum {
	RSPF0_FMSIDRV_IO_NONE = 0,
	RSPF0_FMSIDRV_IO_SET  = 1
};

struct RSPF0_FMSIDRV_Port_ {
	unsigned long opened;
	unsigned long port;
	RSPF0_FMSIDRV_Clock clock_param;
	RSPF0_FMSIDRV_PortParam port_param;
	RSPF0_FMSIDRV_IoParam io_param;
	unsigned long io_setflag;
};

static void (*fmsidrv_reg_read)(unsigned long *addr, unsigned long *data,
	unsigned long size);
static void (*fmsidrv_reg_write)(unsigned long *addr, unsigned long *data,
	unsigned long size);
static void (*fmsidrv_wait)(unsigned long usec);
static unsigned long fmsidrv_address_offset;	/* Support port S only. */
static RSPF0_FMSIDRV_Port fmsidrv_ports[RSPF0_FMSIDRV_DIRECTION_MAX];
							/* 0:SOUT, 1:SIN */

static void write_word(unsigned long addr, unsigned long data);
static void read_word(unsigned long addr, unsigned long *data);
static long check_port_index(RSPF0_FMSIDRV_Port *port);

/*-------------------------------------------------------
Name    : RSPF0_FMSIDRV_Init()
Outline : Initialize FMSI Driver.
Input   : *param      : Pointer of initialing infomation
Output  : error code
-------------------------------------------------------*/
long RSPF0_FMSIDRV_Init(
						RSPF0_FMSIDRV_InitParam * param
						)
{
	unsigned long i;

	/* parameter check */
	if (param == 0 || param->read == 0 || param->write == 0 ||
	param->wait == 0)
		return RSPF0_FMSIDRV_RTN_ERROR_PARAM;

	/* Initial fsidrv_port */
	fmsidrv_address_offset = param->address_offset;

	/* Set other parameters and structure in RSPF0_FMSIDRV_OpenPort(). */
	for (i = 0; i < RSPF0_FMSIDRV_DIRECTION_MAX; i++) {
		fmsidrv_ports[i].opened = RSPF0_FMSIDRV_PORT_CLOSE;
		fmsidrv_ports[i].io_setflag = RSPF0_FMSIDRV_IO_NONE;
		fmsidrv_ports[i].port = 0;	/* 0 is provisionally set. */
	}

	fmsidrv_reg_read = param->read;
	fmsidrv_reg_write = param->write;
	fmsidrv_wait = param->wait;

	return RSPF0_FMSIDRV_RTN_SUCCESS;
}

/*-------------------------------------------------------
Name    : get_port_by_id()
Outline : Get pointer of fsidrv_ports by fsi_id
		and port_id.
Input   : fsi_id      : FSI ID Number(0x00-0x03)
	: port_id     : Port Number(A:0, B:1)
Output  : pointer of fsidrv_ports
-------------------------------------------------------*/
static RSPF0_FMSIDRV_Port *get_port_by_id(unsigned long direction)
{
	return &fmsidrv_ports[direction];
}

/*-------------------------------------------------------
Name    : RSPF0_FMSIDRV_OpenPort()
Outline : Open FMSI Port to use.
Input   : *clock      : Pointer of master clock number
	: direction   : Set object (SOUT:0, SIN:1)
	: **port      : Pointer of Port setting parameter
Output  : error code
-------------------------------------------------------*/
long RSPF0_FMSIDRV_OpenPort(
						RSPF0_FMSIDRV_Clock *clock,
						unsigned long direction,
						RSPF0_FMSIDRV_Port **port
						)
{
	RSPF0_FMSIDRV_Port *p;

	if (clock == 0 || port == 0 || direction >= RSPF0_FMSIDRV_DIRECTION_MAX)
		return RSPF0_FMSIDRV_RTN_ERROR_PARAM;

	/* check pointer & state error */
	p = get_port_by_id(direction);
	if (p->opened)
		return RSPF0_FMSIDRV_RTN_ERROR_STATE;
	*port = p;

	/*memset(&(p->port_param), 0, sizeof(RSPF0_FMSIDRV_PortParam));*/
	p->port_param.audio_clock_mode = 0;
	p->port_param.bit_per_fs_mode = 0;
	p->port_param.master.bclk_inverse = 0;
	p->port_param.master.lrclk_inverse = 0;
	p->port_param.slave.bclk_inverse = 0;
	p->port_param.slave.lrclk_inverse = 0;

	/*memset(&(p->io_param[0]), 0, sizeof(RSPF0_FMSIDRV_IoParam));*/
	p->io_param.bws = 0;
	p->io_param.format = 0;
	p->io_param.msmode = 0;
	p->io_param.tdm_channel = 0;

	/*copy clock mode;*/
	p->clock_param.master_mode = clock->master_mode;
	p->clock_param.slave_mode  = clock->slave_mode;

	p->io_setflag = RSPF0_FMSIDRV_IO_NONE;
	p->port       = direction;
	p->opened     = RSPF0_FMSIDRV_PORT_OPEN;

	return RSPF0_FMSIDRV_RTN_SUCCESS;
}

/*-------------------------------------------------------
Name    : RSPF0_FMSIDRV_ClosePort()
Outline : Close FMSI Port to use.
Input   : port      : Pointer of Port setting parameter
Output  : error code
-------------------------------------------------------*/
long RSPF0_FMSIDRV_ClosePort(RSPF0_FMSIDRV_Port *port)
{
	long result;

	result = check_port_index(port);
	if (result != 0)
		return RSPF0_FMSIDRV_RTN_ERROR_PARAM;

	if (port->opened == RSPF0_FMSIDRV_PORT_CLOSE)
		return RSPF0_FMSIDRV_RTN_ERROR_STATE;
	port->opened = RSPF0_FMSIDRV_PORT_CLOSE;

	return RSPF0_FMSIDRV_RTN_SUCCESS;
}

/*-------------------------------------------------------
Name    : RSPF0_FMSIDRV_SetPortParam()
Outline : Set parameters of FSI Port.
Input   : *port      : Pointer of Port setting parameter
	: *param     : Pointer of parameters
Output  : error code
-------------------------------------------------------*/
long RSPF0_FMSIDRV_SetPortParam(
						RSPF0_FMSIDRV_Port *port,
						RSPF0_FMSIDRV_PortParam *param
						)
{
	long result;

	result = check_port_index(port);
	if (result != 0)
		return RSPF0_FMSIDRV_RTN_ERROR_PARAM;
	if (port->opened == RSPF0_FMSIDRV_PORT_CLOSE)
		return RSPF0_FMSIDRV_RTN_ERROR_STATE;
	if (param == 0)
		return RSPF0_FMSIDRV_RTN_ERROR_PARAM;
	if (param->audio_clock_mode >= RSPF0_FMSIDRV_AUDIOCLK_MAX    ||
		 param->bit_per_fs_mode >= RSPF0_FMSIDRV_BITPERFS_MAX      ||
		 param->master.lrclk_inverse >= RSPF0_FMSIDRV_POLARITY_MAX ||
		 param->master.bclk_inverse >= RSPF0_FMSIDRV_POLARITY_MAX  ||
		 param->slave.lrclk_inverse >= RSPF0_FMSIDRV_POLARITY_MAX  ||
		 param->slave.bclk_inverse >= RSPF0_FMSIDRV_POLARITY_MAX
	)
		return RSPF0_FMSIDRV_RTN_ERROR_PARAM;

	/*memcpy(&port->port_param, param, sizeof(RSPF0_FMSIDRV_PortParam));*/
	port->port_param.audio_clock_mode     = param->audio_clock_mode;
	port->port_param.bit_per_fs_mode      = param->bit_per_fs_mode;
	port->port_param.master.bclk_inverse  = param->master.bclk_inverse;
	port->port_param.master.lrclk_inverse = param->master.lrclk_inverse;
	port->port_param.slave.bclk_inverse   = param->slave.bclk_inverse;
	port->port_param.slave.lrclk_inverse  = param->slave.lrclk_inverse;

	return RSPF0_FMSIDRV_RTN_SUCCESS;
}

/*-------------------------------------------------------
Name    : RSPF0_FMSIDRV_SetIoParam()
Outline : Set individual parameters for I/O of FMSI Port.
Input   : *port      : Pointer of Port setting parameter
	: *param     : Pointer of individual parameters
Output  : error code
Remarks : It is necessary to execute this API function
		before RSPF0_FMSIDRV_StartPort() is executed.
-------------------------------------------------------*/
long RSPF0_FMSIDRV_SetIoParam(
						RSPF0_FMSIDRV_Port *port,
						RSPF0_FMSIDRV_IoParam *param
						)
{
	long result;

	result = check_port_index(port);
	if (result != 0)
		return RSPF0_FMSIDRV_RTN_ERROR_PARAM;
	if (port->opened == RSPF0_FMSIDRV_PORT_CLOSE)
		return RSPF0_FMSIDRV_RTN_ERROR_STATE;
	if (param == 0)
		return RSPF0_FMSIDRV_RTN_ERROR_PARAM;

	if (param->format >= RSPF0_FMSIDRV_FORMAT_MAX ||
	((param->format == RSPF0_FMSIDRV_FORMAT_TDM ||
	param->format == RSPF0_FMSIDRV_FORMAT_TDM_DELAY) &&
	param->tdm_channel >= RSPF0_FMSIDRV_TDM_CHANNEL_MAX) ||
	param->msmode >= RSPF0_FMSIDRV_MSMODE_MAX ||
	param->bws >= RSPF0_FMSIDRV_BWS_MAX)
		return RSPF0_FMSIDRV_RTN_ERROR_PARAM;

	/* set param->tdm_channel to 0 when param->format = FSIDRV_FORMAT_TDM or
	FSIDRV_FORMAT_TDM_DELAY. */
	/* This is hardware Limitation of DO_FMT/DI_FMT */
	if (!(param->format == RSPF0_FMSIDRV_FORMAT_TDM ||
	param->format == RSPF0_FMSIDRV_FORMAT_TDM_DELAY))
		param->tdm_channel = 0;

	/*memcpy(&port->io_param[direction], param,
					sizeof(RSPF0_FMSIDRV_IoParam));*/
	port->io_param.bws         = param->bws;
	port->io_param.format      = param->format;
	port->io_param.msmode      = param->msmode;
	port->io_param.tdm_channel = param->tdm_channel;

	port->io_setflag = RSPF0_FMSIDRV_IO_SET;

	return RSPF0_FMSIDRV_RTN_SUCCESS;
}

/*-------------------------------------------------------
Name    : RSPF0_FMSIDRV_ResetPort()
Outline : Reset FSI Port.
Input   : *port      : Pointer of Port setting parameter
Output  : error code
-------------------------------------------------------*/
long RSPF0_FMSIDRV_ResetPort(
							RSPF0_FMSIDRV_Port *port
							)
{
	unsigned long swrst, clkrst;
	unsigned long reg;
	long result;

	result = check_port_index(port);
	if (result != 0)
		return RSPF0_FMSIDRV_RTN_ERROR_PARAM;
	if (port->opened == RSPF0_FMSIDRV_PORT_CLOSE)
		return RSPF0_FMSIDRV_RTN_ERROR_STATE;
	if (port->port == RSPF0_FMSIDRV_DIRECTION_SOUTPUT) {
		swrst  = 0x00000100;
		clkrst = 0x00000001;
	} else{	/* port->port == RSPF0_FMSIDRV_DIRECTION_SINTPUT */
		swrst  = 0x00001000;
		clkrst = 0x00000010;
	}

	/* Port software reset */
	read_word(fmsidrv_address_offset + RSPF0_FMSIDRV_REG_SOFT_RST, &reg);
	reg &= ~swrst;
	write_word(fmsidrv_address_offset + RSPF0_FMSIDRV_REG_SOFT_RST, reg);

	/* Weight of bit clock */
	fmsidrv_wait(10); /* The longest is 7.82usec (in Fs=8kHz, 16bit/fs). */

	/* Port clock reset */
	read_word(fmsidrv_address_offset + RSPF0_FMSIDRV_REG_CLK_RST, &reg);
	reg &= ~clkrst;
	write_word(fmsidrv_address_offset + RSPF0_FMSIDRV_REG_CLK_RST, reg);

	/* Weight of audio clock */
	fmsidrv_wait(250); /* The longest is 125usec (FS 8kHz) */

	/* Port software reset release */
	read_word(fmsidrv_address_offset + RSPF0_FMSIDRV_REG_SOFT_RST, &reg);
	reg |= swrst;
	write_word(fmsidrv_address_offset + RSPF0_FMSIDRV_REG_SOFT_RST, reg);


	return RSPF0_FMSIDRV_RTN_SUCCESS;
}

/*-------------------------------------------------------
Name    : RSPF0_FMSIDRV_StartPort()
Outline : Start FMSI Port.
Input   : *port      : Pointer of Port setting parameter
Output  : error code
-------------------------------------------------------*/
long RSPF0_FMSIDRV_StartPort(
							RSPF0_FMSIDRV_Port *port
							)
{
	unsigned long clkrst;
	unsigned long reg;
	long result;

	result = check_port_index(port);
	if (result != 0)
		return RSPF0_FMSIDRV_RTN_ERROR_PARAM;
	if (port->opened == RSPF0_FMSIDRV_PORT_CLOSE)
		return RSPF0_FMSIDRV_RTN_ERROR_STATE;

	if (port->io_setflag == RSPF0_FMSIDRV_IO_NONE)
		return RSPF0_FMSIDRV_RTN_ERROR_STATE;

	if (port->port == RSPF0_FMSIDRV_DIRECTION_SOUTPUT) {	/* SOUTPUT */
		/* SOMI_CKG_MD */
		reg = (0x01 << 20) |		/* MSUSYNCS = Serial LR Clock */
			((port->port_param.audio_clock_mode & 0x7) << 12) |
			((port->port_param.bit_per_fs_mode & 0x7) << 8) |
			(0x00 << 4) |			/* DIPMD = Slave */
			(port->io_param.msmode & 0x1);
		write_word(fmsidrv_address_offset +
		RSPF0_FMSIDRV_REG_SOMI_CKG_MD, reg);
		/* SOMI_CKG_RV */
		reg = ((port->port_param.master.lrclk_inverse & 0x1) << 12) |
			((port->port_param.master.bclk_inverse & 0x1) << 8) |
			((port->port_param.slave.lrclk_inverse & 0x1) << 4) |
			(port->port_param.slave.bclk_inverse & 0x1);
		write_word(fmsidrv_address_offset +
		RSPF0_FMSIDRV_REG_SOMI_CKG_RV, reg);
		/* CKG_SMD */
		reg = ((port->clock_param.master_mode & 0x01) << 20) |
			((port->clock_param.slave_mode & 0x01) << 16);
		write_word(fmsidrv_address_offset +
		RSPF0_FMSIDRV_REG_CKG_SMD, reg);
		/* S_DO_FMT */
		reg = (((port->io_param.bws & 0x01) << 20) |
			((port->io_param.format & 0x7) << 4) |
			(port->io_param.tdm_channel & 0x7));
		write_word(fmsidrv_address_offset +
		RSPF0_FMSIDRV_REG_S_DO_FMT, reg);
		/* DRCT_SRL */
		reg = 0;			/* SIMD:0(Serial input mode),
						 SOMD:0(Serial output mode) */
		write_word(fmsidrv_address_offset +
		RSPF0_FMSIDRV_REG_DRCT_SRL, reg);
	} else{		/* SINPUT */
		/* SIMO_CKG_MD */
		reg = (0x00 << 20) |	/* MSUOG = Gated
					 (fixed to low) until SI-IF */
			((port->port_param.audio_clock_mode & 0x7) << 12) |
			((port->port_param.bit_per_fs_mode & 0x7) << 8) |
			(0x00 << 4) |			/* DOPMD = Slave */
			(port->io_param.msmode & 0x1);
		write_word(fmsidrv_address_offset +
		RSPF0_FMSIDRV_REG_SIMO_CKG_MD, reg);
		/* SIMO_CKG_RV */
		reg = ((port->port_param.master.lrclk_inverse & 0x1) << 12) |
			((port->port_param.master.bclk_inverse & 0x1) << 8) |
			((port->port_param.slave.lrclk_inverse & 0x1) << 4) |
			(port->port_param.slave.bclk_inverse & 0x1);
		write_word(fmsidrv_address_offset +
		RSPF0_FMSIDRV_REG_SIMO_CKG_RV, reg);
		/* CKG_SMD */
		reg = ((port->clock_param.master_mode & 0x01) << 20) |
			((port->clock_param.slave_mode & 0x01) << 16);
		write_word(fmsidrv_address_offset +
		RSPF0_FMSIDRV_REG_CKG_SMD, reg);
		/* S_DI_FMT */
		reg = (((port->io_param.bws & 0x01) << 20) |
			((port->io_param.format & 0x7) << 4) |
			(port->io_param.tdm_channel & 0x7));
		write_word(fmsidrv_address_offset +
		RSPF0_FMSIDRV_REG_S_DI_FMT, reg);
		/* DRCT_SRL */
		reg = 0;			/* SIMD:0(Serial input mode),
						 SOMD:0(Serial output mode) */
		write_word(fmsidrv_address_offset +
		RSPF0_FMSIDRV_REG_DRCT_SRL, reg);
	}

	/* CLK_RST */
	if (port->port == RSPF0_FMSIDRV_DIRECTION_SOUTPUT)
							/* SO-IF Clock Reset */
		clkrst = 0x00000001;
	else						/* SI-IF Clock Reset */
		clkrst = 0x00000010;
	read_word(fmsidrv_address_offset + RSPF0_FMSIDRV_REG_CLK_RST, &reg);
	reg |= clkrst;
	write_word(fmsidrv_address_offset + RSPF0_FMSIDRV_REG_CLK_RST, reg);

	return RSPF0_FMSIDRV_RTN_SUCCESS;
}

/*-------------------------------------------------------
Name    : RSPF0_FMSIDRV_SoftwareReset()
Outline : Softwarereset FMSI Port.
Input   : none
Output  : error code
Note    :
-------------------------------------------------------*/
long RSPF0_FMSIDRV_SoftwareReset(void)
{

	/* Int Reset/Soft Reset*/
	write_word(fmsidrv_address_offset + RSPF0_FMSIDRV_REG_SOFT_RST,
								0x00001100);
	write_word(fmsidrv_address_offset + RSPF0_FMSIDRV_REG_SOFT_RST,
								0x00001111);
					/* Interrupt Reset & Software Reset */

	return RSPF0_FMSIDRV_RTN_SUCCESS;
}

/*-------------------------------------------------------
Name    : write_word()
Outline : Write data to FMSI in 4 byte.
Input   : addr      : address for writting
	: data      : data for writting
Output  : none
-------------------------------------------------------*/
static void write_word(unsigned long addr, unsigned long data)
{
	fmsidrv_reg_write((unsigned long *)addr, &data, 1);	/* pgr0703 */
}

/*-------------------------------------------------------
Name    : read_word()
Outline : Read data to FMSI in 4 byte.
Input   : addr      : address for reading
	: data      : data for reading
Output  : none
-------------------------------------------------------*/
static void read_word(unsigned long addr, unsigned long *data)
{
	fmsidrv_reg_read((unsigned long *)addr, data, 1);	/* pgr0703 */
}

/*-------------------------------------------------------
Name    : check_port_index()
Outline : Check FSI Port
Input   : *port   : Pointer of Port setting parameter
Output  : none
-------------------------------------------------------*/
static long check_port_index(RSPF0_FMSIDRV_Port *port)
{
	long i;

	if (port) {
		for (i = 0; i < RSPF0_FMSIDRV_DIRECTION_MAX; i++) {
			if (&fmsidrv_ports[i] == port)
				return 0;
		}
	}
	return -1;						/* error */
}

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("SPU OPEN SOUND I/F Ver 1.0");
MODULE_AUTHOR("Renesas Electronics");
